{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@ame autofocus}
{@description A tool to implement autofocus.
{@longdescription
Calculates sharpness of an image and interpolates the best focus position using
a position history. Optimum is considered to be found if distance to previous
optimum is less then the resolution defined by -r.

Output is either "next <pos>" to test a new position or "final <pos>" if final
position has been found.

This algorithm works with varying seeing conditions but requires exactly reproducible
focus position.}

{$I directives.pas}
uses simage,getparam,urmatrix2;

const af_abs_limit_fact = 0.985;
      af_hist_size      = 12;   { history size }

var par                 : tparams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    mi                  : CSImg_MetaInfo;
    fi,fh               : ansistring;
    bps                 : shortstring;
    pos,res,th,shf,zw   : double;
    sn                  : longint;
    pp,alimit,shfd      : boolean;

    bi                  : TSImg_BayerInfo;

    t                   : text;
    fe                  : boolean;
    s,ss                : ansistring;
    x,y,y0,w            : double;
    ymax                : double;       { absolute maximum multiplied with af_abs_limit_fact }
    pos1,pos2,posa      : double;       { solution of history and at current position, average position }
    pos1v,pos2v         : boolean;      { pos1/pos2 valid }
    smax                : double;       { maximum focuas step }
    a                   : array[0..4] of double;
    b                   : array[0..2] of double;
    zt                  : array[-2..2] of boolean; { untested zones }
    i,j,line            : longint;
    n                   : longint;      { number of usable measurements }
    m                   : longint;      { number of measurements }
    v                   : matrix;
    hist_x,hist_y       : array[0..af_hist_size] of double;     { index 0 is reserved for current position }

label finish;
begin
    with par do
    begin
        init;
        paramusage:='<histfile> [<infile>]';
                        defparam('<histfile>','Focus history file',@fh,tpastr,rapdreq);
        fi:='';         defparam('<infile>','Source image (default: stdin)',fi);
                        defparam('-h','Help',helpf);
                        defparam('-p','Current position (required)',@pos,tpdouble,rapdreq);
        res:=1.0;       defparam('-r','Resolution limit of focus position used as stop condition (default: 1)',res);
        zw:=7;          defparam('-z','Zone width (default: 7, recommended: 5 to 9 times the resolution)',zw);
        th:=1;          defparam('-t','Threshold quantile for sharpness calculation (default 1.0; valid: 0..1; 0.999..1 for stars)',th);
        pp:=false;      defparam('-pp','Position must be positive',pp);
        alimit:=false;  defparam('-a','Final position must (close to) absolute maximum',alimit);
        sn:=1;          defparam('-s','Sharpness algorithm (default: 1)\\'+
                                      '<2: Edge detect for coarse focusing and if sensor resolution is at least twice as high as optical resolution\\'+
                                      ' 2: Hires edge detect, for fine focusing and if optical resolution is close to or higher then sensor resolution\\'+
                                      '>2: Peak detect, for fine focusing and if optical resolution is close to or higher then sensor resolution',sn);
        shf:=1.2;       defparam('-f','Use star FWHM instead of sharpness indicators mentioned above. Parameter is the minimum FWHM. \\'+
                                      'As larger as better the noise immunity. Values less then 1 should be avoided.',shf);
        bps:='';        defparam('-bp','Bayer pattern for sharpness calculation',bps);
        run;
        done;
        shfd:=vardef(@shf);
    end;
    res:=abs(res);
    zw:=abs(zw);
    if zw<4*res then zw:=4*res;
    zw+=1e-6*res;  // prevent rounding issues
    smax:=zw*4;

    cc:=CSImg_Config.create(nil);
    cc.debug('autofocus started');

    { read history file and calculate previous maximum }
    assign(t,fh);
    fe:=fileexists(fh);
    m:=0;
    ymax:=0;
    hist_x[0]:=0; hist_y[0]:=0;         { makes the compiler happy }
    if fe then
    begin
        reset(t);
        line:=0;
        while (ioresult=0) and (not eof(t)) do
        begin
            line+=1;
            readln(t,s);
            s:=trim(s);
            if (length(s)>0) and (s[1]<>'#') then
            begin
                i:=2;
                while (i<length(s)) and (s[i]>' ') do i+=1;
                if i>=length(s) then cc.error('Syntax error at line '+int2str(line)+': Two numbers separated by space expected.');
                ss:=trim(copy(s,1,i-1));
                val(ss,x,j);
                if j<>0 then cc.error('Syntax error at line '+int2str(line)+': Not a number: "'+ss+'"');
                ss:=trim(copy(s,i+1,length(s)-i));
                val(ss,y,j);
                if j<>0 then cc.error('Syntax error at line '+int2str(line)+': Not a number: "'+ss+'"');
                if shfd then y:=1/sqr(y);
                if y>ymax then ymax:=y;
                ymax*=af_abs_limit_fact;
                x-=pos;                         { positions are relative to actual position }
                if m<af_hist_size then m+=1
                else begin
                    for j:=1 to af_hist_size-1 do
                    begin
                        hist_x[j]:=hist_x[j+1];
                        hist_y[j]:=hist_y[j+1];
                    end;
                end;
                hist_x[m]:=x;
                hist_y[m]:=y;
            end;
        end;
        close(t);
    end;

    { calculate old maximum -> pos1 }
    n:=0;
    for i:=0 to 4 do a[i]:=0;
    for i:=0 to 2 do b[i]:=0;
    for i:=1 to m do
    begin
        x:=hist_x[i];
        y:=hist_y[i];
        if abs(x)<=0.75*zw then w:=1    { weight }
        else w:=0.75*zw/abs(x);
        if w>0.1 then n+=1;
        for j:=0 to 4 do
        begin
            if j<3 then b[j]+=w*y;
            a[j]+=w;
            w*=x;
        end;
    end;
    v.pt:=nil;
    pos1v:=false;
    pos1:=0;
    posa:=a[1]/(a[0]+0.1);
    if (a[0]>0.3) and (n>2) then
    begin
        am(v,solve( rmatrix(3,3, [ a[0], a[1], a[2],
                                   a[1], a[2], a[3],
                                   a[2], a[3], a[4] ]),
                    rmatrix(1,3, [ b[0], b[1], b[2] ]) ));
        if v.pt[2]<-1e-30 then
        begin
            pos1v:=true;
            pos1:=-v.pt[1]/(2*v.pt[2]);
        end;
    end;

    { read image and calculate sharpness }
    mi:=CSImg_MetaInfo.create();

    cc.Info('autofocus: Reading input image '+fi);
    SImg_ReadImageFile(cc,fi,img,mi);
    if (bps='') and mi.get('BAYERPAT',s) then bps:=trimquotes(s);
    if bps<>'' then
    begin
        bi.cf[0]:=1;
        bi.cf[1]:=1;
        bi.cf[2]:=1;
        bi.cf[3]:=1;
        bi.pattern:=SImg_BayerParsePattern(cc,bps);
        if not SImg_BayerCheckInfo(bi)
        then cc.warning('autofocus: Invalid Bayer pattern')
        else begin
            cc.Info('autofocus: Converting Bayer pattern image');
            SImg_BayerDownscale(cc,bi,img,img);
        end;
    end;

    if shfd then y0:=SImg_GetSharpnessFWHM(cc,img,shf,round(shf*20))
    else if sn<2 then y0:=SImg_GetSharpness1(cc,img, th)
    else if sn=2 then y0:=SImg_GetSharpness2(cc,img, th)
    else y0:=SImg_GetSharpness3(cc,img, th);
    y:=y0;
    if shfd then
    begin
        if (y<0.25) or (y>shf*20) then { invalid sharpness }
        begin
            x:=0;
            if pos1v then
            begin
                s:='Using optimum found so far';
                x:=pos1;
            end
            else begin
                y:=1e-100;
                for i:=1 to m do
                    if hist_y[i]>y then
                    begin
                        y:=hist_y[i];
                        x:=hist_x[i];
                    end;
                if abs(x)>1e-10 then s:='Using best position found so far'
                else s:='Using last position'
            end;
            cc.warning(cc.name+': Unable to calculate sharpness. '+s);
            if abs(x)<=1e-10 then writeln('final')
            else writeln('final '+real2str(x+pos,5));
            goto finish;
        end;
        if y<0.7 then y:=0.7;
        y:=1/sqr(y);
    end;
    SImg_Free(cc,img);
    cc.info('autofocus: Sharpness at '+real2str(pos,5)+': '+real2str(y0,5));

    { calculate new maximum -> pos2 }
    hist_x[0]:=0;
    hist_y[0]:=y;
    a[0]+=1;
    b[0]+=y;
    pos2:=0;
    pos2v:=false;
    if (a[0]>1.3) and (n>1) then
    begin
        am(v,solve( rmatrix(3,3, [ a[0], a[1], a[2],
                                   a[1], a[2], a[3],
                                   a[2], a[3], a[4] ]),
                    rmatrix(1,3, [ b[0], b[1], b[2] ]) ));
        if v.pt[2]<-1e-30 then
        begin
            pos2v:=true;
            pos2:=-v.pt[1]/(2*v.pt[2]);
        end;
    end;

    { calculate zone indicators }
    for i:=-2 to 2 do zt[i]:=true;
    for i:=0 to m do
    begin
        x:=hist_x[i]-pos2;
        j:=round(x/zw);
        if j<-2 then j:=-2
        else if j>2 then j:=2;
        zt[j]:=false;
    end;

    { info }
    cc.info('autofocus:'+condstring(pos1v,' oldMaximum='+real2str(pos1+pos,5),'') +
                         condstring(pos2v,' newMaximum='+real2str(pos2+pos,5),'') +
//                         ' n='+int2str(n)+' m='+int2str(m) +
                         ' avg='+real2str(posa+pos,5) +
                         ' zi='+condstring(zt[-2],'-','+')+condstring(zt[-1],'-','+')+condstring(zt[0],'-','+')+condstring(zt[1],'-','+')+condstring(zt[2],'-','+'));

    { new focus step }
    if (not pos2v) and (a[0]>1.1) then
    begin
        am(v,solve( rmatrix(2,2, [ a[0], a[1],
                                   a[1], a[2] ]),
                    rmatrix(1,2, [ b[0], b[1] ]) ));
        if v.pt[1]<0 then pos2:=-smax
        else pos2:=smax
    end
    else if not pos2v then pos2:=smax*0.5
    else if pos2<-smax then pos2:=-smax
    else if pos2>smax then pos2:=smax
    else if abs(pos2)<0.75*zw then
    begin
        if zt[-2] or zt[2] then  { enforce zone -2 or 2 }
        begin
            if (zt[-2]) and ((pos2<0) or not zt[2]) then pos2:=-2*zw
            else pos2:=2*zw
        end
        else if zt[-1] or zt[1] then  { enforce zone -1 or  1 }
        begin
            if (zt[-1]) and ((pos2<0) or not zt[1]) then pos2:=-zw
            else pos2:=zw
        end;
    end;

    { positive constraint }
    if pp then
    begin
        if pos1<-pos then pos1:=-pos;
        if pos2<-pos then pos2:=-pos;
    end;
    { stop condition, output }
    if pos2v and pos1v and (not zt[0]) and (abs(pos2-pos1)<=res) and (abs(pos2)<0.5*zw) and (abs(posa)<0.5*zw) and ((y>ymax) or not alimit) then
    begin
        if alimit then writeln('final')
        else writeln('final '+real2str(pos2+pos,5))
    end
    else begin
        if (pos2<0) and (pos2>-res) then pos2:=-res
        else if (pos2>0) and (pos2<res) then pos2:=res;
        writeln('next '+real2str(pos2+pos,5));
    end;

    { append results }
finish:
    assign(t,fh);
    if fe then append(t)
    else rewrite(t);
    writeln(t,real2str(pos,5),' ',real2str(y0,5));
    close(t);

    cc.destroy;
end.
