{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name fbgapprox}
{@description Backround approximation.}
{@longdescription
Finds an approximation for the background. The approximation size can be defined
using the parameters -s, -xs and -ys. If parameter -f is given the difference between input
and approximation is written, otherwise the approximation.}

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    img                 : TSImg_FPImage;
    mask                : TSImg_RDImage;
    ifn,ofn,mfn         : ansistring;
    sa,fa               : TAAStr;
    fg                  : boolean;
    xyr,cs              : double;
    s,f                 : array of double;
    frad,fquant         : double;
    i,tn                : longint;
    err                 : word;
    mi                  : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';        defparam('<infile>','Input image (default: stdin)',ifn);
        ofn:='';        defparam('<outfile>','Output image (default: stdout)',ofn);
        mfn:='';        defparam('-mask','Mask image calculated by getmask', mfn);
        sa.size:=0;     defparam('-s','Approximation size as radius (default: 100). Can be specified multiple times if -f is used.', sa);
        xyr:=1;         defparam('-a','X/Y ratio approximation size. Value given by -s is the geometric average. Default: 1', xyr);
        cs:=0.5;        defparam('-c','Relative approximation in corners (default: 0.5, feasible: 0..1)', cs);
        frad:=0;        defparam('-m','Radius of quantile pre-filter. Default: 0',frad);
        fquant:=0.5;    defparam('-q','Quantile of quantile pre-filter. Default: 0.5',fquant);
        fa.size:=0;     defparam('-f','Subtraction factor. Can be specified multiple times. Default: approximation is output.',fa);
        fg:=false;      defparam('-fg','Foreground approximation rather then background approximation',fg);
                        defparam('-o','Output image',ofn); deffmt('-o','<outfile>');
                        defparam('-h','Help',helpf);
        run;
        done;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('fbgapprox: values: source='+ifn+' dest='+ofn+' mask='+mfn+' s.size='+int2str(sa.size)+' xyr='+real2str(xyr)+' cs='+real2str(cs)+' f.size='+int2str(fa.size) + ' frad='+real2str(frad) + ' fquant='+real2str(fquant) + ' fg='+boolean2str(fg) );
    if (sa.size<>fa.size) and ((fa.size>0) or (sa.size<>1)) then cc.error('fbgapprox: Number of approximation sizes must be equal to number of subtraction factors');

    { evaluate parameters }
    if sa.size<1 then
    begin
        setlength(s,1);
        s[0]:=100;
    end
    else begin
        setlength(s,sa.size);
        for i:=0 to sa.size-1 do
        begin
            val(sa.aa[i],s[i],err);
            if err>0 then cc.error('fbgapprox: Number expected after -s, got `'+sa.aa[i]+'''');
        end;
    end;
    setlength(f,fa.size);
    for i:=0 to fa.size-1 do
    begin
        val(fa.aa[i],f[i],err);
        if err>0 then cc.error('fbgapprox: Number expected after -f. got: `'+fa.aa[i]+'''');
    end;

    mi:=CSImg_MetaInfo.create;
    img.pt:=nil;
    cc.Info('fbgapprox: Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,img,mi);

    mask.pt:=nil;
    if mfn<>'' then
    begin
        cc.Info('fbgapprox: Reading mask image '+mfn);
        SImg_ReadImageFile(cc,mfn,mask);
    end;

    if frad>=0.95 then tn:=cc.threads
    else tn:=cc.threads;
    cc.Info('fbgapprox: Background approximation using ' +int2str(tn) + ' threads');
    SImg_FBGApprox(cc, img,img,mask, s,xyr,cs, f, frad,fquant, fg);

    cc.Info('fbgapprox: Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,img,mi);

    SImg_Free(cc,img);
    if mask.pt<>nil then SImg_Free(cc,mask);

    cc.destroy;
end.

