{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@ame overlap}
{@description Tests whether two images overlap.}
{@longdescription Tests whether two images overlap. This requires wcs data for both images (only the headers are read).
Exitcode is 0 if images overlap, 10 if not or another value is an error occurred}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    fir,fid             : ansistring;
    rmi,dmi             : CSImg_MetaInfo;
    rwcs,dwcs           : TSImg_WCS;
    ovr,ovd             : double;
    rnx,rny,dnx,dny     : double;
    rri,dri             : TSImg_RDImage;

    gx,gy,xi,yi         : longint;
    ox,oy,dx,dy         : double;
    a,b,ra,dec          : double;
    match               : boolean;



begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        fir:='';         defparam('<infile1>','Reference image (wcs data)',@fir,tpastr,rapdreq);
        fid:='';         defparam('<infile2>','Destination image (wcs data, default: stdin)',fid);
        ovr:=0.05;       defparam('-r','Minumum overlapping area relative to destination. Default: 5%.',ovr);
        ovd:=0;          defparam('-d','Minumum overlapping in degree. Default: 0.',ovd);
                         defparam('-h','Help',helpf);
        run;
        done;
    end;

    cc.debug(cc.name+': in_ref='+fir+' in_dest='+fid+'  ovr='+real2str(ovr));

    { reference }
    rri.pt:=nil;        { only used to read dimensions }
    rmi:=CSImg_MetaInfo.create();
    SImg_RHeader(cc, fir, rri, rmi);
    if not SImg_GetWCSData(cc, rmi, rwcs) then cc.error(cc.name+': Invalid or incomplete WCS data for reference image.');
    rnx:=rri.nx; rny:=rri.ny;
    if rmi.get('IBNDXMAX',a) and rmi.get('IBNDXMIN',b) then rnx:=abs(a-b)+1;
    if rmi.get('IBNDYMAX',a) and rmi.get('IBNDYMIN',b) then rny:=abs(a-b)+1;
    rmi.get('IMAGEW', rnx);
    rmi.get('IMAGEH', rny);
    if ( (rnx<1) or (rny<1) ) then cc.error(cc.name+': No image dimensions given for reference.');

    { destination }
    dri.pt:=nil;        { only used to read dimensions }
    dmi:=CSImg_MetaInfo.create();
    SImg_RHeader(cc, fid, dri, dmi);
    if not SImg_GetWCSData(cc, dmi, dwcs) then cc.error(cc.name+': Invalid or incomplete WCS data for destination image.');
    dnx:=dri.nx; dny:=dri.ny;
    if dmi.get('IBNDXMAX',a) and dmi.get('IBNDXMIN',b) then dnx:=abs(a-b)+1;
    if dmi.get('IBNDYMAX',a) and dmi.get('IBNDYMIN',b) then dny:=abs(a-b)+1;
    dmi.get('IMAGEW', dnx);
    dmi.get('IMAGEH', dny);
    if ( (dnx<1) or (dny<1) ) then cc.error(cc.name+': No image dimensions given for destination.');

    { calculate overlap }
    ox:=sqrt(0.5*dnx*dny*ovr);
    oy:=ox;
    if ox<1 then ox:=1;
    if ovd>0 then
    begin
        SImg_GetPixelScale(dwcs, a, b);
        a:=abs(a); b:=abs(b);
        if (a<1e-3) or (b<1e-3) then cc.error(cc.name+': Invalid pixel scale (less than 0.001 arcsec).');
        a:=ovd*3600/a;
        if a>ox then ox:=a;
        b:=ovd*3600/b;
        if b>oy then oy:=b;
    end;
    a:=0.499*dnx;
    if ox>a then
    begin
        cc.warning(cc.name+': Overlapping in X direction is reduced from '+real2str(ox,-1)+' to '+real2str(a,-1)+' pixel');
        ox:=a;
    end;
    a:=0.499*dny;
    if oy>a then
    begin
        cc.warning(cc.name+': Overlapping in Y direction is reduced from '+real2str(oy,-1)+' to '+real2str(a,-1)+' pixel');
        oy:=a;
    end;

    { calculate number of grid points }
    gx:=round((dnx-ox*2)/ox+1.5);
    if gx<3 then gx:=3;
    if gx>100  then gx:=100;
    dx:=(dnx-ox*2)/(gx-1);
    gy:=round(2*(dny-oy*2)/oy+1.5);
    if gy<3 then gy:=3;
    if gy>100  then gy:=100;
    dy:=(dny-oy*2)/(gy-1);

//    writeln(stderr,'ox=',ox:0:1,' gx=',gx,'  oy=',oy:0:1,' gy=',gy,'   size=',dnx:0:1,'x',dny:0:1);

    { test overlapping }
    match:=false;
    for yi:=0 to gy-1 do
        for xi:=0 to gx-1 do
            if (not match) and ( ((xi<>0) and (xi<>gx)) or ((yi<>0) and (yi<>gy)) ) then { ignore corners }
            begin
                SImg_GetRaDec(cc, dwcs, ox+xi*dx, oy+yi*dy, dny, ra,dec);
                if not SImg_GetPos(rwcs, ra,dec,rny, a,b) then cc.error('overlap: transformation matrix of reference is singular');
                match:=(a>=0) and (a<=rnx-1) and (b>=0) and (b<=rny-1);
{               if match then
                begin
                    writeln(stderr,ox+xi*dx:0:1,',',oy+yi*dy:0:1,'  -->  ',ra:0:3,'°, ',dec:0:3, '°  -->  ',a:0:1,', ',b:0:1,'  size=',rnx:0:1,'x',rny:0:1);
                    SImg_GetRaDec(cc, rwcs, a,b, rny, ra,dec);
                    writeln(stderr,a:0:1,',',b:0:1, '  -->  ', ra:0:3,'°,',dec:0:3,'°');
                end;}
            end;
//    writeln(stderr);
    if not match then cc.info('overlap: No match for '+fid);
    rmi.destroy;
    dmi.destroy;
    cc.destroy;
    if not match then SImg_Halt(10);
end.
