{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name staradd}
{@description Transforms star image and optionally add it to an input image.}

{$I directives.pas}
uses simage,getparam,sysutils;

var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,ofn,dfn,sfn     : ansistring;
    pfn,prfn            : ansistring;
    fi,fs,fp,fpd        : TSImg_FPImage;  { tile images }
    pe,rd,rm,ri,thr     : double;

    i,x,y,xc,yc         : longint;
    p,q,r,s             : double;
    mi                  : CSImg_MetaInfo;
    inf,infd            : TSImg_PSFInfo;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

    with par do
    begin
        init;
        paramusage:='[<outfile>]]';
        ofn:='';     defparam('<outfile>','Destination image (default: stdout)', ofn);
        ifn:='';     defparam('-i','Input image added to star image', ifn); deffmt('-i','<infile>');
        dfn:='';     defparam('-d','Difference image', dfn);                deffmt('-d','<infile>');
        sfn:='';     defparam('-s','Deconvoluted star image', sfn);         deffmt('-s','<infile>');
        pfn:='';     defparam('-p','Input PSF image', pfn);                 deffmt('-p','<infile>');
        prfn:='';    defparam('-pr','Reference PSF image which is used to calculate the desired PSF. By default input PSF is used', prfn);      deffmt('-pr','<infile>');
        pe:=1;       defparam('-e','Exponent used for PSF modification. Default: 1', pe);
        rd:=1;       defparam('-f','Factor for unmodified component of PSF and the difference image. Default: 1', rd);
        rm:=-1;      defparam('-fm','Factor for modified component of the PSF. Default: same as -f', rm);
        ri:=1;       defparam('-fi','Factor for input image. Default: 1', ri);
        thr:=0.01;   defparam('-t','PSF modification threshold relative to peak. Default: 1%',thr);
                     defparam('-o','Output image',ofn);                         deffmt('-o','<outfile>');
                     defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': dfn='+dfn+' sfn='+sfn+' ifn='+ifn+' pfn='+pfn+' prfn='+prfn+' ofn='+ofn+' pe='+real2str(pe)+' rd='+real2str(rd)+' rm='+real2str(rm)+' ri='+real2str(ri)+' thr='+real2str(thr));
    if rm<-1e-10 then rm:=rd;
    if (rd<1e-3) and (rm<1e-3) then cc.error(cc.name+': Both PSF factors (-r and -rm) to small. Nothing to do.');
    if (rd<1e-3) and (dfn<>'') then
    begin
        dfn:='';
        cc.warning('Factor for difference image to small. Difference image is not used');
    end;
    if (abs(ri)<1e-3) and (ifn<>'') then
    begin
        ifn:='';
        cc.warning('Factor for input image to close to zero. Ignoring input image.');
    end;
    if (sfn='') and (dfn='') then cc.error('Neither difference image nor deconvoluted star image is given');
    if (sfn<>'') and (pfn='') then cc.error('Deconvoluted star image cannot be used without PSF');
    if (pfn<>'') and (sfn='') then cc.warning('PSF image cannot be used without deconvoluted star image');

    mi:=CSImg_MetaInfo.create();

    { star image }
    if sfn<>'' then
    begin
        fp.pt:=nil;

        cc.info(cc.name+': Reading PSF '+pfn);
        SImg_ReadImageFile(cc,pfn,fp);
        if fp.nc<>1 then cc.error(cc.name+': PSF must only have one color');
        SImg_GetPSFInfo(cc,fp,inf);

        fpd.pt:=nil;
        if prfn='' then SImg_Duplicate(cc, fp, fpd, true)
        else begin
            cc.info(cc.name+': Reading reference PSF '+prfn);
            SImg_ReadImageFile(cc,prfn,fpd);
            if fpd.nc<>1 then cc.error(cc.name+': Reference PSF must only have one color');
        end;
        SImg_GetPSFInfo(cc,fpd,infd);
//      cc.info('fp : '+SImg_PSFInfoString(inf));
//      cc.info('fpd: '+SImg_PSFInfoString(infd));

        { adjust sizes and centers }
        xc:=round(max(inf.xc,infd.xc));  x:=round(max(fp.nx-inf.xc, fpd.nx-infd.xc))-1;
        yc:=round(max(inf.yc,infd.yc));  y:=round(max(fp.ny-inf.yc, fpd.ny-infd.yc))-1;
        SImg_Cut(cc, fp ,fp , round(inf .xc)-xc, round(inf .yc)-yc,  round(inf .xc)+x, round(inf .yc)+y,  0,0, 0, 0);
        SImg_Cut(cc, fpd,fpd, round(infd.xc)-xc, round(infd.yc)-yc,  round(infd.xc)+x, round(infd.yc)+y,  0,0, 0, 0);
//      writeln(stderr,xc,'   ',yc);

        { calculate destination PSF }
        if (abs(pe-1)>1e-3) and (pe>1e-3) then
        begin
            for i:=0 to fpd.size-1 do
            begin
                r:=fpd.pt[i];
                if r<simg_mina then r:=0
                else r:=exp(ln(r)*pe);
                fpd.pt[i]:=r;
            end;
        end;
        s:=simg_mina;
        for i:=0 to fpd.size-1 do
            s+=fpd.pt[i];
        if s<0 then cc.error(cc.name+': Invalid desired PSF');
        s:=rm/s;
        for i:=0 to fpd.size-1 do
            fpd.pt[i]*=s;

        { lower constraint in order to reduce artifacts if sources are not centered to pixel }
        if dfn<>'' then
        begin
            { calculate restraints }
            fi.pt:=nil;
            SImg_New(cc, fi, fp.nx,fp.ny,1);
            p:=0;
            for y:=1 to fp.ny-2 do
            begin
                i:=y*fp.sl;
                for x:=1 to fp.nx-2 do
                begin
                    q:=fp.pt[i];
                    r:=q;
                    s:=abs(fp.pt[i-fp.sl]-q);  if s<r then r:=s;
                    s:=abs(fp.pt[i+fp.sl]-q);  if s<r then r:=s;
                    s:=abs(fp.pt[i-    1]-q);  if s<r then r:=s;
                    s:=abs(fp.pt[i+    1]-q);  if s<r then r:=s;
                    r*=rd;
                    fi.pt[i]:=r;
                    if fpd.pt[i]<r then fpd.pt[i]:=r;
                    p+=fpd.pt[i];
                    i+=1;
                end;
            end;
            { constraint iterations }
            x:=20;
            while (x>0) and (p-rm>1e-4) and (rm>1e-4) do
            begin
                q:=rm/p;
//              writeln(stderr,q:0:5);
                p:=0;
                for i:=0 to fpd.size-1 do
                begin
                    r:=fpd.pt[i]*q;
                    s:=fi.pt[i];
                    if r<s then r:=s;
                    fpd.pt[i]:=r;
                    p+=r;
                end;
                x-=1;
            end;
            SImg_Free(cc, fi);
        end;
//      SImg_WriteImageFile(cc,'staradd-fpd0.fits',fpd);

        { mix PSF's }
        if thr>1e-12 then
        begin
            for i:=0 to fp.size-1 do
            begin
                r:=fp.pt[i];
                s:=r;
                if s<0 then s:=0;
                fpd.pt[i]:=r*(fpd.pt[i]+thr*rd)/(s+thr);
            end;
        end;
//      SImg_WriteImageFile(cc,'staradd-fpd1.fits',fpd);

        { correction PSF instead of normal one }
        if dfn<>'' then
        begin
            for i:=0 to fp.size-1 do
                fpd.pt[i]-=rd*fp.pt[i];
//          SImg_WriteImageFile(cc,'staradd-fpd2.fits',fpd);
        end;

        { read deconvoluted star image and convolute it }
        cc.info(cc.name+': Reading deconvoluted star image '+sfn);
        SImg_ReadImageFile(cc, sfn, fs, mi);
        if fs.nc<>1 then cc.error(cc.name+': Deconvoluted star image must be a 1-color image');

        cc.info(cc.name+': Convolution star image '+sfn);
        SImg_Convolute(cc, fs,fpd,fs, xc,yc);

        { read and process difference image }
        if dfn<>'' then
        begin
            fi.pt:=nil;
            cc.info(cc.name+': Reading and processing difference image '+dfn);
            mi.reset;
            SImg_ReadImageFile(cc, dfn, fi, mi);
            if (fi.nx<>fs.nx) or (fi.ny<>fs.ny) or (fi.nc<>1) then cc.error(cc.name+': Difference image and deconvoluted star image must have the same dimensions');
            for i:=0 to fi.size-1 do
            begin
                r:=fs.pt[i] + rd*fi.pt[i];
                if r<0 then r:=0;
                fs.pt[i]:=r;
            end;
            SImg_Free(cc, fi);
        end;
        rd:=1;
        SImg_Free(cc,fp);
        SImg_Free(cc,fpd);
    end else
    begin
        fi.pt:=nil;
        cc.info(cc.name+': Reading difference image '+dfn);
        SImg_ReadImageFile(cc,dfn,fs);
        if fs.nc<>1 then cc.error(cc.name+': Difference image must be a 1-color image');
    end;

    { input image }
    if ifn<>'' then
    begin
        cc.info(cc.name+': Reading and processing input image '+ifn);
        fi.pt:=nil;
        mi.reset;
        SImg_ReadImageFile(cc, ifn, fi, mi);
        if (fi.nx<>fs.nx) or (fi.ny<>fs.ny) or (fi.nc<>1) then cc.error(cc.name+': Input image ans star image (or difference image) must have the same dimensions');
        for i:=0 to fi.size-1 do
            fs.pt[i]:=fs.pt[i]*rd + fi.pt[i]*ri;
        SImg_Free(cc,fi);
    end
    else if abs(rd-1)>1e-4 then
        for i:=0 to fs.size-1 do
            fs.pt[i]*=rd;


    { output image }
    cc.Info(cc.name+': Writing output image '+ofn);
    SImg_WriteImageFile(cc, ofn, fs, mi);
    SImg_Free(cc,fs);

    mi.destroy;
    cc.destroy;
end.
