{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name starsubtract}
{@description Extract stars form image and subtract or interpolate them.}

{$I directives.pas}
uses simage,getparam,sysutils;

var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,mfn,ofn         : ansistring;
    omfn,osfn,opfn,odfn : ansistring;
    fi,fs,fb,fd         : TSImg_FPImage;  { tile images }
    fpsf,fpsfs          : TSImg_FPImage;  { tile PSF, PSF sum }
    fim,fsm,fbm,fdm     : TSImg_FPImage;  { mosaic images }
    rmm,rmo,rm          : TSImg_RDImage;
    tol,tolc,ff,t1,t2   : double;
    ithr,iwdt,bf,sig    : double;
    cwdt                : double;
    rad,getmask         : longint;
    ma,mf               : boolean;
    i,x,y,opfnx,opfny   : longint;
    ti                  : TSImg_TileInfo;
    name                : shortstring;
    ssi                 : TSImg_StarSubtract_Info;
    mi                  : CSImg_MetaInfo;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

    with par do
    begin
        init;
        paramusage:='[<infile>] [<outfile>]';
        ifn:='';     defparam('<infile>', 'Source image (default: stdin)', ifn);
        ofn:='';     defparam('<outfile>','Destination image (default: stdout)', ofn);
        mfn:='';     defparam('-m','Input star mask (bit 1 and 2 / mask 2 and 4) as written by -om.', mfn);  deffmt('-m','<infile>');
        ma:=false;   defparam('-ma','Allow to find new stars if -m is given.',ma);
        mf:=false;   defparam('-mf','Filter stars given by -m. By default they are treated as save stars.',mf);
        rad:=50;     defparam('-r','PSF radius: Default: 50', rad);
        tol:=1.4;    defparam('-t','PSF tolerance as factor in center. Default: 1.4', tol);
        tolc:=0;     defparam('-tc','PSF tolerance as factor in corners. Default: same as -t', tolc);
        ithr:=5;     defparam('-i','Interpolation threshold relative to noise level. Default: 10', ithr);
        iwdt:=0.5;   defparam('-w','Interpolation width. Influences the maximum Interpolation radius. Valid: >0. Default 0.5', iwdt);
        cwdt:=0.5;   defparam('-c','PSF correction width relative to PSF size. Valid: 0..1, Default 0.5', cwdt);
        ff:=0.5;     defparam('-f','Filter strength. (Default: 0.5. Smaller values to consider more objects as stars.', ff);
        bf:=1;       defparam('-b','Noise correction factor. Default and recommended: 1. Detection threshold can be increased manually if noise is not constant.', bf);
        t1:=1;       defparam('-n1','Split into this amount of tiles into direction of larger dimension for overall calculation. (default: 1)',t1);
        t2:=-1001;   defparam('-n2','Split into this amount of tiles into direction of larger dimension for PSF correction. <=0 to disable\\'+
                                    'Default: 0 if input mask (-m) is given without -ma, otherwise 3',t2);
                     defparam('-o','Output image',ofn);                         deffmt('-o','<outfile>');
        omfn:='';    defparam('-om','Output mask: bit0: interpolation, bit 1: star mask, bit 2: save star',omfn);       deffmt('-om','<outfile>');
        osfn:='';    defparam('-os','Output deconvoluted star image',osfn);                                             deffmt('-os','<outfile>');
        odfn:='';    defparam('-od','Output difference, i.e. convoluted and extrapolated stars',odfn);          deffmt('-od','<outfile>');
        opfn:='';    defparam('-op','Output PSF. %x and %y are replaced by tile indices.', opfn);                       deffmt('-op','<outfile>');
                     defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug('starsubtract: source='+ifn+' mask_in='+mfn + ' out='+ofn+ ' star_out='+osfn+ ' mask_out='+omfn+' psf_out='+opfn+' diff_out='+odfn+
            ' tol='+real2str(tol)+' tolc='+real2str(tolc)+' rad='+real2str(rad)+' ithr='+real2str(ithr)+' iwdt'+real2str(iwdt,-2)+
            ' ff='+real2str(ff)+' bf='+real2str(bf)+' n1='+realstr(t1,-1)+' n2='+real2str(t2,-1)+' ma='+boolean2str(ma));
    if (ofn='') and (omfn='') and (osfn='') and (opfn='') then
    begin
        cc.warning('No output defined.');
        SImg_Halt;
    end;

    if rad<3 then rad:=3;
    name:=cc.name;
    if t1<1 then t1:=1;
    if t2<-1000 then
    begin
        if ma or (mfn='') then t2:=3
        else t2:=0;
    end;

    { input image }
    fim.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info(cc.name+': Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,fim, mi);
    mi.deleteRaw;
    if fim.nc<>1 then cc.error(cc.name+': Only one color images are supported.');

    { tiles }
    if fim.nx>fim.ny then
    begin
        ti.nx:=round(t1);
        ti.ny:=round(fim.ny/fim.nx*t1);
    end
    else begin
        ti.ny:=round(t1);
        ti.nx:=round(fim.nx/fim.ny*t1);
    end;
    if ti.ny<1 then ti.ny:=1;
    if ti.nx<1 then ti.nx:=1;
    ti.bb:=rad;
    ti.bu:=rad;
    if (ti.nx>1) or (ti.ny>1) then cc.info(name+': Splitting image into '+int2str(ti.nx)+' x '+int2str(ti.ny)+' tiles');
    opfnx:=0;
    opfny:=0;

    if opfn<>'' then
    begin
        opfnx:=pos('%x',opfn); if opfnx<0 then opfnx:=0;
        opfny:=pos('%y',opfn); if opfny<0 then opfny:=0;
        if ((ti.nx>1) and (opfnx<=0)) or ((ti.ny>1) and (opfny<=0)) then cc.error(name+': Multiple PSF''s cannot be written to single file. Use %x and %y.');
        if opfnx>opfny then
        begin
            if opfnx>0 then delete(opfn,opfnx,2);
            if opfny>0 then
            begin
                delete(opfn,opfny,2);
                opfnx-=2;
            end;
        end
        else begin
            if opfny>0 then delete(opfn,opfny,2);
            if opfnx>0 then
            begin
                delete(opfn,opfnx,2);
                opfny-=2;
            end;
        end;
    end;

    { input mask }
    getmask:=1;
    rmm.pt:=nil;
    if mfn<>'' then
    begin
        cc.Info(cc.name+': Reading mask '+mfn);
        SImg_ReadImageFile(cc,mfn,rmm);
        if (rmm.fmt<>SImg_RawByte) and (rmm.fmt<>SImg_RawFP) then cc.error(cc.name+': Invalid format of star mask');
        if (rmm.nx<>fim.nx) and (rmm.ny<>fim.ny) or (rmm.nc<>1) then cc.error(cc.name+': Invalid size of star mask');
        if not ma then getmask:=0;
        if mf then getmask:=getmask or 2;
    end;

    { process tiles }
    fi.pt:=nil; rm.pt:=nil; fpsf.pt:=nil;       { tiles images }
    fbm.pt:=nil; fsm.pt:=nil; fdm.pt:=nil;      { output mosaic images }
    fpsf.pt:=nil; fpsfs.pt:=nil;
    rmo:=rmm;                                   { output mask }
    sig:=0;
    for y:=0 to ti.ny-1 do
        for x:=0 to ti.nx-1 do
        begin
            if (ti.nx>1) or (ti.ny>1) then cc.name:=name+'('+int2str(x+1)+','+int2str(y+1)+')';
            SImg_GetTile(cc, ti, fim,fi, x,y);
            if rmm.pt<>nil then SImg_GetTile(cc, ti, rmm,rm, x,y)
            else rm.pt:=nil;

            { calculate PSF }
            {if mfn='' then i:=4
            else }i:=0;
            SImg_GetPSF(cc, fi, fpsf, tol, rad, rm,i);

            { output psf }
            if opfn<>'' then
            begin
                ifn:=opfn;
                if opfnx>opfny then insert(int2str(x+1),ifn,opfnx);
                if opfny>0 then insert(int2str(y+1),ifn,opfny);
                if (opfnx>0) and (opfnx<opfny) then insert(int2str(x+1),ifn,opfnx);
                cc.Info(cc.name+': Writing output psf '+ifn);
                SImg_WriteImageFile(cc,ifn,fpsf);
            end;

            if (omfn<>'') or (ofn<>'') or (osfn<>'') or (odfn<>'') then
            begin
                { find stars }
                fb.pt:=nil; fs.pt:=nil; fd.pt:=nil;
{ Subtract stars
    fi     : input (1 color image)
    fpsf   : input PSF
    rms    : mask input and output. Possible inputs:
        1. undefined
        2. mask
            bit 0: interpolated pixels (output only)
            bit 1: star (input and output)
            bit 2: save star (input and output )
        3. floating point (SImg_RawFP) containing deconvoluted stars that will be brightness filtered
    fob    : output image: background
    fos    : output image: deconvoluted stars
    fod    : output image: difference between input and fob where saturated regions are extrapolated
    ofob   : true: output fob; false: do not touch fob
    ofos   : true: output fos; false: do not touch fos
    ofod   : true: output fod; false: do not touch fod
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars.
    tol    : psf width tolerance as factor in center
    tolc   : psf width tolerance as factor in corners
    ff     : Source filter strength >=0, high values ==> less objects are considered as stars
    nf     : Noise correction factor. Usually 1.
    ithr   : Interpolation threshold relative to noise
    iwdt   : Interpolation width
    cwdt   : PSF correction width relative to PSF. Valid 0..1
    tn     : PSF correction tiles into direction of larger dimension; <0.5 to disable PSF correction
}
                ssi:=SImg_StarSubtract(cc, fi,fpsf,rm, fb,fs,fd, ofn<>'',osfn<>'',odfn<>'', getmask, tol,tolc, ff,bf, ithr,iwdt,cwdt, t2);
                SImg_Free(cc,fi);
                sig+=ssi.sig;

                if (ti.nx>1) or (ti.ny>1) then
                begin
                    if (ti.nx>1) or (ti.ny>1) then
                    if fb.pt<>nil then
                    begin
                        if fbm.pt=nil then SImg_New(cc, fbm, fim.nx,fim.ny,1);
{                       for i:=0 to fb.size-1 do
                            fb.pt[i]*=1+(x+y)*0.5; }
                        SImg_BlendTile(cc, ti, fb,fbm, x,y);
                        SImg_Free(cc,fb);
                    end;
                    if fs.pt<>nil then
                    begin
                        if fsm.pt=nil then SImg_New(cc, fsm, fim.nx,fim.ny,1);
                        SImg_CopyTile(cc, ti, fs,fsm, x,y);
                        SImg_Free(cc,fs);
                    end;
                    if fd.pt<>nil then
                    begin
                        if fdm.pt=nil then SImg_New(cc, fdm, fim.nx,fim.ny,1);
                        SImg_BlendTile(cc, ti, fd,fdm, x,y);
                        SImg_Free(cc,fd);
                    end;
                    if rmo.pt=rmm.pt then rmo.pt:=nil;
                    if rmo.pt=nil then SImg_New(cc, rmo, SImg_RawByte, fim.nx,fim.ny,1, true);
                    SImg_CopyTile (cc, ti, rm,rmo, x,y);
                    SImg_Free(cc,rm);
                end
                else begin
                    SImg_Move(cc, fb,fbm);
                    SImg_Move(cc, fs,fsm);
                    SImg_Move(cc, fd,fdm);
                    SImg_Move(cc, rm,rmo);
                end;
            end
            else begin
                SImg_Free(cc,rm);
                SImg_Free(cc,fi);
            end;

            if fpsfs.pt=nil then
            begin
                fpsfs:=fpsf;
                fpsf.pt:=nil;
            end
            else begin
                for i:=0 to fpsf.size-1 do
                    fpsfs.pt[i]+=fpsf.pt[i];
                SImg_Free(cc, fpsf);
            end;
        end;
    SImg_Free(cc, fim);
    cc.name:=name;
    SImg_Move(cc, rmo,rmm);

    { add meta info }
    sig/=ti.nx*ti.ny;
    t1:=SImg_GetPSFWidth(cc,fpsfs, 0.25)*sqrt(0.5);
    SImg_Free(cc,fpsfs);
    mi.add('NOISEFLR', real2str(sig,10),'Noise floor');
    mi.add('FWHMPIX',real2str(t1,-4), 'Star FWHM in pixels');
    cc.info(cc.name+': Added meta info: NOISEFLR='+real2str(sig)+'  FWHMPIX='+real2str(t1,-2));

    { output image }
    if ofn<>'' then
    begin
        cc.Info(cc.name+': Writing output image '+ofn);
        SImg_WriteImageFile(cc,ofn,fbm,mi);
    end;
    SImg_Free(cc, fbm);

    { output deconvoluted stars }
    if osfn<>'' then
    begin
        cc.Info(cc.name+': Writing deconvoluted star image '+osfn);
        SImg_WriteImageFile(cc,osfn,fsm,mi);
    end;
    SImg_Free(cc, fsm);

    { output convoluted stars }
    if odfn<>'' then
    begin
        cc.Info(cc.name+': Writing difference / convoluted star image '+osfn);
        SImg_WriteImageFile(cc,odfn,fdm,mi);
    end;
    SImg_Free(cc, fdm);

    { output mask }
    if omfn<>'' then
    begin
        cc.Info(cc.name+': Writing mask '+omfn);
        SImg_WriteImageFile(cc,omfn,rmm,mi);
    end;
    SImg_Free(cc, rmm);

    cc.destroy;
end.
