{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name tilemanager}
{@description Manages tile sets. }

{$I directives.pas}
uses simage,getparam,sysutils,math;

const grid = 20;

type string15 = string[15];

function get_ra_n(size,dec:double):longint;
const deg2rad = pi/180;
var r,s         : double;
begin
    r:=dec-0.5*size;
    s:=dec+0.5*size;
    if s*r<0 then r:=0
    else if abs(r)>abs(s) then r:=s;
    r:=pi/arctan(tan(0.5*size*deg2rad)/cos(r*deg2rad));
    result:=round(r+0.499999999);
end;

{ ra2, dec2 : position of tile
  ra,dec    : test position }
function testMatch(size, ra2,dec2, ra,dec:double):boolean;
const deg2rad = pi/180;
var y,z,s,c     : double;
begin
    result:=false;
    size:=sin(0.5*size*deg2rad);
    { 1. rotate -ra2 and convert to spherical coordinates }
    sincos(dec*deg2rad,y,z);
    sincos((ra-ra2)*deg2rad,s,c);
    if abs(s*z)>size then exit;         // x=s*z
    z:=c*z;
    { 2. rotate -dec_ref }
    sincos(dec2*deg2rad,s,c);
    y:=y*c - s*z;
    if abs(y)>size then exit;
    result:=true;
end;

var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,ofn             : ansistring;
    size,d1,d2          : double;
    tout                : text;
    dec_i,dec_n,dec_j   : longint;
    ra_i,ra_j,ra_n      : longint;
    xi,yi,i,j,k,mn      : longint;
    dec,ra,nx,ny,a,b    : double;
    ra_s                : double;
    ma_ra_i,ma_dec_i    : array[0..sqr(grid+1)-1] of smallint;  { list of matching tile indices }
    mi                  : CSImg_MetaInfo;
    wcs                 : TSImg_WCS;
    ri                  : TSImg_RDImage;


begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='';
        size:=2.5;       defparam('-s','Tile size in degree (default: 2.5)',size);
        ifn:='';         defparam('-w','Check in which tiles the given WCS header lies. By default, a tile list is created.',ifn);
        d1:=90;          defparam('-d1','DEC start position in degree, default: 90',d1);
        d2:=-20;         defparam('-d2','DEC end position, default: -20',d2);
                         defparam('-o','Output file',ofn);    deffmt('-o','<outfile>');
                         defparam('-h','Help',helpf);
        run;
        done;
    end;

    cc.debug('tilemanager: size='+real2str(size)+' in='+ifn+' out='+ofn);
    dec_n:=round(abs(d1-d2)/size+0.49);
    d2:=sign(d2-d1);
    if size<0.1 then size:=0.1;

    if (ofn='') or (ofn='-') then
    begin
        flush(output);
        tout:=output;
    end else
    begin
        {$I-}
        assign(tout,ofn);
        rewrite(tout);
        if ioresult<>0 then cc.error('Error opening '+ofn);
        {$I+}
    end;

    if ifn='' then
    begin
        writeln(tout,'# DEC_IDX RA_IDX  DEC     RA');
        for dec_i:=0 to dec_n-1 do
        begin
            dec:=d1+d2*(dec_i+0.5)*size;
            ra_n:=get_ra_n(size,dec);
            ra_s:=360/ra_n;
            for ra_i:=0 to ra_n-1 do
                writeln(tout, dec_i,'   ',ra_i,'        ', dec:0:3,'    ', (ra_i+0.5)*ra_s:0:3);
        end;
    end
    else begin
        ri.pt:=nil;             // dimensions only
        mi:=CSImg_MetaInfo.create;
        SImg_RHeader(cc, ifn, ri, mi);
        if not SImg_GetWCSData(cc, mi, wcs) then cc.error('tilemanager: Invalid or incomplete WCS data.');
        nx:=ri.nx; ny:=ri.ny;
        if mi.get('IBNDXMAX',a) and mi.get('IBNDXMIN',b) then nx:=abs(a-b)+1;
        if mi.get('IBNDYMAX',a) and mi.get('IBNDYMIN',b) then ny:=abs(a-b)+1;
        mi.get('IMAGEW', nx);
        mi.get('IMAGEH', ny);
        if ( (nx<1) or (ny<1) ) then cc.error('tilemanager: No image dimensions given.');
        mn:=0;
        ma_ra_i[0]:=0; ma_dec_i[0]:=0;   // makes the compiler happy
        for yi:=0 to grid do
            for xi:=0 to grid do
            begin
                SImg_GetRaDec(cc,wcs, xi/grid*(nx-1), yi/grid*(ny-1), ny, ra, dec);
                dec_i:=round((dec-d1)*d2/size);  { test dec_i and dec_i-1 }
                for i:=0 to 1 do
                begin
                    dec_j:=dec_i-i;
                    b:=ra;
                    if dec_j>=dec_n then continue
                    else if dec_j<0 then        // possible flip at pole
                    begin
                        b+=180;
                        dec_j:=0;
                    end;
                    a:=d1+d2*(dec_j+0.5)*size;
                    ra_n:=get_ra_n(size, a);
                    ra_s:=360/ra_n;
                    ra_i:=round(b/ra_s);  { test ra_i and ra_i-1 }
                    for j:=0 to 1 do
                    begin
                        ra_j:=(ra_i-j) mod ra_n;
                        if ra_j<0 then ra_j+=ra_n;
                        k:=0;
                        while (k<mn) and ((ma_dec_i[k]<>dec_j) or (ma_ra_i[k]<>ra_j)) do
                            k+=1;
                        if (k>=mn) and testMatch(size, (ra_j+0.5)*ra_s,a, ra,dec) then
                        begin
                            ma_dec_i[mn]:=dec_j;
                            ma_ra_i[mn]:=ra_j;
                            mn+=1;
                        end;
                    end;
                end
            end;

        writeln(tout,'# DEC_IDX RA_IDX');
        for i:=0 to mn-1 do
        begin
            writeln(tout, ma_dec_i[i],' ',ma_ra_i[i]);
        end;
    end;
    close(tout);

end.
