{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name zocorrect}
{@description Bias-/flatfield corrections and calculation of pixel weight image}

{
    photon count image: 2nd pass (unit conversion)
}

{$I directives.pas}
uses simage,getparam;
var par                         : tparams;
    cc                          : CSImg_Config;
    fni,fno,fnw                 : ansistring;
    fnb,fn0,fn1,fnd,fng,fnm     : ansistring;
    ma                          : TAAStr;
    om,bl,fmin                  : double;
    bayer,nc,ue,ues,us,o1,nn    : boolean;

    mi                          : CSImg_MetaInfo;
    fi,f0,f1,fsn,fm             : TSImg_FPImage;
    str                         : ansistring;
    r,s,u,dmin,dmax             : double;
    rn,bn,dn,gain,adr           : double;
    i                           : longint;

begin
    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        fni:='';      defparam('<infile>','Input image (default: stdin)',fni);
        fno:='';      defparam('<outfile>','Output image (default: stdout)',fno);
        fnw:='';      defparam('-w','Output pixel weight image',fnw);                   deffmt('-w','<outfile>');

        fnb:='';      defparam('-b','Bias image',fnb);                                  deffmt('-b','<infile>');
        fn0:='';      defparam('-0','Fixed dark current image',fn0);                    deffmt('-0','<infile>');
        fnd:='';      defparam('-d','Variable dark current image',fnd);                 deffmt('-d','<infile>');
        fng:='';      defparam('-g','Variable amp glow image (only use it if necessary)',fng);                          deffmt('-g','<infile>');
        fn1:='';      defparam('-1','Flatfield image',fn1);                             deffmt('-1','<infile>');
        fnm:='';      defparam('-pm','Dead pixel mask as calculated by avgset -pm',fnm);        deffmt('-pm','<infile>');

        bl:=0.0;      defparam('-bl','Fixed bias level subtracted from image (default: 0)',bl);

        fmin:=-1;     defparam('-s','Power supply noise reduction (parameter is the frequncy limit, recommended: 10..100).',fmin);
        nc:=false;    defparam('-nc','Do not crop values less than -2*<readout noise>',nc);
        bayer:=false; defparam('-bp','Bayer pattern image (default: autodetect)',bayer);

        o1:=false;    defparam('-1o','Flatfield offset correction (ignored if no flatfield defined)',o1);
        om:=0;        defparam('-om','Mask multiplier for overflown pixels (default: 0)',om);
        nn:=false;    defparam('-nn','Disable photon noise estimator.',nn);

        ma.size:=0;   defparam('-m','Add meta info (can be defined multiple times)',ma);        par.deffmt('-m','<key>=<value>[/<comment>]');

        ue:=false;    defparam('-ue','Output image unit is electrons',ue);
        ues:=false;   defparam('-ues','Output image unit is electrons per second',ues);
        us:=false;    defparam('-us','Output image unit is ADU per second',us);

                      defparam('-h','Help',helpf);
                      defparam('-o','Output image',fno); deffmt('-o','<outfile>');
        run;
        done;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('zocorrect: in='+fni+' out='+fno+' weight='+fnw+' bias='+fnb+' fixed_dark='+fn0+' variable_dark='+fnd+' variable_ampglow='+fng+' flat='+fn1+' mask='+fnm+
        ' bl='+real2str(bl)+' om='+real2str(om)+' nc='+boolean2str(nc)+' nn='+boolean2str(nn)+' bayer='+boolean2str(bayer)+' fmin='+real2str(fmin) +
        ' ue='+boolean2str(ue)+' ues='+boolean2str(ues)+' us='+boolean2str(us)+' 1o='+boolean2str(o1));

    i:=0;
    if ue then i+=1;
    if ues then i+=1;
    if us then i+=1;
    if i>1 then cc.error('More than one output units defined');
    o1:=o1 and (fn1<>'');

    { input image }
    fi.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info('zocorrect: Reading input image '+fni);
    SImg_ReadImageFile(cc,fni,fi,mi);

    { meta infos }
    for i:=0 to ma.size-1 do
    begin
        str:=mi.add(trim(ma.aa[i]));
        if str<>'' then  cc.warning('zocorrect: Error parsing -m '+ma.aa[i]+': '+str+ '  (expected syntax: -m <key>=<value>[/<comment>])');
    end;
    if not bayer then bayer:=mi.get('BAYERPAT',str,nil);
    if not mi.get('RDNOISE',rn) then rn:=-1;
    if not mi.get('BGNOISE',bn) then bn:=-1;
    if not mi.get('EGAIN',gain) then gain:=-1
    else if gain<1e-10 then gain:=-1;
    adr:=1/gain;  { ADU/e}

    { initialize square noise image with zeros }
    fsn.pt:=nil;
    if (fnw<>'') or o1 then
    begin
        if fnw<>'' then
        begin
            if rn<-1e-30 then cc.error('zocorrect: Read out noise not defined');
            if gain<1e-10 then cc.error('zocorrect: Gain not defined');
        end;
        SImg_Duplicate(cc,fi,fsn,false);
    end;

    { dead pixel mask }
    fm.pt:=nil;
    if (fsn.pt<>nil) and (fnm<>'') then
    begin
        cc.Info('zocorrect: Reading pixel mask '+fnm);
        SImg_ReadImageFile(cc,fnm,fm);
        if fm.size<>fi.size then cc.error('Invalid size of pixel mask');
    end;
    if (fsn.pt<>nil) and (fm.pt=nil) then
    begin
        SImg_New(cc,fm, fi.nx,fi.ny,fi.nc, false);
        for i:=0 to fm.size-1 do
            fm.pt[i]:=1.0;
    end;

    { overflow/underflow }
    if not mi.get('DATAMIN',dmin) then dmin:=-1e300;
    if not mi.get('DATAMAX',dmax) then dmax:=1e300;
    if (fm.pt<>nil) and ((dmin>-1e299) or (dmax<1e299)) then
    begin
        cc.Info('zocorrect: Calculating overflow/underflow mask');
        for i:=0 to fi.size-1 do
        begin
            r:=fi.pt[i];
            if r<dmin then fm.pt[i]:=0
            else if r>dmax then fm.pt[i]*=om;
        end;
     end;

    { bias }
    f0.pt:=nil;
    if fnb<>'' then
    begin
        cc.Info('zocorrect: Reading and subtracting bias image '+fn0);
        SImg_ReadImageFile(cc,fnb,f0);
        if f0.size<>fi.size then cc.error('Invalid size of bias image');
        for i:=0 to fi.size-1 do
            fi.pt[i]-=f0.pt[i]+bl;
        SImg_Free(cc,f0);
    end else if abs(bl)>1e-30 then
        for i:=0 to fi.size-1 do
            fi.pt[i]-=bl;

    { fixed dark current }
    f0.pt:=nil;
    if fn0<>'' then
    begin
        cc.Info('zocorrect: Reading and subtracting fixed dark current image '+fn0);
        SImg_ReadImageFile(cc,fn0,f0);
        if f0.size<>fi.size then cc.error('Invalid size of fixed dark current image');
        for i:=0 to fi.size-1 do
        begin
            r:=f0.pt[i];
            fi.pt[i]-=r;
            if (fsn.pt<>nil) and (r>0) then fsn.pt[i]+=r*adr;
        end;
        SImg_Free(cc,f0);
    end;

    { variable dark current image }
    f0.pt:=nil; f1.pt:=nil;
    if fnd<>'' then
    begin
        SImg_ReadImageFile(cc,fnd,f0);
        if bayer then i:=2
        else i:=1;
        r:=SImg_DarkCurrentSubtraction(cc, fi,f0, fm,f1, fi, i);
        cc.Info('zocorrect: Variable dark current subtraction with factor '+real2str(r,5));
        if fsn.pt<>nil then
        begin
            r*=adr;
            for i:=0 to fsn.size-1 do
                fsn.pt[i]+=abs(f0.pt[i]*r);
        end;
        SImg_Free(cc,f0);
    end;

    { variable amp glow image }
    f0.pt:=nil; f1.pt:=nil;
    if fng<>'' then
    begin
        SImg_ReadImageFile(cc,fng,f0);
        r:=SImg_AmpGlowSubtraction(cc, fi,f0, fm,f1, fi);
        cc.Info('zocorrect: Amplow subtraction with factor '+real2str(r,5));
        if fsn.pt<>nil then
        begin
            r*=adr;
            for i:=0 to fsn.size-1 do
                fsn.pt[i]+=abs(f0.pt[i]*r);
        end;
        SImg_Free(cc,f0);
    end;

    { cropping to small values }
    if (not nc) and (rn>1e-50) then
    begin
        r:=-rn*2;
        cc.Info('zocorrect: Cropping values less than '+real2str(r,5));
        for i:=0 to fi.size-1 do
            if fi.pt[i]<r then fi.pt[i]:=r;
    end;

    { dark current estimation }
    if fsn.pt=nil then dn:=-1
    else begin
        r:=SImg_GetQuantile(cc,fsn,0.5);
        if r<0 then r:=0;
        dn:=sqrt(r);
        cc.info(cc.name+': Estimated dark current noise: '+real2str(dn)+' ADU = '+real2str(dn*gain)+' e-');
        if (bn>simg_mina) and (dn>bn) then
        begin
            cc.warning(cc.name+': Background noise less than dark current noise. Corrected background noise');
            bn:=dn;
        end;
    end;
//    cc.error(cc.name+': BG noise: '+real2str(SImg_BGNoise(cc,fi,rn,gain))+' ADU  n25='+real2str(sqrt(SImg_GetQuantile(cc,fi,0.25)*adr)) );

    { photon noise estimation }
    if (fsn.pt<>nil) and (rn>1e-30) and (gain>1e-10) then
    begin
        r:=SImg_BGNoise(cc,fi,rn,gain);
        if nn then
            for i:=0 to fsn.size-1 do
                fsn.pt[i]+=r
        else SImg_NoiseEstimator(cc, fi, fsn, r,gain, true);  { does not use bn because it contains dark current noise }
    end;

    { combine mask image (fm) and square noise image (fsn) }
    if (fsn.pt<>nil) and (fm.pt<>nil) then
    begin
        s:=SImg_GetQuantile(cc,fsn,0.25);
        if s<simg_mina then s:=simg_mina;
        u:=1/s;
        for i:=0 to fm.size-1 do
        begin
            r:=fsn.pt[i];
            if r>s then fm.pt[i]/=r
            else fm.pt[i]*=u;
        end;
        SImg_Free(cc,fsn);
    end;

    { image unit }
    u:=1;
    if (ues or us) and (not mi.get('EXPTIME',u)) then cc.warning('zocorrect: Exposure time not defined. Assuming 1s');
    if u<1e-9 then
    begin
        cc.warning('zocorrect: Invalid exposure time. Assuming 1s');
        u:=1;
    end;
    r:=u*gain;
    if ue or ues then
    begin
        r:=u;
        if gain<1e-10 then cc.warning('zocorrect: Gain not defined, assuming 1 (image unit is electrons)')
        else u*=adr;
    end;
    str:='';
    if ues then str:='''electron/s'''
    else if ue then str:='''electron'''
    else if us then str:='''adu/s''';
    if str<>'' then mi.add('BUNIT',str,'Image unit');
    if r>1e-10 then mi.add('EGAIN',real2str(r,12),'ADC resolution in electrons per image unit')
    else mi.delete('EGAIN');
    if bn>1e-50 then
    begin
        bn/=u;
        mi.add('BGNOISE',real2str(bn,15),'Background noise in image units');
    end
    else mi.delete('BGNOISE');
    if rn>1e-50 then
    begin
        rn/=u;
        mi.add('RDNOISE',real2str(rn,15),'Readout noise in image units');
    end
    else mi.delete('RDNOISE');
    if dn>-simg_mina then
    begin
        dn/=u;
        mi.add('DCNOISE',real2str(dn,15),'Dark current noise in image units');
    end
    else mi.delete('DCNOISE');

    { read flatfield (also used for power supply noise reduction) }
    if fn1<>'' then
    begin
        cc.Info('zocorrect: Reading flatfield image '+fn1);
        SImg_ReadImageFile(cc,fn1,f1);
    end;

    { power supply noise reduction }
    f0.pt:=nil;
    if fmin>-1e-10 then
    begin
        cc.Info('zocorrect: Reduction of power supply noise');
        SImg_PSDenoise(cc,fi,f0,f1,fi,fmin,true);
    end;

    { flatfield + unit correction }
    if fn1<>'' then
    begin
        cc.Info('zocorrect: Reading and dividing flatfield image '+fn1);
        SImg_ReadImageFile(cc,fn1,f1);
        if f1.size<>fi.size  then cc.error('zocorrect: invalid dimensions: '+fn1);
        s:=SImg_GetQuantile(cc, f1, 0.99);
        if s<1e-20 then cc.error('zocorrect: non-positive flatfield');
        s:=1/s;
        for i:=0 to fi.size-1 do
        begin
            r:=f1.pt[i]*s;
            if r<0 then r:=0;
            if fm.pt<>nil then
            begin
                if r<1.02 then fm.pt[i]*=sqr(r*u)
                else fm.pt[i]*=sqr(u*1.02*sqr(sqr(1.02/r)));  { this is most likely a hot pixel }
            end;
            if r<1e-2 then r:=1e-2;
            r:=1/(r*u);
            fi.pt[i]*=r;
            if o1 then f1.pt[i]:=r;
        end;
        if o1 then
        begin
            r:=SImg_BGApprox4(cc, fi,f1,fm);
            cc.Info('zocorrect: Subtraction of flatfield offset '+real2str(r));
            for i:=0 to fi.size-1 do
                fi.pt[i]-=r*f1.pt[i];
        end;
        SImg_Free(cc,f1);
    end else
    begin
        r:=1/u;
        s:=sqr(u);
        for i:=0 to fi.size-1 do
            fi.pt[i]*=r;
        r:=sqr(u);
        if fm.pt<>nil then
            for i:=0 to fm.size-1 do
                fm.pt[i]*=r;
    end;

    { output image }
    mi.delete('DATAMIN');
    mi.delete('DATAMAX');
    cc.Info('zocorrect: Writing output image '+fno);
    SImg_WriteImageFile(cc,fno,fi,mi);
    SImg_Free(cc,fi);

    { output weight image }
    if fnw<>'' then
    begin
        cc.Info('zocorrect: Writing pixel weight image '+fnw);
        mi.delete('BGNOISE');
        mi.delete('RDNOISE');
        mi.delete('EGAIN');
        mi.delete('BUNIT');
        SImg_WriteImageFile(cc,fnw,fm,mi);
        SImg_Free(cc,fm);
    end;

    cc.destroy;
end.
