{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[align_debug]

{*******************************************************************************}
{********* SImg_ReadAlignmentParams ********************************************}
{*******************************************************************************}
{ return true if alignment parameters are set }
function SImg_ReadAlignmentParams(cc:CSImg_Config; var res:TSImg_AlignmentParams; mi:CSImg_MetaInfo):boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    res.xmin:=-1000;
    res.xmax:= 1000;
    res.ymin:=-1000;
    res.ymax:= 1000;
    if not mi.get('IBNDXMIN', res.xmin) then cc.warning('SImg_ReadAlignmentParams: Attribute not found: "IBNDXMIN"');
    if not mi.get('IBNDXMAX', res.xmax) then cc.warning('SImg_ReadAlignmentParams: Attribute not found: "IBNDXMAX"');
    if not mi.get('IBNDYMIN', res.ymin) then cc.warning('SImg_ReadAlignmentParams: Attribute not found: "IBNDYMIN"');
    if not mi.get('IBNDYMAX', res.ymax) then cc.warning('SImg_ReadAlignmentParams: Attribute not found: "IBNDYMAX"');

    res.ax:=0; result:=mi.get('ALGNAX', res.ax);
    res.ay:=0; result:=result and mi.get('ALGNAY', res.ay);
    res.b11:=1.0;  mi.get('ALGNB11', res.b11);
    res.b12:=0.0;  mi.get('ALGNB12', res.b12);
    res.b21:=0.0;  mi.get('ALGNB21', res.b21);
    res.b22:=1.0;  mi.get('ALGNB22', res.b22);
    res.c1 :=0.0;  mi.get('ALGNC1',  res.c1);
    res.c2 :=0.0;  mi.get('ALGNC2',  res.c2);
    res.c3 :=0.0;  mi.get('ALGNC3',  res.c3);
    res.c4 :=0.0;  mi.get('ALGNC4',  res.c4);
    res.c5 :=0.0;  mi.get('ALGNC5',  res.c5);
    res.c6 :=0.0;  mi.get('ALGNC6',  res.c6);
    res.d  :=0.0;  mi.get('ALGND',   res.d);
    res.fact:=1;   mi.get('ALGNBSC', res.fact);
    res.weight:=1; mi.get('ALGNWGHT',res.weight);
end;

{*******************************************************************************}
{********* SImg_WriteAlignmentParams *******************************************}
{*******************************************************************************}
procedure SImg_WriteAlignmentParams(const res:TSImg_AlignmentParams; mi:CSImg_MetaInfo);
begin
    mi.add('ALGNAX', real2str(res.ax ,15),'Alignment: Shift in X direction');
    mi.add('ALGNAY', real2str(res.ay ,15),'Alignment: Shift in Y direction');
    mi.add('ALGNB11',real2str(res.b11,15),'Alignment: Component (1,1) of scaling/rotation matrix');
    mi.add('ALGNB12',real2str(res.b12,15),'Alignment: Component (1,2) of scaling/rotation matrix');
    mi.add('ALGNB21',real2str(res.b21,15),'Alignment: Component (2,1) of scaling/rotation matrix');
    mi.add('ALGNB22',real2str(res.b22,15),'Alignment: Component (2,2) of scaling/rotation matrix');
    mi.add('ALGNC1', real2str(res.c1 ,15),'Alignment: 1st component of 1st order distortion');
    mi.add('ALGNC2', real2str(res.c2 ,15),'Alignment: 2dn component of 1st order distortion');
    mi.add('ALGNC3', real2str(res.c3 ,15),'Alignment: 3rd component of 1st order distortion');
    mi.add('ALGNC4', real2str(res.c4 ,15),'Alignment: 4th component of 1st order distortion');
    mi.add('ALGNC5', real2str(res.c5 ,15),'Alignment: 5th component of 1st order distortion');
    mi.add('ALGNC6', real2str(res.c6 ,15),'Alignment: 6th component of 1st order distortion');
    mi.add('ALGND',  real2str(res.d ,15), 'Alignment: 2nd order distortion');
    mi.add('IBNDXMIN',real2str(res.xmin,-1),'Lower image boundary in X direction');
    mi.add('IBNDXMAX',real2str(res.xmax,-1),'Upper image boundary in X direction');
    mi.add('IBNDYMIN',real2str(res.ymin,-1),'Lower image boundary in Y direction');
    mi.add('IBNDYMAX',real2str(res.ymax,-1),'Upper image boundary in Y direction');
    mi.add('ALGNBSC', real2str(res.fact,15),'Brightness scaling');
    mi.add('ALGNWGHT', real2str(res.weight,15),'Weight');
end;


{*******************************************************************************}
{********* SImg_GetAlignmentConstraints ****************************************}
{*******************************************************************************}
{ tries to calculate constraints using wcs data
  rmi : meta info from reference
  dmi : meta info from destination
  tol : tolerance (typically 0.05 to 0.2)
  result: non-empty string on failure containing the reason
}
function SImg_GetAlignmentConstraints(cc:CSImg_Config; rmi,dmi:CSImg_MetaInfo; tol:double; var con:TSImg_AlignmentConstraints):ansistring;
const rad2n   = 1/(2*pi);
var rwcs,dwcs                   : TSImg_WCS;
    rx,ry,rxx,ryx,rxy,ryy,ax,ay : double;
    dxmin,dxmax,dymin,dymax     : double;
    rxmin,rymin,rymax           : double;
begin
    result:='';
    if cc=nil then cc:=SImg_Config_Default;
    tol:=abs(tol);
    if tol<0.05 then tol:=0.05;
    if rmi=nil then
    begin
        result:='No meta information for reference given';
        exit;
    end;
    if dmi=nil then
    begin
        result:='No meta information for destination given';
        exit;
    end;
    if not SImg_GetWCSData(cc, rmi, rwcs) then
    begin
        result:='Reference contains no or invalid WCS data';
        exit;
    end;
    if not SImg_GetWCSData(cc, dmi, dwcs) then
    begin
        result:='Destination contains no or invalid WCS data';
        exit;
    end;
    { read reference image boundaries }
    if not rmi.get('IBNDXMIN', rxmin) then result+=' "IBNDXMIN"';
    if not rmi.get('IBNDYMIN', rymin) then result+=' "IBNDYMIN"';
    if not rmi.get('IBNDYMAX', rymax) then result+=' "IBNDYMAX"';
    if result<>'' then
    begin
        result:='Missing attribute(s) for reference:'+result;
        exit;
    end;
    { read destination image boundaries }
    if not dmi.get('IBNDXMIN', dxmin) then result+=' "IBNDXMIN"';
    if not dmi.get('IBNDXMAX', dxmax) then result+=' "IBNDXMAX"';
    if not dmi.get('IBNDYMIN', dymin) then result+=' "IBNDYMIN"';
    if not dmi.get('IBNDYMAX', dymax) then result+=' "IBNDYMAX"';
    if result<>'' then
    begin
        result:='Missing attribute(s) for destination:'+result;
        exit;
    end;
    { destination image center -> reference }
    SImg_GetRaDec(cc, dwcs, -dxmin,-dymin,dymax-dymin+1, ax,ay);
    if not SImg_GetPos(rwcs, ax,ay, rymax-rymin+1, rx,ry) then
    begin
        result:='Transformation matrix of destination is singular';
        exit;
    end;
    { destination image center + (1,0) -> reference }
    SImg_GetRaDec(cc, dwcs, 1-dxmin,-dymin,dymax-dymin+1, ax,ay);
    if not SImg_GetPos(rwcs, ax,ay, rymax-rymin+1, rxx,ryx) then
    begin
        result:='Transformation matrix of destination is singular';
        exit;
    end;
    { destination image center + (0,1) -> reference }
    SImg_GetRaDec(cc, dwcs, -dxmin,1-dymin,dymax-dymin+1, ax,ay);
    if not SImg_GetPos(rwcs, ax,ay, rymax-rymin+1, rxy,ryy) then
    begin
        result:='Transformation matrix of destination is singular';
        exit;
    end;
    { positions }
    con.rx:=rx+rxmin;
    con.ry:=ry+rymin;
    con.dmax:=sqrt(sqr(dxmax-dxmin)+sqr(dymax-dymin))*tol;
    { angles and scales }
    rxx-=rx; ryx-=ry;
    rxy-=rx; ryy-=ry;
    ax:=arctan2(ryx,rxx)*rad2n;
    ay:=arctan2(ryy,rxy)*rad2n-0.25-ax;
    ay-=round(ay);
    rx:=sqrt(sqr(rxx)+sqr(ryx));
    if rx<1e-12 then rx:=1e-12;
    ry:=sqrt(sqr(rxy)+sqr(ryy));
    if ry<1e-12 then ry:=1e-12;
    con.flip:=abs(ay)>0.25;
    if con.flip then
    begin
        ay+=0.5;
        ay-=round(ay);
    end;
    con.sh:=ay*360*2 + 30*tol;
    con.frc:=-ax*360;
    con.frd:=180*tol;
    con.nq:=abs(rx/ry-1)>0.01;
    rx:=1/rx;
    ry:=1/ry;
    con.smin:=min(rx,ry)*(1-0.5*tol);
    con.smax:=max(rx,ry)*(1+0.5*tol);
//ifdef[align_debug]
    cc.info('SImg_GetAlignmentConstraints: rx='+real2str(con.rx,-1)+' ry='+real2str(con.ry,-1)+' dmax='+real2str(con.dmax,4)+' flip='+boolean2str(con.flip)+' smin='+real2str(con.smin)
        + ' smax=' + real2str(con.smax) + ' nq='+boolean2str(con.nq)+' sh='+real2str(con.sh,-1)+' frc='+real2str(con.frc,-1)+' frd='+real2str(con.frd,-1) );
//endif
end;


{*******************************************************************************}
{********* SImg_GetAlignment ***************************************************}
{*******************************************************************************}
const TSImg_Alignment_matchtmp_size = 8;

type TSImg_Alignment_matchval = record
        v      : word;
        i1,i2  : word; { index in source list }
        j1,j2  : byte; { index in chain }
        end;
     PSImg_Alignment_matchval = ^TSImg_Alignment_matchval;

     TSImg_Alignment_matchtmp = record
        len     : longint;
        ridx    : array[0..TSImg_Alignment_matchtmp_size-1] of word;  { index in reference source list }
        didx    : array[0..TSImg_Alignment_matchtmp_size-1] of word;  { index in destination source list }
        cnt     : array[0..TSImg_Alignment_matchtmp_size-1] of byte;  { number of hits }
        end;

     TSImg_Alignment_matches = record
        cnt             : longint;   { number of hits, or -1 if ambiguous (=false-match) }
        ref,dst         : longint;   { reference / destination index, destination index according to find2 }
        w,v,rx,ry,sx,sy : double;    { weight, proto-weight, destination and reference position }
        sdist           : double;    { square distance, calculated by corr and find2 }
{        tx,ty           : double; }   { for temporary calculations }
        end;
     PSImg_Alignment_matches = ^TSImg_Alignment_matches;

     TSImg_Alignment_find2buf = record  { closest reference per destination }
        ref      : longint;   { reference index or -1 if no known match}
        sdist    : double;    { square distance to reference }
        end;
     PSImg_Alignment_find2buf = ^TSImg_Alignment_find2buf;


{********* SImg_GetAlignment_checkConstraints **********************************}
{ res  : solution
  con  : constraints
  fast : no distortion
  cc   : can be null in order to disable debug output
  return true if constraint are met
  only uses res.a and res.b
}
function SImg_GetAlignment_checkConstraints(const res:TSImg_AlignmentParams; const con:TSImg_AlignmentConstraints; fast:boolean; cc:CSImg_Config):boolean;
const rad2n = 1/(2*pi);
      deg2n = 1/360.0;
var a,x1,y1,x2,y2,x,y,xx,yx,xy,yy       : double;
begin
    result:=false;
    if fast then
    begin
        { transform (con.rx,con.ry) }
        x1:=con.rx;
        y1:=con.ry;
        x:=res.b11*x1 + res.b12*y1 + res.ax;
        y:=res.b21*x1 + res.b22*y1 + res.ay;
        if sqr(x)+sqr(y)>sqr(con.dmax) then exit;
        { transform (con.rx+1,con.ry) }
        xx:=res.b11;
        yx:=res.b21;
        { transform (con.rx+1,con.ry+1) }
        xy:=res.b12;
        yy:=res.b22;
    end
    else begin
        { transform (con.rx,con.ry) }
        x1:=con.rx; x2:=sqr(x1);
        y1:=con.ry; y2:=sqr(y1);
        a:=res.d*(x2+y2) + res.c1*x1 + res.c2*y1;
        x1+=x1*a + res.c3*x2 + res.c5*y2;
        y1+=y1*a + res.c4*y2 + res.c6*x2;
        x:=res.b11*x1 + res.b12*y1 + res.ax;
        y:=res.b21*x1 + res.b22*y1 + res.ay;
        if sqr(x)+sqr(y)>sqr(con.dmax) then
        begin
//ifdef[align_debug]
            if cc<>nil then cc.info('SImg_GetAlignmentConstraints: dmax constraint violated');
//endif
            exit;
        end;
        { transform (con.rx+1,con.ry) }
        x1:=con.rx+1; x2:=sqr(x1);
        y1:=con.ry; y2:=sqr(y1);
        a:=res.d*(x2+y2) + res.c1*x1 + res.c2*y1;
        x1+=x1*a + res.c3*x2 + res.c5*y2;
        y1+=y1*a + res.c4*y2 + res.c6*x2;
        xx:=res.b11*x1 + res.b12*y1 + res.ax - x;
        yx:=res.b21*x1 + res.b22*y1 + res.ay - y;
        { transform (con.rx+1,con.ry+1) }
        x1:=con.rx; x2:=sqr(x1);
        y1:=con.ry+1; y2:=sqr(y1);
        a:=res.d*(x2+y2) + res.c1*x1 + res.c2*y1;
        x1+=x1*a + res.c3*x2 + res.c5*y2;
        y1+=y1*a + res.c4*y2 + res.c6*x2;
        xy:=res.b11*x1 + res.b12*y1 + res.ax - x;
        yy:=res.b21*x1 + res.b22*y1 + res.ay - y;
    end;
    { angles and scaling }
    x1:=sqrt(sqr(xx)+sqr(yx));
    if (x1<con.smin) or (x1>con.smax) then
    begin
//ifdef[align_debug]
        if cc<>nil then cc.info('SImg_GetAlignmentConstraints: smin/smax constraint violated (1)');
//endif
        exit;
    end;
    y1:=sqrt(sqr(xy)+sqr(yy));
    if (y1<con.smin) or (y1>con.smax) then
    begin
//ifdef[align_debug]
        if cc<>nil then cc.info('SImg_GetAlignmentConstraints: smin/smax constraint violated (1)');
//endif
        exit;
    end;
    x2:=arctan2(yx,xx)*rad2n;
    a:=arctan2(yy,xy)*rad2n-0.25-x2;
    a:=a-round(a);
    y2:=x2-con.frc*deg2n;
    y2-=round(y2);
    if con.flip then
    begin
        x2+=con.frc*deg2n;
        x2-=round(x2);
        if abs(a)>0.25 then
        begin
            a+=0.5;
            a-=round(a);
        end;
    end
    else x2:=y2;
//ifdef[align_debug]
    if cc<>nil then cc.info('SImg_GetAlignmentCheckConstraints: dx='+real2str(x,-1)+' dy='+real2str(y,-1)+'   rx='+real2str(x1,4)+' ry='+real2str(y1,4)+'    da1='+real2str(y2*360,-1)+' da2='+real2str(x2*360,-1)+'  sh='+real2str(a*360,-1));
//endif
    if (abs(x2)*360>con.frd) and (abs(y2)*360>con.frd) then exit;
    if abs(a)*360>con.sh then exit;
    result:=true;
end;


{********* SImg_GetAlignment_matchVals *****************************************}
{ calculate match values
    pl    : source list
    nm    : max number of match values per source
    depth : depth
    acc   : accuracy (0.1 = 10%)
    mv    : output match values
    flip  : allow flipping
    result: number of match values }
function SImg_GetAlignment_matchvals(pl:TSImg_RDImage; nm,depth:longint; acc:double; mv:PSImg_Alignment_matchval; flip:boolean):longint;
var i,j,k,l,m,nn,i1,i2 : longint;
    r,a,b              : double;
    r1,r2,x1,x2,y1,y2  : double;
    d                  : ^double;
begin
    r:=1/acc;
    if r>127 then r:=127;
    nn:=(pl.nx-5) div 3;
    if nn>256 then nn:=256;
    m:=0;
    for i:=0 to pl.ny-1 do
    begin
        l:=0;
        for k:=1 to depth do
        begin
            d:=pl.ptd + i*pl.sl + 5;
            for j:=0 to nn-k-1 do
            begin
                if l>=nm then break;
                x1:=d[0];
                y1:=d[1];
                i1:=round(d[2]);
                r1:=sqr(x1)+sqr(y1);
                if (r1<1e-3) then break;
                if (i1<0) or (i1>=pl.ny) then continue;
                x2:=d[k*3];
                y2:=d[k*3+1];
                i2:=round(d[k*3+2]);
                d+=3;
                r2:=sqr(x2)+sqr(y2);
                if r2<r1 then break;
                if (i2<0) or (i2>=pl.ny) or (r1<0.2*r2) then continue;
                r2:=1/r2;
                a:=(x1*x2+y1*y2)*r2;
                b:=(y1*x2-x1*y2)*r2;
                if flip then b:=abs(b);
//ifdef[align_debug]
                if abs(a)>1 then Simg_warning('internal error: |a| > 1');
                if abs(b)>1 then Simg_warning('internal error: |b| > 1');
//endif
                mv[l].v:=(round(a*r) shl 8) or (round(b*r) and 255);
                if mv[l].v<>0 then
                begin
                    mv[l].i1:=i1;
                    mv[l].i2:=i2;
                    mv[l].j1:=j;
                    mv[l].j2:=j+k;
                    l+=1;
                    m+=1;
                end;
            end;
        end;
        for l:=l to nm-1 do
            mv[l].v:=0;
        mv+=nm;
    end;
    result:=m;
end;

{********* SImg_GetAlignment_addMatch ******************************************}
procedure SImg_GetAlignment_addMatch(mr2:PSImg_Alignment_matches; idx:longint);
begin
    with mr2^ do
    begin
        if (cnt=0) or ((cnt>0) and (dst=idx)) then
        begin
            cnt+=1;
            dst:=idx;
        end
        else cnt:=-1;
    end;
end;

{********* SImg_GetAlignment_find **********************************************}
{ find initial matches
    pref  : reference source list
    pdst  : destination source list
    mvr   : reference match values
    mvd   : destination match values
    nref  : number reference match values per source
    ndst  : number reference match values per source
    ml    : match limit ( 1: 3 source match; 2: 4 source match; ...), must be smaller or equal than depth*2
    tt    : thread number
    tn    : number of threads
    mr2   : output match values
    con   : constraints
}
threadprocedure SImg_GetAlignment_find(pref:TSImg_RDImage; pdst:TSImg_RDImage; mvr:PSImg_Alignment_matchval; mvd:PSImg_Alignment_matchval; nref:longint; ndst:longint; ml:longint; tt:longint; tn:longint; mr2:PSImg_Alignment_matches; con:TSImg_AlignmentConstraints);
const rad2n = 1/(2*pi);
      deg2n = 1/360.0;
var i,j,k,l,m,n                 : longint;
    ci,ri,di                    : longint;
    ref,dst                     : PSImg_Alignment_matchval;
    mr                          : array[0..SImg_Sources_MaxChainLength-1] of TSImg_Alignment_matchtmp;
    b                           : boolean;
    rx0,ry0,rx1,ry1,rx2,ry2     : double;
    dx0,dy0,dx1,dy1,dx2,dy2     : double;
    r,rr0,ar0,dsrmin,dsrmax     : double;
    res                         : TSImg_AlignmentParams;
label nomatch;
begin
    with res do
    begin
        c1:=0;
        c2:=0;
        c3:=0;
        c4:=0;
        c5:=0;
        c6:=0;
        d:=0;
    end;
    i:=tt;
    while i<pref.ny do
    begin
        n:=i*pref.sl;
        rx0:=pref.ptd[n];
        ry0:=pref.ptd[n+1];
        rr0:=sqrt(sqr(rx0-con.rx)+sqr(ry0-con.ry));
        ar0:=arctan2(ry0-con.ry,rx0-con.rx)*rad2n + con.frc*deg2n;
        dsrmin:=rr0*con.smin-con.dmax;
        dsrmin*=abs(dsrmin);    // preserve negative sign
        dsrmax:=sqr(rr0*con.smax+con.dmax);
        for j:=0 to pdst.ny-1 do
        begin
            n:=j*pdst.sl;
            dx0:=pdst.ptd[n];
            dy0:=pdst.ptd[n+1];
            { pre-check constraints }
            dx1:=sqr(dx0)+sqr(dy0);
            if (dx1<dsrmin) or (dx1>dsrmax) then continue;
            if con.frd<179 then
            begin
                dx1:=arctan2(dy0,dx0)*rad2n;
                dx2:=dx1-ar0;
                dx2-=round(dx2);
                if con.flip then
                begin
                    dx1:=-dx1-ar0;
                    dx1-=round(dx1);
                end
                else  dx1:=dx2;
                if (abs(dx2)*360>con.frd) and (abs(dx1)*360>con.frd)  then continue;
            end;
            { find 3 source matches by comparing match vals }
            fillchar(mr,sizeof(mr),#0);
            ref:=mvr + i*nref;
            for k:=0 to nref-1 do
            begin
                if ref^.v=0 then break;
                dst:=mvd + j*ndst;
                for l:=0 to ndst-1 do
                begin
                    if dst^.v=0 then break;
                    if ref^.v=dst^.v then   { match }
                    begin
                        { check constraints }
                        n:=ref^.i1*pref.sl;
                        rx1:=pref.ptd[n  ] - rx0;
                        ry1:=pref.ptd[n+1] - ry0;

                        n:=ref^.i2*pref.sl;
                        rx2:=pref.ptd[n  ] - rx0;
                        ry2:=pref.ptd[n+1] - ry0;

                        n:=dst^.i1*pdst.sl;
                        dx1:=pdst.ptd[n  ] - dx0;
                        dy1:=pdst.ptd[n+1] - dy0;

                        n:=dst^.i2*pdst.sl;
                        dx2:=pdst.ptd[n  ] - dx0;
                        dy2:=pdst.ptd[n+1] - dy0;

                        r:=rx1*ry2-rx2*ry1;
                        if abs(r)<1e-50 then goto nomatch;
                        r:=1/r;
                        res.b11:=r*( ry2*dx1 -ry1*dx2);
                        res.b12:=r*(-rx2*dx1 +rx1*dx2);
                        res.b21:=r*( ry2*dy1 -ry1*dy2);
                        res.b22:=r*(-rx2*dy1 +rx1*dy2);
                        res.ax:=dx0 - res.b11*rx0 - res.b12*ry0;
                        res.ay:=dy0 - res.b21*rx0 - res.b22*ry0;
                        if not SImg_GetAlignment_checkConstraints(res,con,true,nil) then goto nomatch;
                        { store the matches }
                        ri:=ref^.i1;
                        di:=dst^.i1;
                        ci:=ref^.j1;
                        for m:=0 to 1 do
                        begin
                            with mr[ci] do
                            begin
                                n:=0;
                                while (n<len) and (didx[n]<>di) do
                                    n+=1;
                                if n<TSImg_Alignment_matchtmp_size then
                                begin
                                    cnt[n]+=1;
                                    ridx[n]:=ri;
                                    didx[n]:=di;
                                    if n=len then len:=n+1;
                                end;
                            end;
                            ci:=ref^.j2;
                            ri:=ref^.i2;
                            di:=dst^.i2;
                        end;
                    end;
nomatch:
                    dst+=1;
                end;
                ref+=1;
            end;

            { check for (ml+2)-source matches }
            b:=false;
            for k:=0 to SImg_sources_maxChainLength-1 do
            begin
                for l:=0 to TSImg_Alignment_matchtmp_size-1 do
                if mr[k].cnt[l]>=ml then
                begin
//                  writeln(i,',',j,':  ',mr[k].ridx[m],' = ',mr[k].didx[m],' (',mr[k].cnt[m],')');
                    b:=true;  { i,j are matches too }
                    SImg_GetAlignment_addMatch(mr2+mr[k].ridx[l], mr[k].didx[l]);
                end;
            end;
            if b then SImg_GetAlignment_addMatch(mr2+i, j);
        end;

    i+=tn;
    end;
end;

{********* SImg_GetAlignment_info **********************************************}
function SImg_GetAlignment_info(const res:TSImg_AlignmentParams; sep:string7):ansistring;
var a1,r1,a2,r2,s,rs  : double;
begin
    r1:=sqrt(sqr(res.b11)+sqr(res.b21));
    r2:=sqrt(sqr(res.b12)+sqr(res.b22));
    if abs(res.b11)>1e-50 then a1:=arctan2(res.b21,res.b11)*180/pi
    else if res.b11>0 then a1:=90
    else a1:=-90;

    if abs(res.b22)>1e-50 then a2:=arctan2(-res.b12,res.b22)*180/pi
    else if res.b22>0 then a2:=-90
    else a2:=90;

    if res.srad<1e-3 then s:=100
    else s:=res.srad;
    rs:=sqrt(s);
    result:='shift: ('+real2str(res.ax,-1)+', '+real2str(res.ay,-1)+')';
    result+=sep + 'rotation: (' +real2str(a1,-3)+'°, ' +real2str(a2,-3)+'°)   scale: ('+real2str(r1,-3)+', '+real2str(r2,-3)+')';
    result+=sep + '1st and 2nd order distortions at r='+real2str(rs,-1)+': ('+ real2str(res.c1*rs*100,-1)+'%, '+ real2str(res.c2*rs*100,-1)+'%, '+ real2str(res.c3*rs*100,-1)+'%, '+ real2str(res.c4*rs*100,-1)+'%, '+ real2str(res.c5*rs*100,-1)+'%, '+real2str(res.c6*rs*100,-1)+'%), ' + real2str(res.d*100*s,-2)+'%';
    result+=sep + 'correlation: ('+real2str(res.j1*1e-5,5)+', '+real2str(res.j2*1e-5,5)+', '+real2str(res.j3*1e-5,5)+')';
end;

{********* SImg_GetAlignment_solve *********************************************}
{ solve for alignment parameters
    matches  : matches
    ia,iz    : indexes
    dof      : degrees of freedom: 4, 6, 9, 13
    horeg    : regularization for polynomials 10..13. Ignored ifdef[align_nopc]
    res      : result }
procedure SImg_GetAlignment_solve(matches:PSImg_Alignment_matches; ia,iz:longint; dof:longint; var res:TSImg_AlignmentParams; horeg:double);
var i,j,jz,vn                   : longint;
    r                           : TSImg_AlignmentParams;
    v4                          : matrix;
// define[align_noopt]
//ifdef[align_noopt]
    p1x,p2x,p4x,p5x,p6x,p7x     : double;
    p1y,p2y,p4y,p5y,p6y,p7y     : double;
//endif
    x2,y2,xy,p3x,p3y            : double;
    a,x,y                       : double;
    pcr,pcs,spcs                : double;  { preconditioner for r[xy], s[xy] and sqr(s[xy]) }
    { sums }
    s_w                                         : double;
    s_sx,s_sy                                   : double;
    s_sxrx,s_sxry,s_syrx,s_syry                 : double;
    s_sp1,s_sp2,s_sp3,s_sp4                     : double;
    s_sp5,s_sp6,s_sp7                           : double;
    s_rx,s_ry                                   : double;
    s_rxx,s_rxy,s_ryy                           : double;
    s_p1x,s_p1y,s_p2x,s_p2y                     : double;
    s_p3x,s_p3y,s_p4x,s_p5y,s_p6x               : double;
    s_p7y                                       : double;
    s_p1xrx,s_p1xry,s_p1yrx,s_p1yry             : double;
    s_p2xrx,s_p2xry,s_p2yrx,s_p2yry             : double;
    s_p3xrx,s_p3xry,s_p3yrx,s_p3yry             : double;
    s_p4xrx,s_p4xry,s_p5yrx,s_p5yry             : double;
    s_p6xrx,s_p6xry,s_p7yrx,s_p7yry             : double;
    s_p11,s_p12,s_p13,s_p14,s_p15,s_p16,s_p17   : double;
    s_p22,s_p23,s_p24,s_p25,s_p26,s_p27         : double;
    s_p33,s_p34,s_p35,s_p36,s_p37               : double;
    s_p44,s_p46,s_p55,s_p57,s_p66,s_p77         : double;
//ifdef[align_noopt]
    s_p4y,s_p5x,s_p6y,s_p7x                     : double;
    s_p4yrx,s_p4yry,s_p5xrx,s_p5xry             : double;
    s_p6yrx,s_p6yry,s_p7xrx,s_p7xry             : double;
    s_p45,s_p47,s_p56,s_p67                     : double;
//endif
begin
    r:=res; v4.pt:=nil;

    vn:=0;
    for i:=ia to iz do
        with matches[i] do
            if cnt>0 then vn+=1;

    if vn<3 then i:=4
    else if vn<5 then i:=6
    else if vn<9 then i:=9
    else i:=13;
    if dof>i then dof:=i;

// define[align_nopc]
//ifdef[align_nopc]
    pcr:=1;
    spcs:=1;
    pcs:=1;
//else
    pcr:=4/(1+abs(res.xmin)+abs(res.xmax)+abs(res.ymin)+abs(res.ymax));
    spcs:=1;
    if res.srad>0.1 then spcs:=1/res.srad;
    pcs:=sqrt(spcs);
//endif

    jz:=0;
    for j:=0 to jz do
    begin
        { solve for shift, scaling, rotation; see notes }
        s_w:=0;
        s_sx:=0;  s_sy:=0;
        s_rx:=0;  s_ry:=0;
        s_sxrx:=0;  s_sxry:=0;  s_syrx:=0;  s_syry:=0;
        s_sp1:=0;  s_sp2:=0;  s_sp3:=0;  s_sp4:=0;
        s_sp5:=0;  s_sp6:=0;  s_sp7:=0;
        s_rxx:=0;  s_rxy:=0;  s_ryy:=0;
        s_p1x:=0;  s_p1y:=0;  s_p2x:=0;  s_p2y:=0;
        s_p3x:=0;  s_p3y:=0;  s_p4x:=0;
        s_p5y:=0;  s_p6x:=0;
        s_p7y:=0;
        s_p1xrx:=0;  s_p1xry:=0;  s_p1yrx:=0;  s_p1yry:=0;
        s_p2xrx:=0;  s_p2xry:=0;  s_p2yrx:=0;  s_p2yry:=0;
        s_p3xrx:=0;  s_p3xry:=0;  s_p3yrx:=0;  s_p3yry:=0;
        s_p4xrx:=0;  s_p4xry:=0;
        s_p5yrx:=0;  s_p5yry:=0;
        s_p6xrx:=0;  s_p6xry:=0;
        s_p7yrx:=0;  s_p7yry:=0;

        s_p11:=0;  s_p12:=0;  s_p13:=0;  s_p14:=0;  s_p15:=0;  s_p16:=0;  s_p17:=0;
        s_p22:=0;  s_p23:=0;  s_p24:=0;  s_p25:=0;  s_p26:=0;  s_p27:=0;
        s_p33:=0;  s_p34:=0;  s_p35:=0;  s_p36:=0;  s_p37:=0;
        s_p44:=0;  s_p46:=0;
        s_p57:=0;
//ifdef[align_nopc]
        s_p55:=0;
        s_p66:=0;
        s_p77:=0;
//else
        s_p44:=horeg;
        s_p55:=horeg;
        s_p66:=horeg;
        s_p77:=horeg;
//endif

//ifdef[align_noopt]
        s_p4y:=0;  s_p5x:=0;  s_p6y:=0;  s_p7x:=0;
        s_p4yrx:=0;  s_p4yry:=0;  s_p5xrx:=0;  s_p5xry:=0;
        s_p6yrx:=0;  s_p6yry:=0;  s_p7xrx:=0;  s_p7xry:=0;
        s_p45:=0;  s_p47:=0;  s_p56:=0;  s_p67:=0;
//endif

        for i:=ia to iz do
            with matches[i] do
                if cnt>0 then
                begin
                    { p1..p6 }
                    x:=sx*pcs;  y:=sy*pcs;
                    x2:=sqr(x); y2:=sqr(y);
                    xy:=y*x;
                    a:=x2+y2;
//ifdef[align_noopt]
                    p1x:=x*x;   p1y:=y*x;                { 1st order, c1 }
                    p2x:=x*y;   p2y:=y*y;                { 1st order, c2 }
                    p4x:=x*x;   p4y:=0;                  { 1st order, c3 }
                    p5x:=0;     p5y:=y*y;                { 1st order, c4 }
                    p6x:=y*y;   p6y:=0;                  { 1st order, c5 }
                    p7x:=0;     p7y:=x*x;                { 1st order, c6 }
//endif
                    p3x:=x*a;   p3y:=y*a;                { 2nd order, d }
                    { sums }
                    x:=rx*pcr;  y:=ry*pcr;
                    a:=w;
                    s_w+=a;
                    s_sx+=sx*a;
                    s_sy+=sy*a;
                    s_sxrx+=sx*x*a;
                    s_sxry+=sx*y*a;
                    s_syrx+=sy*x*a;
                    s_syry+=sy*y*a;
//ifdef[align_noopt]
                    s_sp1+=(sx*p1x+sy*p1y)*a;
                    s_sp2+=(sx*p2x+sy*p2y)*a;
                    s_sp3+=(sx*p3x+sy*p3y)*a;
                    s_sp4+=(sx*p4x+sy*p4y)*a;
                    s_sp5+=(sx*p5x+sy*p5y)*a;
                    s_sp6+=(sx*p6x+sy*p6y)*a;
                    s_sp7+=(sx*p7x+sy*p7y)*a;
                    s_rx+=x*a;
                    s_ry+=y*a;
                    s_rxx+=sqr(x)*a;
                    s_rxy+=x*y*a;
                    s_ryy+=sqr(y)*a;
                    s_p1x+=p1x*a;
                    s_p1y+=p1y*a;
                    s_p2x+=p2x*a;
                    s_p2y+=p2y*a;
                    s_p3x+=p3x*a;
                    s_p3y+=p3y*a;
                    s_p4x+=p4x*a;
                    s_p4y+=p4y*a;
                    s_p5x+=p5x*a;
                    s_p5y+=p5y*a;
                    s_p6x+=p6x*a;
                    s_p6y+=p6y*a;
                    s_p7x+=p7x*a;
                    s_p7y+=p7y*a;
                    s_p1xrx+=x*p1x*a;
                    s_p1xry+=y*p1x*a;
                    s_p1yrx+=x*p1y*a;
                    s_p1yry+=y*p1y*a;
                    s_p2xrx+=x*p2x*a;
                    s_p2xry+=y*p2x*a;
                    s_p2yrx+=x*p2y*a;
                    s_p2yry+=y*p2y*a;
                    s_p3xrx+=x*p3x*a;
                    s_p3xry+=y*p3x*a;
                    s_p3yrx+=x*p3y*a;
                    s_p3yry+=y*p3y*a;
                    s_p4xrx+=x*p4x*a;
                    s_p4xry+=y*p4x*a;
                    s_p4yrx+=x*p4y*a;
                    s_p4yry+=y*p4y*a;
                    s_p5xrx+=x*p5x*a;
                    s_p5xry+=y*p5x*a;
                    s_p5yrx+=x*p5y*a;
                    s_p5yry+=y*p5y*a;
                    s_p6xrx+=x*p6x*a;
                    s_p6xry+=y*p6x*a;
                    s_p6yrx+=x*p6y*a;
                    s_p6yry+=y*p6y*a;
                    s_p7xrx+=x*p7x*a;
                    s_p7xry+=y*p7x*a;
                    s_p7yrx+=x*p7y*a;
                    s_p7yry+=y*p7y*a;
                    s_p11+=(p1x*p1x+p1y*p1y)*a;
                    s_p12+=(p1x*p2x+p1y*p2y)*a;
                    s_p13+=(p1x*p3x+p1y*p3y)*a;
                    s_p14+=(p1x*p4x+p1y*p4y)*a;
                    s_p15+=(p1x*p5x+p1y*p5y)*a;
                    s_p16+=(p1x*p6x+p1y*p6y)*a;
                    s_p17+=(p1x*p7x+p1y*p7y)*a;
                    s_p22+=(p2x*p2x+p2y*p2y)*a;
                    s_p23+=(p2x*p3x+p2y*p3y)*a;
                    s_p24+=(p2x*p4x+p2y*p4y)*a;
                    s_p25+=(p2x*p5x+p2y*p5y)*a;
                    s_p26+=(p2x*p6x+p2y*p6y)*a;
                    s_p27+=(p2x*p7x+p2y*p7y)*a;
                    s_p33+=(p3x*p3x+p3y*p3y)*a;
                    s_p34+=(p3x*p4x+p3y*p4y)*a;
                    s_p35+=(p3x*p5x+p3y*p5y)*a;
                    s_p36+=(p3x*p6x+p3y*p6y)*a;
                    s_p37+=(p3x*p7x+p3y*p7y)*a;
                    s_p44+=(p4x*p4x+p4y*p4y)*a;
                    s_p45+=(p4x*p5x+p4y*p5y)*a;
                    s_p46+=(p4x*p6x+p4y*p6y)*a;
                    s_p47+=(p4x*p7x+p4y*p7y)*a;
                    s_p55+=(p5x*p5x+p5y*p5y)*a;
                    s_p56+=(p5x*p6x+p5y*p6y)*a;
                    s_p57+=(p5x*p7x+p5y*p7y)*a;
                    s_p66+=(p6x*p6x+p6y*p6y)*a;
                    s_p67+=(p6x*p7x+p6y*p7y)*a;
                    s_p77+=(p7x*p7x+p7y*p7y)*a;

//else
                    s_sp1+=(sx*x2+sy*xy)*a;
                    s_sp2+=(sx*xy+sy*y2)*a;
                    s_sp3+=(sx*p3x+sy*p3y)*a;
                    s_sp4+=sx*x2*a;
                    s_sp5+=sy*y2*a;
                    s_sp6+=sx*y2*a;
                    s_sp7+=sy*x2*a;
                    s_rx+=x*a;
                    s_ry+=y*a;
                    s_rxx+=sqr(x)*a;
                    s_rxy+=x*y*a;
                    s_ryy+=sqr(y)*a;
                    s_p1x+=x2*a;
                    s_p1y+=xy*a;
                    s_p2x+=xy*a;
                    s_p2y+=y2*a;
                    s_p3x+=p3x*a;
                    s_p3y+=p3y*a;
                    s_p4x+=x2*a;
                    s_p5y+=y2*a;
                    s_p6x+=y2*a;
                    s_p7y+=x2*a;
                    s_p1xrx+=x*x2*a;
                    s_p1xry+=y*x2*a;
                    s_p1yrx+=x*xy*a;
                    s_p1yry+=y*xy*a;
                    s_p2xrx+=x*xy*a;
                    s_p2xry+=y*xy*a;
                    s_p2yrx+=x*y2*a;
                    s_p2yry+=y*y2*a;
                    s_p3xrx+=x*p3x*a;
                    s_p3xry+=y*p3x*a;
                    s_p3yrx+=x*p3y*a;
                    s_p3yry+=y*p3y*a;
                    s_p4xrx+=x*x2*a;
                    s_p4xry+=y*x2*a;
                    s_p5yrx+=x*y2*a;
                    s_p5yry+=y*y2*a;
                    s_p6xrx+=x*y2*a;
                    s_p6xry+=y*y2*a;
                    s_p7yrx+=x*x2*a;
                    s_p7yry+=y*x2*a;
                    s_p11+=(x2*x2+xy*xy)*a;
                    s_p12+=(x2*xy+xy*y2)*a;
                    s_p13+=(x2*p3x+xy*p3y)*a;
                    s_p14+=x2*x2*a;
                    s_p15+=xy*y2*a;
                    s_p16+=x2*y2*a;
                    s_p17+=xy*x2*a;
                    s_p22+=(xy*xy+y2*y2)*a;
                    s_p23+=(xy*p3x+y2*p3y)*a;
                    s_p24+=xy*x2*a;
                    s_p25+=y2*y2*a;
                    s_p26+=xy*y2*a;
                    s_p27+=y2*x2*a;
                    s_p33+=(p3x*p3x+p3y*p3y)*a;
                    s_p34+=p3x*x2*a;
                    s_p35+=p3y*y2*a;
                    s_p36+=p3x*y2*a;
                    s_p37+=p3y*x2*a;
                    s_p44+=x2*x2*a;
                    s_p46+=x2*y2*a;
                    s_p55+=y2*y2*a;
                    s_p57+=y2*x2*a;
                    s_p66+=y2*y2*a;
                    s_p77+=x2*x2*a;
//endif
                end;

        if dof<6
        then am(v4,solve( rmatrix(4,4, [ s_w,    0,         s_rx,         -s_ry,
                                           0,  s_w,         s_ry,          s_rx,
                                        s_rx, s_ry,  s_rxx+s_ryy,             0,
                                       -s_ry, s_rx,            0,   s_rxx+s_ryy ]),
                         rmatrix(1,4, [ s_sx, s_sy, s_sxrx+s_syry, s_syrx-s_sxry ])) )
        else if dof<9
        then am(v4,solve( rmatrix(6,6, [  s_w,     0,    s_rx,    s_ry,       0,       0,
                                            0,   s_w,       0,       0,    s_rx,    s_ry,
                                         s_rx,     0,   s_rxx,   s_rxy,       0,       0,
                                         s_ry,     0,   s_rxy,   s_ryy,       0,       0,
                                            0,  s_rx,       0,       0,   s_rxx,   s_rxy,
                                            0,  s_ry,       0,       0,   s_rxy,   s_ryy ] ),
                          rmatrix(1,6, [ s_sx, s_sy, s_sxrx, s_sxry, s_syrx, s_syry ])) )
        else if dof<13
        then am(v4,solve( rmatrix(9,9,    [  s_w,     0,    s_rx,    s_ry,       0,       0,   s_p1x,   s_p2x,   s_p3x,
                                               0,   s_w,       0,       0,    s_rx,    s_ry,   s_p1y,   s_p2y,   s_p3y,
                                            s_rx,     0,   s_rxx,   s_rxy,       0,       0, s_p1xrx, s_p2xrx, s_p3xrx,
                                            s_ry,     0,   s_rxy,   s_ryy,       0,       0, s_p1xry, s_p2xry, s_p3xry,
                                               0,  s_rx,       0,       0,   s_rxx,   s_rxy, s_p1yrx, s_p2yrx, s_p3yrx,
                                               0,  s_ry,       0,       0,   s_rxy,   s_ryy, s_p1yry, s_p2yry, s_p3yry,
                                           s_p1x, s_p1y, s_p1xrx, s_p1xry, s_p1yrx, s_p1yry,   s_p11,   s_p12,   s_p13,
                                           s_p2x, s_p2y, s_p2xrx, s_p2xry, s_p2yrx, s_p2yry,   s_p12,   s_p22,   s_p23,
                                           s_p3x, s_p3y, s_p3xrx, s_p3xry, s_p3yrx, s_p3yry,   s_p13,   s_p23,   s_p33 ]),
                             rmatrix(1,9, [ s_sx, s_sy, s_sxrx, s_sxry, s_syrx, s_syry, s_sp1, s_sp2, s_sp3 ])) )
//ifdef[align_noopt]
        else am(v4,solve( rmatrix(13,13,  [  s_w,     0,    s_rx,    s_ry,       0,       0,   s_p1x,   s_p2x,   s_p3x,   s_p4x,   s_p5x,   s_p6x,   s_p7x,
                                               0,   s_w,       0,       0,    s_rx,    s_ry,   s_p1y,   s_p2y,   s_p3y,   s_p4y,   s_p5y,   s_p6y,   s_p7y,
                                            s_rx,     0,   s_rxx,   s_rxy,       0,       0, s_p1xrx, s_p2xrx, s_p3xrx, s_p4xrx, s_p5xrx, s_p6xrx, s_p7xrx,
                                            s_ry,     0,   s_rxy,   s_ryy,       0,       0, s_p1xry, s_p2xry, s_p3xry, s_p4xry, s_p5xry, s_p6xry, s_p7xry,
                                               0,  s_rx,       0,       0,   s_rxx,   s_rxy, s_p1yrx, s_p2yrx, s_p3yrx, s_p4yrx, s_p5yrx, s_p6yrx, s_p7yrx,
                                               0,  s_ry,       0,       0,   s_rxy,   s_ryy, s_p1yry, s_p2yry, s_p3yry, s_p4yry, s_p5yry, s_p6yry, s_p7yry,
                                           s_p1x, s_p1y, s_p1xrx, s_p1xry, s_p1yrx, s_p1yry,   s_p11,   s_p12,   s_p13,   s_p14,   s_p15,   s_p16,   s_p17,
                                           s_p2x, s_p2y, s_p2xrx, s_p2xry, s_p2yrx, s_p2yry,   s_p12,   s_p22,   s_p23,   s_p24,   s_p25,   s_p26,   s_p27,
                                           s_p3x, s_p3y, s_p3xrx, s_p3xry, s_p3yrx, s_p3yry,   s_p13,   s_p23,   s_p33,   s_p34,   s_p35,   s_p36,   s_p37,
                                           s_p4x, s_p4y, s_p4xrx, s_p4xry, s_p4yrx, s_p4yry,   s_p14,   s_p24,   s_p34,   s_p44,   s_p45,   s_p46,   s_p47,
                                           s_p5x, s_p5y, s_p5xrx, s_p5xry, s_p5yrx, s_p5yry,   s_p15,   s_p25,   s_p35,   s_p45,   s_p55,   s_p56,   s_p57,
                                           s_p6x, s_p6y, s_p6xrx, s_p6xry, s_p6yrx, s_p6yry,   s_p16,   s_p26,   s_p36,   s_p46,   s_p56,   s_p66,   s_p67,
                                           s_p7x, s_p7y, s_p7xrx, s_p7xry, s_p7yrx, s_p7yry,   s_p17,   s_p27,   s_p37,   s_p47,   s_p57,   s_p67,   s_p77 ]),
                            rmatrix(1,13, [ s_sx, s_sy, s_sxrx, s_sxry, s_syrx, s_syry, s_sp1, s_sp2, s_sp3, s_sp4, s_sp5, s_sp6, s_sp7])) );
//else
        else am(v4,solve( rmatrix(13,13,  [  s_w,     0,    s_rx,    s_ry,       0,       0,   s_p1x,   s_p2x,   s_p3x,   s_p4x,       0,   s_p6x,       0,
                                               0,   s_w,       0,       0,    s_rx,    s_ry,   s_p1y,   s_p2y,   s_p3y,       0,   s_p5y,       0,   s_p7y,
                                            s_rx,     0,   s_rxx,   s_rxy,       0,       0, s_p1xrx, s_p2xrx, s_p3xrx, s_p4xrx,       0, s_p6xrx,       0,
                                            s_ry,     0,   s_rxy,   s_ryy,       0,       0, s_p1xry, s_p2xry, s_p3xry, s_p4xry,       0, s_p6xry,       0,
                                               0,  s_rx,       0,       0,   s_rxx,   s_rxy, s_p1yrx, s_p2yrx, s_p3yrx,       0, s_p5yrx,       0, s_p7yrx,
                                               0,  s_ry,       0,       0,   s_rxy,   s_ryy, s_p1yry, s_p2yry, s_p3yry,       0, s_p5yry,       0, s_p7yry,
                                           s_p1x, s_p1y, s_p1xrx, s_p1xry, s_p1yrx, s_p1yry,   s_p11,   s_p12,   s_p13,   s_p14,   s_p15,   s_p16,   s_p17,
                                           s_p2x, s_p2y, s_p2xrx, s_p2xry, s_p2yrx, s_p2yry,   s_p12,   s_p22,   s_p23,   s_p24,   s_p25,   s_p26,   s_p27,
                                           s_p3x, s_p3y, s_p3xrx, s_p3xry, s_p3yrx, s_p3yry,   s_p13,   s_p23,   s_p33,   s_p34,   s_p35,   s_p36,   s_p37,
                                           s_p4x,     0, s_p4xrx, s_p4xry,       0,       0,   s_p14,   s_p24,   s_p34,   s_p44,       0,   s_p46,       0,
                                               0, s_p5y,       0,       0, s_p5yrx, s_p5yry,   s_p15,   s_p25,   s_p35,       0,   s_p55,       0,   s_p57,
                                           s_p6x,     0, s_p6xrx, s_p6xry,       0,       0,   s_p16,   s_p26,   s_p36,   s_p46,       0,   s_p66,       0,
                                               0, s_p7y,       0,       0, s_p7yrx, s_p7yry,   s_p17,   s_p27,   s_p37,       0,   s_p57,       0,   s_p77 ]),
                            rmatrix(1,13, [ s_sx, s_sy, s_sxrx, s_sxry, s_syrx, s_syry, s_sp1, s_sp2, s_sp3, s_sp4, s_sp5, s_sp6, s_sp7])) );
//endif

        r.b22:=v4.pt[2]*pcr;
        r.b12:=-v4.pt[3]*pcr;
        if v4.y>4 then r.b21:=-v4.pt[4]*pcr else begin r.b21:=r.b12; r.b12:=-r.b12; end;
        if v4.y>5 then r.b11:= v4.pt[5]*pcr else r.b11:=r.b22;
        a:=r.b11*r.b22-r.b12*r.b21;
        if abs(a)<1e-100 then a:=0
        else a:=1/a;
        r.b11*=a;  r.b12*=a;
        r.b21*=a;  r.b22*=a;
        r.ax:=-r.b11*v4.pt[0]-r.b12*v4.pt[1];
        r.ay:=-r.b21*v4.pt[0]-r.b22*v4.pt[1];
        if v4.y>6  then r.c1:=-v4.pt[6]*spcs     else r.c1:=0;
        if v4.y>7  then r.c2:=-v4.pt[7]*spcs     else r.c2:=0;
        if v4.y>8  then r.d :=-v4.pt[8]*spcs*pcs else r.d :=0;
        if v4.y>9  then r.c3:=-v4.pt[9]*spcs     else r.c3:=0;
        if v4.y>10 then r.c4:=-v4.pt[10]*spcs    else r.c4:=0;
        if v4.y>11 then r.c5:=-v4.pt[11]*spcs    else r.c5:=0;
        if v4.y>12 then r.c6:=-v4.pt[12]*spcs    else r.c6:=0;

        free(v4);
    end;
    res:=r;
end;

{********* SImg_GetAlignment_corr **********************************************}
{ calculate correlation
    matches  : matches
    ny       : size of matches
    res      : align parameters (solution)
    rad      : radius
    ddz      : re-calculate weights
}
function SImg_GetAlignment_corr(matches:PSImg_Alignment_matches; ny:longint; var res:TSImg_AlignmentParams; rad:double; ddz:boolean):double;
var i            : longint;
    r            : TSImg_AlignmentParams;
    x1,y1,x2,y2  : double;
    a,b,r2       : double;
begin
    r:=res;
    r.j1:=0;  r.j2:=0;  r.j3:=0;
    r2:=ln(0.5)/sqr(rad);

    for i:=0 to ny-1 do
        if matches[i].cnt>0 then
        begin
            try
                x1:=matches[i].sx; x2:=sqr(x1);
                y1:=matches[i].sy; y2:=sqr(y1);
                a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
                x1+=x1*a + r.c3*x2 + r.c5*y2;
                y1+=y1*a + r.c4*y2 + r.c6*x2;
                x2:=r.b11*x1 + r.b12*y1 + r.ax;
                y2:=r.b21*x1 + r.b22*y1 + r.ay;
                a:=sqr(matches[i].rx-x2)+sqr(matches[i].ry-y2);
                matches[i].sdist:=a;
                a:=exp(r2*a);
            except
                a:=0;
                matches[i].sdist:=1e100;
            end;
            b:=matches[i].v;
            if ddz then matches[i].w:=b*a;

            r.j1+=b*a;
            a:=sqr(a);
            r.j2+=b*a;
            r.j3+=b*sqr(a);
        end;

    if ddz then
    begin
        res.j1:=r.j1;
        res.j2:=r.j2;
        res.j3:=r.j3;
    end;
    result:=r.j1 + r.j2 + r.j3;
end;

{********* SImg_GetAlignment_J *************************************************}
{ calculate minimization functional
    matches  : matches
    ny       : size of matches
    res      : align parameters (solution)
}
function SImg_GetAlignment_J(matches:PSImg_Alignment_matches; ny:longint; r:TSImg_AlignmentParams):double;
var i                   : longint;
    x1,y1,x2,y2         : double;
    j,a,ax,ay           : double;
    b11,b12,b21,b22     : double;
begin
    a:=1/(r.b11*r.b22-r.b12*r.b21);
    b11:= r.b22*a;
    b12:=-r.b12*a;
    b21:=-r.b21*a;
    b22:= r.b11*a;
    ax:=b11*r.ax+b12*r.ay;
    ay:=b21*r.ax+b22*r.ay;
    j:=0;
    for i:=0 to ny-1 do
        if matches[i].cnt>0 then
        begin
            x1:=matches[i].sx; x2:=sqr(x1);
            y1:=matches[i].sy; y2:=sqr(y1);
            a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
            x1+=x1*a + r.c3*x2 + r.c5*y2 + ax;
            y1+=y1*a + r.c4*y2 + r.c6*x2 + ay;
            x2:=matches[i].rx;
            y2:=matches[i].ry;
            j+=(sqr(b11*x2+b12*y2-x1)+sqr(b21*x2+b22*y2-y1))*matches[i].w;
        end;
    result:=j;
end;


{********* SImg_GetAlignment_factor ********************************************}
{ calculate factor
    matches  : matches
    ny       : size of matches
    res      : result
    result   : number of matches
}
function SImg_GetAlignment_factor(pref,pdst:TSImg_RDImage; matches:PSImg_Alignment_matches; ny:longint; var res:double):longint;
var i,ri,di             : longint;
    sdm,rb,db,rs,ds,w   : double;
begin
    result:=0;
    sdm:=sqr(0.5*SImg_sources_mindist);
    rs:=1e-100;
    ds:=1e-100;
    for i:=0 to ny-1 do
        if (matches[i].cnt>0) and (matches[i].sdist<sdm) then
        begin
            ri:=matches[i].ref;
            di:=matches[i].dst;
            db:=pdst.ptd[pdst.sl*di+2];
            rb:=pref.ptd[pref.sl*ri+2];
//          writeln(stderr, i:3,': db=',db:0:3,'  rb=',rb:0:3,'  dw=',pdst.ptd[pdst.sl*di+4]:0:3,'  rw=',prefy.ptd[pref.sl*ri+4]:0:3,'  dbg=',pdst.ptd[pdst.sl*di+3]:0:3,'  rng=',pref.ptd[pref.sl*ri+3]:0:3);
            if (db>0) and (rb>0) then
            begin
                w:=sqrt(sqrt(rb*db));
                w*=sqrt(w);  { = (rb*db)^0.375 }
                w:=pdst.ptd[pdst.sl*di+4]*pref.ptd[pref.sl*ri+4]/w;
                rs+=rb*w;
                ds+=db*w;
                result+=1;
            end;
        end;
    res:=rs/ds;
end;


{********* SImg_GetAlignment_find2 *********************************************}
{ Find matches based on a solution
    pref     : reference sources
    pdst     : destination sources
    rad      : radius
    res      : align parameters (solution)
    matches  : where to store matches, can be nil, will be ordered (index=ref)
    dbuf     : buffer to determine closest reference per destination in order to find most likely match, can be nil
    ia,iz    : thread index und number of threads
}
threadprocedure SImg_GetAlignment_find2_thread(pref:TSImg_RDImage; pdst:TSImg_RDImage; rad:double; res:PSImg_AlignmentParams; matches:PSImg_Alignment_matches; dbuf:PSImg_Alignment_find2buf; ia:longint; iz:longint);
var i,j,k        : longint;
    r            : TSImg_AlignmentParams;
    x1,y1,x2,y2  : double;
    a,b,r2,srmin : double;
    d            : ^double;
begin
    r:=res^;
    r2:=ln(0.5)/sqr(rad);
    srmin:=sqr(rad*4);

    for i:=ia to iz do
    begin
        d:=pref.ptd + i*pref.sl;
        x1:=d[0]; x2:=sqr(x1);
        y1:=d[1]; y2:=sqr(y1);
        a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
        x1+=x1*a + r.c3*x2 + r.c5*y2;
        y1+=y1*a + r.c4*y2 + r.c6*x2;
        x2:=r.b11*x1 + r.b12*y1 + r.ax;
        y2:=r.b21*x1 + r.b22*y1 + r.ay;

//      if (x2<r.xmin) or (x2>r.xmax) or (y2<r.ymin) or (y2>r.ymax) then continue;

        k:=0;
        d:=pdst.ptd;
        a:=1e100;
        for j:=0 to pdst.ny-1 do
        begin
            b:=sqr(d[0]-x2)+sqr(d[1]-y2);
            if b<a then
            begin
                k:=j;
                a:=b;
            end;
            d+=pdst.sl;
        end;

        if (dbuf<>nil) and (a<dbuf[k].sdist) then
        begin
            dbuf[k].sdist:=a;
            dbuf[k].ref:=i;
        end;

        if a<srmin then
        begin
            b:=pref.ptd[i*pref.sl+2]*pdst.ptd[k*pdst.sl+2];
            if b<0 then b:=0
            else b:=sqrt(sqrt(b));
            with matches[i] do
            begin
                cnt:=1;
                dst:=k;
                sdist:=a;               { square distance }
                v:=b;                   { <proto weight> = <brightness>^0.25 }
                w:=exp(r2*a);           { not the weight, weight will be calculated later as b*(a+a^2+a^4) }
                d:=pref.ptd+pref.sl*i;
                sx:=d[0];  sy:=d[1];    { reference position }
                d:=pdst.ptd+pdst.sl*k;
                rx:=d[0];  ry:=d[1];    { destination position }
            end;
        end;

    end;
end;


{ Find matches based on a solution
    pref     : reference sources
    pdst     : destination sources
    rad      : radius
    res      : align parameters (solution)
    matches  : where to store new matches
    dbuf     : buffer to determine closest reference per destination in order to find most likely match, can bu nil
}
procedure SImg_GetAlignment_find2(pref,pdst:TSImg_RDImage; rad:double; var res:TSImg_AlignmentParams; matches:PSImg_Alignment_matches; dbuf:PSImg_Alignment_find2buf);
var cc          : CSImg_Config;
    i,ia,iz,tn  : longint;
    c           : double;
threadvars;
begin
    cc:=SImg_Config_Default;
    res.j1:=0;
    res.j2:=0;
    res.j3:=0;

    for i:=0 to pref.ny-1 do
    with matches[i] do
    begin
        cnt:=0;
        ref:=i;
    end;

    if dbuf<>nil then
        for i:=0 to pdst.ny-1 do
            with dbuf[i] do
            begin
                ref:=-1;
                sdist:=1e300;
            end;

    tn:=cc.threads;
    i:=(pdst.ny*pref.ny) div 1000000;
    if i<tn then tn:=i;

    if tn<2 then SImg_GetAlignment_find2_thread(pref,pdst, rad, @res, matches,dbuf, 0,pref.ny-1);
    begin
        ia:=0;
        for i:=0 to tn-1 do
        begin
            iz:=(i+1)*pref.ny div tn;
            dothread SImg_GetAlignment_find2_thread(pref,pdst, rad, @res, matches,dbuf, ia,iz-1);
            ia:=iz;
        end;
        threadwait;
    end;

    { sort out the most likely matches and calculate correlation and weight }
    if dbuf<>nil then
    for i:=0 to pref.ny-1 do
        if (matches[i].cnt>0) and (dbuf[matches[i].dst].ref<>i) then matches[i].cnt:=0;
    for i:=0 to pref.ny-1 do
        with matches[i] do
            if cnt>0 then
            begin
                c:=sqr(w);
                res.j1+=v*w;
                res.j2+=v*c;
                res.j3+=v*sqr(c);
                w:=v*(w+c+sqr(c));
            end;

    {a:=(res.b11*res.b22-res.b12*res.b21);
    if a<0 then a:=0;
    a:=sqrt(a);
    res.j1*=a; res.j2*=a; res.j3*=a;}
end;

{********* SImg_GetAlignment_find3 *********************************************}
{ Find matches based on a solution
    pref     : reference sources
    pdst     : destination sources
    rad      : radius
    res      : align parameters (solution)
    omatches : matches, must be ordered (index=ref),
    returns  : number of matches equal to matches in omatches
}

function SImg_GetAlignment_find3(pref,pdst:TSImg_RDImage; r:TSImg_AlignmentParams; omatches:PSImg_Alignment_matches):longint;
var i,j,k,j1     : longint;
    x1,y1,x2,y2  : double;
    a,b          : double;
    d            : ^double;
begin
    j1:=0;
//    writeln(stderr,'xmin=',r.xmin:0:1,'  xmin=',r.xmax:0:1,'  ymin=',r.ymin:0:1,'  xmax=',r.ymax:0:1);
    for i:=0 to pref.ny-1 do
    begin
        d:=pref.ptd + i*pref.sl;
        x1:=d[0]; x2:=sqr(x1);
        y1:=d[1]; y2:=sqr(y1);
        a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
        x1+=x1*a + r.c3*x2 + r.c5*y2;
        y1+=y1*a + r.c4*y2 + r.c6*x2;
        x2:=r.b11*x1 + r.b12*y1 + r.ax;
        y2:=r.b21*x1 + r.b22*y1 + r.ay;

        if (x2<r.xmin) or (x2>r.xmax) or (y2<r.ymin) or (y2>r.ymax) then continue;

        k:=0;
        d:=pdst.ptd;
        a:=1e100;
        for j:=0 to pdst.ny-1 do
        begin
            b:=sqr(d[0]-x2)+sqr(d[1]-y2);
            if b<a then
            begin
                k:=j;
                a:=b;
            end;
            d+=pdst.sl;
        end;

        if (omatches[i].cnt>0) and (omatches[i].dst=k) then j1+=1;
    end;
    result:=j1;
end;

{********* SImg_GetAlignment_solveinit *****************************************}
{ find best initial solution by splitting inital match list into fragments which are solved independently.
  Reason: Due to quadratic error a single wrong match makes the result unusable.
    pref     : reference sources
    pdst     : destination
    omatches : matches, must be ordered (index=ref)
    smatches : matches, must be sorted by cnt
    it       : iteration (0: dof=4..9; 1: dof=9, 2: dof=13)
    tt,tn    : thread index and number of threads
    res      : best result }
threadprocedure SImg_GetAlignment_solveinit(pref:TSImg_RDImage; pdst:TSImg_RDImage; omatches:PSImg_Alignment_matches; smatches:PSImg_Alignment_matches; it:longint; tt:longint; tn:longint; res:PSImg_AlignmentParams; con:PSImg_AlignmentConstraints);
const desired_fragments1 = 32;   { desired number of fragments for first iteration, total match count per fragment is approximately constant }
      desired_fragments2 = 16;   { desired number of fragments for other iterations }
var i,j,k,l,m,n : longint;
    r           : TSImg_AlignmentParams;
    dof         : longint;   { degrees of freedom }
    df          : longint;   { desired fragments }
    mm          : longint;   { minimum matches }
    cr          : longint;   { remaining counts }
    fc          : longint;   { fragment counter }
    mr          : longint;   { total match count }
    solve       : boolean;

begin
    res^.j1:=-1;
    r:=res^;

    if it>1 then
    begin
        mm:=10;
        dof:=13;
    end
    else if it>0 then
    begin
        mm:=5;
        dof:=9;
    end
    else begin
        mm:=2;
        dof:=4;
        if con^.nq then mm+=1;
        if con^.flip then mm+=1;
        if mm>2 then dof:=6;
    end;

    if it>0 then df:=desired_fragments2
    else df:=desired_fragments1;

    mr:=0; cr:=0;
    while (mr<pref.ny) and (smatches[mr].cnt>0) do
    begin
        cr+=smatches[mr].cnt;
        mr+=1;
    end;

    i:=0;  fc:=0;
    while i<mr do
    begin
        solve:=(fc mod tn)=tt;

        l:=cr;
        if fc<df then l:=l div (df-fc);

        { find fragment }
        j:=i; k:=0;
        while (j<mr) and ((j-i<mm) or (k<l) or (mr-j<mm)) do
            with smatches[j] do
            begin
                k+=cnt;
                if solve then
                begin
                    m:=ref*pref.sl;
                    sx:=pref.ptd[m];
                    sy:=pref.ptd[m+1];
                    n:=dst*pdst.sl;
                    rx:=pdst.ptd[n];
                    ry:=pdst.ptd[n+1];
//                  if (i=0) then writeln(i,'->',j,':  ',ref,'   ',dst,'  (',rx:0:2,', ',ry:0:2,')  <-  (',sx:0:2,', ',sy:0:2,')');
                    w:=cnt;
                end;
                j+=1;
            end;

        { solve fragment }
        if solve and (j-i>=mm) then
        begin
//          writeln(stderr,'solveinit(',it,'): ',i,'  ',j,'   ',mr);
            SImg_GetAlignment_solve(smatches,i,j-1, dof, r, 1);
            if SImg_GetAlignment_checkConstraints(r,con^,false,nil) then
            begin
                r.j1:=SImg_GetAlignment_find3(pref,pdst, r, omatches);
//              SImg_Info(int2str(it)+':'+int2str(i)+'..'+int2str(j-1)+':  '+SImg_GetAlignment_info(r,';   '));
                if r.j1>res^.j1 then res^:=r;
            end;
        end;
        i:=j; cr-=k; fc+=1;
    end;
end;

{********* SImg_GetAlignment_sort **********************************************}
{ sort matches by cnt
  result   : number of matches
  omatches : ordered matched (ref=index, input)
  smatches : sorted matches (output)
  ny       : refs.ny
}
function SImg_GetAlignment_sort(ny:longint; omatches,smatches:PSImg_Alignment_matches):longint;
var i,j,k,l,m,n : longint;
begin
    { sort matches }
    m:=0;
    for i:=0 to ny-1 do
        if omatches[i].cnt>0 then
        begin
            j:=m;
            k:=omatches[i].cnt;
            l:=omatches[i].dst;
            n:=omatches[i].ref;
            smatches[j].cnt:=k;
            smatches[j].ref:=n;
            smatches[j].dst:=l;
            while (j>0) and (smatches[j-1].cnt<k) do
            begin
                smatches[j].cnt:=smatches[j-1].cnt;
                smatches[j].ref:=smatches[j-1].ref;
                smatches[j].dst:=smatches[j-1].dst;
                smatches[j-1].cnt:=k;
                smatches[j-1].ref:=n;
                smatches[j-1].dst:=l;
                j-=1;
            end;
            m+=1;
        end;
    for i:=m to ny-1 do
        smatches[i].cnt:=0;
    result:=m;
end;

{********* SImg_GetAlignment ***************************************************}
{ calculates how reference must be aligned to fit to destination
    pref0 : reference sources
    pdst0 : destination
    dl    : estimate distortion level
    res   : result
    con   : constraints

    dl settings:

    dl  depth   acc  ml   sh
    0       1    2%   2   1
    1       1    4%   2   2
    2       2    8%   2   4
    3       3   15%   2   8
}
procedure SImg_GetAlignment(cc:CSImg_Config; pref0,pdst0:TSImg_RDImage; dl:longint; var res:TSImg_AlignmentParams; con:TSImg_AlignmentConstraints);
const density_factor = 2;  { max. source  density ratio }
var pref,pdst                   : TSImg_RDImage;
    i,j,k,m                     : longint;
    ndst,nref,tn                : longint;
    depth,ml                    : longint;
    mvr,mvd                     : ^TSImg_Alignment_matchval;
    omatches,smatches           : ^TSImg_Alignment_matches;
    dbuf                        : ^TSImg_Alignment_find2buf;
    a,r                         : double;
{   b,s                         : double;}
    j1,j2,j3                    : double;
    xmin,ymin,xmax,ymax         : double;
    res2                        : ^TSImg_AlignmentParams;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetAlignment: pref='+SImg_InfoStr2(pref0)+' pdst='+SImg_InfoStr2(pdst0)+'  dl='+int2str(dl));
    if not (SImg_Recover(cc,pref0) and (pref0.fmt=SImg_RawDouble) and (pref0.nx>=5) and (pref0.nc=1)) then cc.Error('SImg_GetAlignment: Invalid format of reference source data array');
    if not (SImg_Recover(cc,pdst0) and (pdst0.fmt=SImg_RawDouble) and (pdst0.nx>=5) and (pdst0.nc=1)) then cc.Error('SImg_GetAlignment: Invalid format of destination source data array');

    { check boundaries of destination source list }
    xmin:=res.xmin;
    ymin:=res.ymin;
    xmax:=res.xmax;
    ymax:=res.ymax;
    for i:=0 to pdst0.ny-1 do
    begin
        j1:=pdst0.ptd[i*pdst0.sl  ];
        j2:=pdst0.ptd[i*pdst0.sl+1];
        if j1<xmin then xmin:=j1;
        if j1>xmax then xmax:=j1;
        if j2<ymin then ymin:=j2;
        if j2>ymax then ymax:=j2;
    end;
    if (abs(xmin-res.xmin)>simg_mina) or (abs(ymin-res.ymin)>simg_mina) or (abs(xmax-res.xmax)>simg_mina) or (abs(ymax-res.ymax)>simg_mina) then cc.warning('SImg_GetAlignement: Found destination sources out of boundaries, adjusted boundaries');
    res.xmin:=xmin;
    res.ymin:=ymin;
    res.xmax:=xmax;
    res.ymax:=ymax;

    { check boundaries of destination source list }
    xmin:=pref0.ptd[0]; xmax:=xmin;
    ymin:=pref0.ptd[1]; ymax:=ymin;
    for i:=1 to pref0.ny-1 do
    begin
        j1:=pref0.ptd[i*pref0.sl  ];
        j2:=pref0.ptd[i*pref0.sl+1];
        if j1<xmin then xmin:=j1;
        if j1>xmax then xmax:=j1;
        if j2<ymin then ymin:=j2;
        if j2>ymax then ymax:=j2;
    end;

    { limit number os sources such that source density ration is approximately equal }
    j1:=(res.xmax-res.xmin+1)*(res.ymax-res.ymin+1);
    j2:=(xmax-xmin+1)*(ymax-ymin+1);
    { pdst0.ny/j1 < density_factor*pref0.ny/(j2*con.smin) }
    if con.smin>simg_mina then
    begin
        i:=round(density_factor*pref0.ny*j1/(j2*sqr(con.smin)));
//      writeln(stderr,'i=',i,'  pdst0.ny=',pdst0.ny);
        if i<pdst0.ny then
        begin
            cc.info(cc.name+': Reduced number of destination sources from '+int2str(pdst0.ny)+' to '+int2str(i)+' in order to achieve similar source density as in reference source list');
            pdst0.ny:=i;
        end;
    end;
    { pref0.ny/(j2*con.smax) < density_factor*pdst0.ny/j1 }
    j:=round(density_factor*pdst0.ny*sqr(con.smax)*j2/j1);
//    writeln(stderr,'j=',j,'  pref0.ny=',pref0.ny);
    if j<pref0.ny then
    begin
        cc.info(cc.name+': Reduced number of reference sources from '+int2str(pref0.ny)+' to '+int2str(j)+' in order to achieve similar source density as in destination source list');
        pref0.ny:=i;
    end;

    { calculate chains if necessary }
    pref.pt:=nil; pdst.pt:=nil;
    if pref0.nx<11 then SImg_GetSourceChains(cc,pref0,pref)
    else pref:=pref0;
    if pdst0.nx<11 then SImg_GetSourceChains(cc,pdst0,pdst)
    else pdst:=pdst0;

    { res.srad }
    j1:=0;  j2:=0;
    for i:=0 to pref.ny-1 do
    begin
        j1+=sqr(pref.ptd[i*pref.sl  ]);
        j2+=sqr(pref.ptd[i*pref.sl+1]);
    end;
    j1/=pref.ny;
    j2/=pref.ny;
    res.srad:=j1+j2;
    res.fact:=1;
    res.weight:=1;

    { dl settings }
    depth:=dl;
    if depth<1 then depth:=1;
    if sqr(con.rx)>2*j1 then depth+=1;
     if sqr(con.ry)>2*j2 then depth+=1;
//ifdef[align_debug]
    cc.info('getalignment: depth='+int2str(depth)+'  xmax='+real2str(sqrt(2*j1),-1)+'  ymax='+real2str(sqrt(2*j2),-1));
//endif

    a:=0.01;
    if dl=1 then a:=0.04
    else if dl=2 then a:=0.08
    else if dl>2 then a:=0.15;

    ml:=2;
//    if dl=2 then ml:=3
//    else if dl>2 then ml:=4;

    if (con.sh<1e-6) then
    begin
        con.sh:=1;
        if dl=1 then con.sh:=2
        else if dl=2 then con.sh:=4
        else if dl>=2 then con.sh:=8;
    end;

    { calculate match values
      each match value consists in position of a source relative of farther+brighter source }
    cc.info('getalignment: Calculating match values');
    i:=depth;
    if pref.ny<500 then i+=1;                      { depth }
    nref:=((pref.nx-8) div 3)*i - (i*(i-1)) div 2; { number of match values per source }
    getmem2(mvr,nref*sizeof(TSImg_Alignment_matchval)*pref.ny);
    i:=SImg_GetAlignment_matchVals(pref,nref,i,a,mvr,con.flip);
//ifdef[align_debug]
    cc.info('Found ' +int2str(i) +' match values for reference');
//endif

    i:=depth;
    if pdst.ny<1000 then i+=1;
    ndst:=((pdst.nx-8) div 3)*i - (i*(i-1)) div 2;
    getmem2(mvd,ndst*sizeof(TSImg_Alignment_matchval)*pdst.ny);
    i:=SImg_GetAlignment_matchVals(pdst,ndst,i,a,mvd,con.flip);
//ifdef[align_debug]
    cc.info('Found ' +int2str(i) +' match values for destination');
//endif

    { find matches }
    cc.info('getalignment: Searching initial matches using '+ int2str(cc.threads) + ' threads');
    getmem2(omatches,sizeof(TSImg_Alignment_matches)*pref.ny);
    for i:=0 to pref.ny-1 do
        with omatches[i] do
        begin
            ref:=i;
            cnt:=0;
        end;
    tn:=cc.threads;
    for i:=0 to tn-1 do
        dothread SImg_GetAlignment_find(pref,pdst, mvr,mvd, nref,ndst, ml, i,tn, omatches, con);
    threadwait;

    freemem2(mvd,ndst*sizeof(TSImg_Alignment_matchval)*pdst.ny);
    freemem2(mvr,nref*sizeof(TSImg_Alignment_matchval)*pref.ny);

    getmem2(smatches,sizeof(TSImg_Alignment_matches)*pref.ny);
    m:=SImg_GetAlignment_sort(pref.ny, omatches, smatches);
    cc.info('getalignment: Found '+int2str(m)+' initial matches');
    if m<3 then cc.error('getalignment: At least two initial matches required');

{ test: shift: 1000,1100: rotation:-53.13°, scale:1.6 }
(*    res.dx:=-1000; res.dy:=300;
    s:=1/res.srad;
    for i:=0 to pdst.ny-1 do
    begin
        j:=i*pdst.sl;
        r:=s*(sqr(pdst.ptd[j])+sqr(pdst.ptd[j+1]));
        r:=0.1*r+0.05*sqr(r)-0.01*sqr(r)*r;
        a:=pdst.ptd[j  ] + (pdst.ptd[j  ])*r;
        b:=pdst.ptd[j+1] + (pdst.ptd[j+1])*r;

        r:=s*(sqr(pdst.ptd[j]-res.dx)+sqr(pdst.ptd[j+1]-res.dy));
        r:=(-0.04*r+0.07*sqr(r)-0.02*sqr(r)*r)*sqrt(r);
        a+=(pdst.ptd[j  ]-res.dx)*r;
        b+=(pdst.ptd[j+1]-res.dy)*r;

{       r:=(sqr(pdst.ptd[j]-30000)+sqr(pdst.ptd[j+1]-res.dy)-10000)*3e-11;
        a+=r*(pdst.ptd[j]-30000);
        b+=r*(pdst.ptd[j+1]-10000); }

        pdst.ptd[j  ]:=a*0.9-b*1.0 + 1000;
        pdst.ptd[j+1]:=a*1.2+b*0.9 + 1100;
    end; *)

{ Find initial solution.
  During iteration initial matches (omatches) are (re-)used and only unlikely matches are sorted out. }
    cc.info('getalignment: Searching initial solution using '+ int2str(cc.threads) + ' threads');
    getmem2(res2,sizeof(TSImg_AlignmentParams)*tn);
    dbuf:=nil;
    for j:=0 to 2 do
    begin
        if j>0 then SImg_GetAlignment_sort(pref.ny, omatches, smatches);
        for i:=0 to tn-1 do
        begin
            (res2+i)^:=res;
            dothread SImg_GetAlignment_solveinit(pref,pdst, omatches, smatches, j, i,tn, res2+i, @con);
        end;
        threadwait;

        k:=0;
        for i:=1 to tn-1 do
            if res2[i].j1>res2[k].j1 then k:=i;
        res:=res2[k];
        if not SImg_GetAlignment_checkConstraints(res,con,false,cc) then cc.error('No initial solution within constraints found');

        SImg_GetAlignment_find2(pref,pdst, SImg_sources_mindist, res, smatches,dbuf);
//ifdef[align_debug]
        cc.info('Initial solution:' +int2str(j+1) + ': ' + SImg_GetAlignment_info(res,';   '));
//endif
        if j<2 then
        begin
            m:=0;
            for i:=0 to pref.ny-1 do
            begin
                if (smatches[i].cnt<1) or (omatches[i].dst<>smatches[i].dst) then omatches[i].cnt:=0;
                if omatches[i].cnt>0 then m+=1;
            end;
//ifdef[align_debug]
            cc.info('Initial solution:' +int2str(j+1) + ': ' + int2str(m) + ' remaining matches');
//endif
            if (j=0) and (m<5) then
            begin
                cc.warning('Less than 5 save matches');
                break;
            end
            else if (j=1) and (m<10) then
            begin
                cc.warning('Less than 10 save matches');
                break;
            end;
        end;

        if (dbuf=nil) then getmem2(dbuf,sizeof(TSImg_Alignment_find2buf)*pdst.ny);
    end;
    freemem2(res2,sizeof(TSImg_AlignmentParams)*tn);

{ find final solution
  Uses newly calculated matches from last initial iteration }
    freemem2(omatches,sizeof(TSImg_Alignment_matches)*pref.ny);
    omatches:=smatches;
    r:=SImg_sources_mindist;
    cc.info('getalignment: Searching final solution using up to '+ int2str(tn) + ' threads');
    j1:=0; j2:=0; j3:=0;
    for i:=0 to 256 do
    begin
        if (r<10) and (i>5) then
        begin
            r:=10;
            j1:=0.7*j1 + 0.3*res.j1;
            j2:=0.7*j2 + 0.3*res.j2;
            j3:=0.7*j3 + 0.3*res.j3;
        end;
// define[align_solvertest]
//ifdef[align_solvertest]
        SImg_GetAlignment_solve(omatches, 0,pref.ny-1, 9, res, 0);
        a:=SImg_GetAlignment_J(omatches, pref.ny, res);
//endif
        SImg_GetAlignment_solve(omatches, 0,pref.ny-1, 13, res, select(i<5,1,0.1));
//ifdef[align_solvertest]
        cc.info('Iteration '+int2str(i+1,2)+': J(9)='+real2str(a)+'  J(13)='+real2str(SImg_GetAlignment_J(omatches, pref.ny, res)));
//endif
        if (i<5) and (not SImg_GetAlignment_checkConstraints(res,con,false,nil)) then
        begin
            cc.warning('Solution with 13 DOF not within constraints, trying 9 DOF');
            SImg_GetAlignment_solve(omatches, 0,pref.ny-1, 9, res, 1);
//          if not SImg_GetAlignment_checkConstraints(res,con,false,nil) then cc.error('Solution not within constraints');
        end;
        SImg_GetAlignment_find2(pref,pdst, r, res, omatches,dbuf);
        if res.j1+res.j2+res.j3<1.001*(j1+j2+j3) then break;
//ifdef[align_debug]
        cc.info('Iteration '+int2str(i+1,2)+': ' + SImg_GetAlignment_info(res,';   ') + condstring(r<10.001, '' ,' *') );
//endif
        r*=0.95;
    end;
    cc.info('getalignment: Final solution:' + lineEnding + SImg_GetAlignment_info(res,lineEnding));
    cc.info('getalignment: Matches: '+int2str(SImg_GetAlignment_factor(pref,pdst,omatches,pref.ny,res.fact)));
    cc.info('getalignment: Brightnes scaling of destination: '+real2str(res.fact,8));
    if not SImg_GetAlignment_checkConstraints(res,con,false,cc) then cc.error('Solution not within constraints');

    if pref.pt<>pref0.pt then SImg_Free(cc,pref);
    if pdst.pt<>pdst0.pt then SImg_Free(cc,pdst);
    freemem2(omatches,sizeof(TSImg_Alignment_matches)*pref.ny);
    freemem2(dbuf,sizeof(TSImg_Alignment_find2buf)*pdst.ny);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_GetAlignment2 **************************************************}
{*******************************************************************************}
{ Aligns a bright diffuse source to center
    fi  : input image
    rad : approximate radius of source
    res : result
    cx  : constraint: x position
    cy  : constraint: y position
    cr  : constraint: radius
}
procedure SImg_GetAlignment2(cc:CSImg_Config; fi:TSImg_FPImage; rad:double; var res:TSImg_AlignmentParams; cx,cy,cr:integer);
var fj,fk       : TSImg_FPImage;
    x,y,xm,ym   : longint;
    x0,x1,y0,y1 : longint;
    r,s,t       : simgfp;
    d,e         : ^simgfp;
    u,xr,yr     : double;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetAlignment2: pref='+SImg_InfoStr2(fi)+' rad='+real2str(rad)+' cx='+int2str(cx)+' cy='+int2str(cy)+' cr='+int2str(cr));
    if not SImg_Recover(cc,fi) then cc.Error('SImg_GetAlignment: Invalid input image');

    { optional 1-color conversion and Gauss filtering }
    fj.pt:=nil; fk.pt:=nil;
    if fi.nc<>1 then
    begin
        SImg_New(cc, fk, fi.nx,fi.ny,1);
        d:=fi.pt-1;
        for x:=0 to fk.size-1 do
        begin
            r:=0;
            for y:=1 to fi.nc do
                r+=d[y];
            d+=fi.nc;
            fk.pt[x]:=r;
        end;
        SImg_Filter(cc,fk,fj,rad);
        SImg_Filter(cc,fk,fk,rad*2);
    end
    else begin
        SImg_Filter(cc,fi,fj,rad);
        SImg_Filter(cc,fi,fk,rad*2);
    end;

    { find maximum within constraints }
    x0:=cx-cr;
    if x0<1 then x0:=1;
    x1:=cx+cr;
    if x1>fj.nx-2 then x1:=fj.nx-2;
    y0:=cy-cr;
    if y0<1 then y0:=1;
     y1:=cy+cr;
    if y1>fj.ny-2 then y1:=fj.ny-2;
    cr:=sqr(cr);
    s:=0; xm:=cx; ym:=cy;
    for y:=y0 to y1 do
    begin
        d:=fj.pt + y*fj.sl;
        e:=fk.pt + y*fk.sl;
        for x:=x0 to x1 do
        begin
            if sqr(x-cx)+sqr(y-cy)<=cr then
            begin
                r:=d[x]-e[x];
                if r>s then
                begin
                    xm:=x;
                    ym:=y;
                    s:=r;
                end;
            end;
        end;
    end;

    { interpolation }
    d:=fj.pt + ym*fj.sl + xm;
    e:=fk.pt + ym*fk.sl + xm;
    r:=d[-fj.sl]-e[-fk.sl];
    t:=d[ fj.sl]-e[ fk.sl];
    u:=2*r-4*s+2*t;
    if abs(u)<1e-100 then yr:=ym
    else yr:=ym + (r-t)/u;

    r:=d[-1]-e[-1];
    t:=d[ 1]-e[ 1];
    u:=2*r-4*s+2*t;
    if abs(u)<1e-100 then xr:=xm
    else xr:=xm+(r-t)/u;

    { write output. The maximum is the center of destination image. Transformation and distortion parameters are trivial. }
    res.xmin:=-xr;
    res.ymin:=-yr;
    res.xmax:=fi.nx-1-xr;
    res.ymax:=fi.ny-1-yr;
    res.ax :=0;
    res.ay :=0;
    res.b11:=1;
    res.b12:=0;
    res.b21:=0;
    res.b22:=1;
    res.c1 :=0;
    res.c2 :=0;
    res.c3 :=0;
    res.c4 :=0;
    res.c5 :=0;
    res.c6 :=0;
    res.d  :=0;
    res.fact:=1;
    res.weight:=1;

    SImg_Free(cc,fj);
    SImg_Free(cc,fk);
    SImg_DecDebugLevel;
end;
