{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

const colorize_l_r    = 0.299;  { red factor for luminance calculation }
      colorize_l_g    = 0.587;  { green factor for luminance calculation }
      colorize_l_b    = 0.114;  { blue factor for luminance calculation }

{*************************************************************************}
{******* SImg_ColorFieldRatio ********************************************}
{*************************************************************************}
{ ti    : thread index
  tn    : thread number
  sig   : noise
  fu    : uncompressed image
  fc    : compressed image
  fd    : de-saturation image
  fo    : output
  b     : black level for fu
}
threadprocedure SImg_ColorFieldRatio_thread(ti:longint; tn:longint; fu:TSImg_FPImage; fc:TSImg_FPImage; fd:TSImg_FPImage; fo:TSImg_FPImage; b:double);
var a,c         : double;
    i,ia,iz     : longint;
begin
    ia:=(fu.size*ti) div tn;
    iz:=(fu.size*(ti+1)) div tn-1;
    if fd.pt=nil then
        for i:=ia to iz do
        begin
            a:=fu.pt[i]-b;
            c:=fc.pt[i];
            if (c>=simg_mina) and (a>c) then a:=1-c/a
            else a:=0;
            fo.pt[i]:=a;
        end
    else
        for i:=ia to iz do
        begin
            a:=fu.pt[i]-b;
            c:=fc.pt[i];
            if (a<simg_mina) or (c<simg_mina) then
            begin
                a:=1;
                c:=1;
            end
            else begin
                a:=1/a;
                c*=a;
                if c>1 then c:=1;
                a*=fd.pt[i];
                if a<0 then a:=0
                else if a>1 then a:=1;
            end;
            fo.pt[i]:=1-c;
            fd.pt[i]:=a;
        end;
end;

{ Get color field (1 color image with values between 0 and 1 which is mapped to a color) from ratio of two images
  fu    : uncompressed image. Black lavel is adjusted to black lavel of fc.
  fc    : compressed image
  fd    : de-saturation image (optional, if given it acts as input and output)
  fo    : output image: 0 if fc<0 or fu<fc, otherwise 1-fc/fu
}
procedure SImg_ColorFieldRatio(cc:CSImg_Config; fu,fc,fd:TSImg_FPImage; var fo:TSImg_FPImage);
var i,j         : longint;
    a,b         : simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('ColorFieldRatio: Called with: fu='+SImg_InfoStr2(fu)+' fc='+SImg_InfoStr2(fc)+' fd='+SImg_InfoStr2(fd)+' fo='+SImg_InfoStr2(fo));
    if not SImg_Recover(cc,fu) then cc.error('ColorFieldRatio: Invalid image: fu');
    if not SImg_Recover(cc,fc) then cc.error('ColorFieldRatio: Invalid image: fc');
    if fu.size<>fc.size then cc.error('ColorFieldRatio: Incompatible image dimension');
    if fd.pt<>nil then
    begin
        if not SImg_Recover(cc,fd) then cc.error('ColorFieldRatio: Invalid image: fd');
        if (fd.size<>fu.size) or (fd.nc<>1) then cc.error('ColorFieldRatio: Invalid dimensions: fd');
    end;
    if (not SImg_OverWriteAlert(fc,fo)) and (not SImg_OverWriteAlert(fu,fo)) then SImg_New(cc, fo, fc.ny,fc.ny,fc.nc, false);

    { black level adjustment }
    j:=1;
    for i:=0 to fc.size-1 do
    begin
        a:=fc.pt[i];
        if a<simg_mina then j+=1;
    end;
    cc.info(cc.name+': Black level adjusted to '+real2str(j/(fc.size+1)*100,-1)+ '% quantile');
    b:=SImg_GetQuantile(cc, fu, j/(fc.size+1));

    j:=cc.threads;
    for i:=0 to j-1 do
        dothread SImg_ColorFieldRatio_thread(i,j, fu,fc,fd,fo, b);
    threadwait;

    SImg_DecDebugLevel;
end;

procedure SImg_ColorFieldRatio(cc:CSImg_Config; fu,fc:TSImg_FPImage; var fo:TSImg_FPImage);
var fd  : TSImg_FPImage;
begin
    fd.pt:=nil;
    SImg_ColorFieldRatio(cc, fu,fc,fd, fo);
end;

{*************************************************************************}
{******* SImg_TransformColorMap ******************************************}
{*************************************************************************}
{ calculate transformed color map from base color map
    fi  : brightness image
    fc  : color field
    bmap: base color map: n+1 rows (n>=1) consisting in 5 values: 1 position, 1 reference luminance, 3 color values (RGB)
    ml  : Quantile of color field where color map starts.
    mu  : Quantile of color field where color map ends.
          If both quantiles are out of range fixed values 0 and 1 are used.
    mi  : Invert color map
    me  : Color map histogram equalization. Valid: 0..1
    mb  : Color map histogram balance. <0 overweight smaller values, >0 overweight larger values. No effect if me=0
    ms  : Saturation. Valid: 0..1, Feasible: 0.25...0.5
    map : output map
}
procedure SImg_TransformColorMap(cc:CSImg_Config; fi,fc:TSImg_FPImage; const bmap:array of simgfp; ml,mu:double; mi:boolean; me,mb,ms:double; var map:TSImg_ColorMap);
const hist_size = 2048;
      hist_reg  = 0.1;  { histogram regularization }
      bmap_ss   = 8;    { bmap sub-sampling }
var n,n2,i,j,k,c        : longint;
    p,q,r,s             : double;
    a,cl,cu             : simgfp;
    mlv,muv             : boolean;
    hist                : array[0..hist_size-1] of double;
    xt                  : array of simgfp;      { bmap -> map translation table }
    cmap                : ^simgfp;
label nohist;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_TransformColorMap: Called with: fi='+SImg_InfoStr2(fi)+' fc='+SImg_InfoStr2(fc)+' bmap.size='+int2str(high(bmap)+1)+' ml='+real2str(ml)+' mu='+real2str(mu)+' mi='+boolean2str(mi)+' me='+real2str(me)+' mb='+real2str(mb)+' ms='+real2str(ms));

    { validate bmap }
    n:=high(bmap) div 5;
    if n<1 then cc.error('SImg_TransformColorMap: Base color map must contain at least 2 rows');
    for i:=1 to n do
        if bmap[i*5]-bmap[(i-1)*5]<simg_mina then cc.error('SImg_TransformColorMap: Invalid positions in base color map at row ' +int2str(i));
    for i:=0 to n do
    begin
        s:=bmap[i*5+1];
        if s<0 then s:=0;
        if s>1 then s:=1;
        r:=0;
        for c:=0 to 2 do
        begin
            if bmap[i*5+2+c]<=-simg_mina then cc.error('SImg_TransformColorMap: Negative color value '+int2str(c)+' in color map at row ' +int2str(i));
            r+=bmap[i*5+c];
        end;
        if (s>simg_mina) and (r<simg_mina) then cc.error('SImg_TransformColorMap: Invalid color or luminance in color map at row '+int2str(i));
    end;
    if mb<-1 then mb:=-1
    else if mb>1 then mb:=1;
    if me<0 then me:=0
    else if me>1 then me:=1;

    mlv:=(ml>-simg_mina) and (ml<1+simg_mina);
    muv:=(mu>-simg_mina) and (mu<1+simg_mina);
    if not mlv then ml:=0
    else if ml<0 then ml:=0
    else if ml>1 then ml:=1;
    if not muv then mu:=1
    else if mu>1 then mu:=1
    else if mu<0 then mu:=0;
    if abs(mu-ml)<simg_mina then cc.error('SImg_TransformColorMap: Quantiles to close');
    if mu<ml then
    begin
        cc.info(cc.name+': upper quantiles smaller then lower quantile. inverting color map and equalization balance');
        r:=mu;
        mu:=ml;
        ml:=r;
        mb*=-1;
    end;
    { interpolate bmap + invert }
    n2:=n*bmap_ss;
    getmem(cmap,sizeof(simgfp)*5*(n2+1));
    s:=1/bmap_ss;
    for i:=0 to n2 do
    begin
        j:=i div bmap_ss;
        if j>=n then j:=n-1;
        q:=i*s-j;
        if mi then
        begin
            for c:=0 to 4 do
                cmap[i*5+c]:=bmap[(n-j)*5+c]*(1-q) + bmap[(n-j-1)*5+c]*q;
            cmap[i*5]:=-cmap[i*5];
        end
        else
            for c:=0 to 4 do
                cmap[i*5+c]:=bmap[j*5+c]*(1-q) + bmap[(j+1)*5+c]*q;
    end;
    map.l:=0; map.u:=1;
    setlength(xt,n2+1);
    r:=1/(cmap[n2*5]-cmap[0]);
    for i:=0 to n2 do
        xt[i]:=(cmap[i*5]-cmap[0])*r;
    if (not muv) and (not mlv) then goto nohist;

    if not SImg_Recover(cc,fi) then cc.error('SImg_TransformColorMap: Invalid image: fi');
    if not SImg_Recover(cc,fc) then cc.error('SImg_TransformColorMap: Invalid image: fc');
    if fi.size<>fc.size then cc.error('SImg_TransformColorMap: Incompatible image dimension');

   { min max }
    cl:=fc.pt[0];
    cu:=fc.pt[0];
    for i:=1 to fc.size-1 do
    begin
        a:=fc.pt[i];
        if a<cl then cl:=a;
        if a>cu then cu:=a;
    end;
    if cu-cl<simg_mina then cc.error('SImg_TransformColorMap: Constant image: fc');

    { histogram }
    for i:=0 to hist_size-1 do
        hist[i]:=0;
    r:=(hist_size-0.001)/(cu-cl);
    s:=0;
    for i:=0 to fc.size-1 do
    begin
        j:=trunc((fc.pt[i]-cl)*r);
        a:=fi.pt[i];
        if a>simg_mina then
        begin
            s+=a;
            hist[j]+=a;
        end;
    end;
    if s<simg_mina then cc.error('SImg_TransformColorMap: Non-positive image: fi');
    r:=s*hist_reg/hist_size;
    s*=1+hist_reg;
    for i:=0 to hist_size-1 do
        hist[i]+=r;

    { quantile }
    j:=0;
    q:=0;
    if mb<-1 then mb:=-1
    else if mb>1 then mb:=1;
    for i:=0 to n2 do
    begin
        p:=xt[i];
        p:=mb*sqr(p) + (1-mb)*p;
        p:=(p*(mu-ml)+ml)*s;    { desired sum }
        while (q+hist[j]<p) and (j<hist_size-1) do
        begin
            q+=hist[j];
            j+=1;
        end;
        xt[i]:=j+(p-q)/hist[j]; { position in histogram }
//c.info(intstr(i)+': xt='+real2str(xt[i])+' q='+real2str(q)+' hist['+int2str(j)+']='+real2str(hist[j])+' p='+real2str(p));
    end;

    { upper / lower limits }
    r:=(cu-cl)/(hist_size-0.001);
    map.l:=xt[0]*r+cl;
    map.u:=xt[n2]*r+cl;

    { convert to map values }
    s:=xt[0];
    r:=1/(xt[n2]-s);
    q:=1/(cmap[n2]-cmap[0]);
    for i:=0 to n2 do
        xt[i]:=(xt[i]-s)*r*me + (cmap[i*5]-cmap[0])*q*(1-me);
{    for i:=0 to n do
        cc.info(intstr(i) + ' --> ' + real2str(xt[i],-3));}

nohist:
    { transform map }
    for c:=0 to 3 do
        map.m[0,c]:=cmap[c+1];
    j:=0;
    for i:=1 to n2 do
    begin
        k:=round(xt[i]*TSImg_ColorMap_size);
        if k>j then
        begin
            r:=1/(k-j);
            for j:=j+1 to k do
            begin
                s:=(k-j)*r;
                for c:=0 to 3 do
                    map.m[j,c]:=cmap[(i-1)*5+c+1]*s + cmap[i*5+c+1]*(1-s);
            end;
        end;
    end;

    { saturation }
    if ms<1-1e-6 then
    begin
        for i:=0 to TSImg_ColorMap_size do
            for c:=1 to 3 do
                map.m[i,c]:=map.m[i,c]*ms + (1-ms);
    end;


    { calibrate luminance }
    for i:=0 to TSImg_ColorMap_size do
    begin
        r:=map.m[i,0];
        if r<0 then r:=0;
        if r>1 then r:=1;
        map.m[i,0]:=r;
        r:=( r + 1e-6 )/( sqrt( colorize_l_r*sqr(map.m[i,1]) + colorize_l_g*sqr(map.m[i,2]) + colorize_l_b*sqr(map.m[i,3])) + 1e-6 );
        for c:=1 to 3 do
            map.m[i,c]*=r;
    end;

    setlength(xt,0);
    freemem(cmap);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{********* SImg_Colorize *************************************************}
{*************************************************************************}
{ ti : thread index
  tn : number of threads
  fi: brightness image
  fc: color map input, range: 0..1
  fd: color de-saturation image, range 0 (no de-saturation) .. 1 (white)
  fj: output image
  map: internal color map
  n : number of entries in color map
  ul: upper limit
}
threadprocedure SImg_Colorize_thread(ti:longint; tn:longint; fi:TSImg_FPImage; fc:TSImg_FPImage; fd:TSImg_FPImage; fj:TSImg_FPImage; l:simgfp; u:simgfp; mapm:psimgfp; ul:simgfp);
var i,j,iz,c    : longint;
    r,s,t,w,v   : double;
    d           : ^simgfp;
    sat         : boolean;
begin
    i:=(fi.size*ti) div tn;
    iz:=(fi.size*(ti+1)) div tn - 1;
    u:=TSImg_ColorMap_size/(u-l);
    for i:=i to iz do
    begin
        s:=(fc.pt[i]-l)*u;
        if s<0 then s:=0;
        if s>TSImg_ColorMap_size then s:=TSImg_ColorMap_size;
        j:=round(s+0.5);
        if j<1 then j:=1;
        if j>TSImg_ColorMap_size then j:=TSImg_ColorMap_size;
        s:=j-s;
        r:=fi.pt[i];
        j*=4;
        sat:=false;
        d:=fj.pt+i*3;
        if fd.pt=nil then
        begin
            for c:=0 to 2 do
            begin
                t:=r*(mapm[j-3+c]*s + mapm[j+1+c]*(1-s));
                if t>ul then
                begin
                    t:=ul;
                    sat:=true;
                end;
                d[c]:=t;
            end;
        end
        else begin
            w:=fd.pt[i];
            if w<0 then w:=0;
            if w>1 then w:=1;
            v:=r*(mapm[j-4]*s + mapm[j]*(1-s))*w;  { <desired brightness> * <de-saturation> }
            w:=1-w;
            for c:=0 to 2 do
            begin
                t:=r*(mapm[j-3+c]*s + mapm[j+1+c]*(1-s))*w + v;
                if t>ul then
                begin
                    t:=ul;
                    sat:=true;
                end;
                d[c]:=t;
            end;
        end;
        { handle overflow by de-saturation) }
        if sat then
        begin
            { multiply reference brightness -> desired brightness }
            r*=mapm[j-4]*s + mapm[j]*(1-s);
            { reduce saturation while trying to archive desired brightness }
            for j:=0 to 1 do
            begin
                s:=ul-sqrt(colorize_l_r*sqr(d[0]) + colorize_l_g*sqr(d[1]) + colorize_l_b*sqr(d[2]));
                if s<simg_mina then t:=0
                else t:=(ul-r)/s;
                for c:=0 to 2 do
                    d[c]:=d[c]*t + (1-t)*ul;
            end;
        end;

    end;
end;

{ Colorize 1-color image
  fi: input image used for brightness
  fc: input image used for color, clipped to range: map.l .. map.u
  fd: color de-saturation image, range: 0 (no de-saturation) .. 1 (white)
  fo: output image
  map: color map
  sat: clip to brightest value and correct brightness by de-saturation
}
procedure SImg_Colorize(cc:CSImg_Config; fi,fc,fd:TSImg_FPImage; var fo:TSImg_FPImage; const map:TSImg_ColorMap; clip:boolean);
var i,j : longint;
    r   : double;
    fj  : TSImg_FPImage;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Colorize: called with: fi='+SImg_InfoStr2(fi)+' fc='+SImg_InfoStr2(fc)+' fd='+SImg_InfoStr2(fd)+' fo='+SImg_InfoStr2(fo)+' clip='+boolean2str(clip));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Colorize: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_Colorize: 1 color image expected: fi');
    if not SImg_Recover(cc,fc) then cc.error('SImg_Colorize: Invalid image: fc');
    if (fc.nx<>fi.nx) or (fc.ny<>fi.ny) or (fc.nc<>1) then cc.error('SImg_Colorize: Invalid dimensions: fc');
    if fd.pt<>nil then
    begin
        if not SImg_Recover(cc,fd) then cc.error('SImg_Colorize: Invalid image: fd');
        if (fd.nx<>fi.nx) or (fd.ny<>fi.ny) or (fd.nc<>1) then cc.error('SImg_Colorize: Invalid dimensions: fd');
    end;
    if not SImg_OverwriteAlert(fi,fo) and not SImg_OverwriteAlert(fc,fo) and not SImg_OverwriteAlert(fd,fo) then SImg_Free(cc, fo);

    if clip then
    begin
        r:=0;
        for i:=0 to fi.size-1 do
            if fi.pt[i]>r then r:=fi.pt[i];
    end
    else r:=simg_large;

    fj.pt:=nil;
    SImg_New(cc,fj, fi.nx,fi.ny,3, false);
    j:=cc.threads;
    for i:=0 to j-1 do
        dothread SImg_Colorize_thread(i,j, fi,fc,fd,fj, map.l,map.u, @map.m, r);
    threadwait;

    SImg_Move(cc, fj,fo);
    SImg_DecDebugLevel;
end;

procedure SImg_Colorize(cc:CSImg_Config; fi,fc:TSImg_FPImage; var fo:TSImg_FPImage; const map:TSImg_ColorMap; clip:boolean);
var fd : TSImg_FPImage;
begin
    fd.pt:=nil;
    SImg_Colorize(cc, fi,fc,fd, fo, map, clip);
end;


{*************************************************************************}
{********* SImg_ColorizeTest *********************************************}
{*************************************************************************}
{ Test image
  fo    : output image
  map   : color map
  sat   : saturate to brightest value
  lx,ly : size
}
procedure SImg_ColorizeTest(cc:CSImg_Config; var fi:TSImg_FPImage; const map:TSImg_ColorMap; lx,ly:longint);
var fc          : TSImg_FPImage;
    lxb,x,y,i   : longint;
    r,s         : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ColorizeTest: called with: fi='+SImg_InfoStr2(fi)+' lx='+int2str(lx)+' ly='+int2str(ly));

    if ly<10 then ly:=10;
    if lx<ly*2 then lx:=ly*2;
    SImg_New(cc, fc, lx,ly,1, false);
    SImg_New(cc, fi, lx,ly,1, false);
    lxb:=ly div 8;
    lx-=lxb*2;

    y:=ly div 2;
    s:=1/(y*(ly-1-y));
    for x:=-lxb to lx+lxb-1 do
    begin
        r:=map.l+ x/(lx-1)*(map.u-map.l);
        if r<0 then r:=0;
        if r>1 then r:=1;
        i:=x+lxb;
        for y:=0 to ly-1 do
        begin
            fi.pt[i]:=sqr(y*(ly-1-y)*s);
            fc.pt[i]:=r;
            i+=fc.sl;
        end;
        end;
    SImg_Colorize(cc, fi,fc, fi, map, true);

    SImg_free(cc,fc);
    SImg_DecDebugLevel;
end;



{*************************************************************************}
{********* SImg_ColorizeLegend *******************************************}
{*************************************************************************}
{ Draw a legend for a given map.
  fo    : output image
  map   : color map
  sat   : saturate to brightest value
  lx,ly : size
  nl    : number of labels (<2 disables labels)
  lm    : Label mode: >0; number of valid digits, 0: percent, <0 relative to maximum in percent
}
procedure SImg_ColorizeLegend(cc:CSImg_Config; var fi:TSImg_FPImage; const map:TSImg_ColorMap; lx,ly,nl,lm:longint);
const lb = 8;           { boundary, tick length }
var fc          : TSImg_FPImage;
    x,y,i,c     : longint;
    r,s,t       : double;
    str         : shortstring;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_ColorizeTest: called with: fi='+SImg_InfoStr2(fi)+' lx='+int2str(lx)+' ly='+int2str(ly)+' nl='+int2str(nl)+' lm='+real2str(lm));

    if lx<80 then lx:=80;
    if ly<30 then ly:=30;
    SImg_New(cc, fc, lx,ly,1, false);
    SImg_New(cc, fi, lx,ly,1, false);

    lx-=4*lb;
    for x:=-2*lb to lx+2*lb-1 do
    begin
        r:=map.l + x/(lx-1)*(map.u-map.l);
        if r<0 then r:=0;
        if r>1 then r:=1;
        i:=x+2*lb;
        for y:=0 to ly-1 do
        begin
            fi.pt[i]:=1;
            fc.pt[i]:=r;
            i+=fc.sl;
        end;
    end;
    SImg_Colorize(cc, fi,fc, fi, map, false);

    if nl>1 then
    begin
        { find maximum }
        r:=0;
        for i:=0 to fi.size-1 do
            if fi.pt[i]>r then r:=fi.pt[i];

        { draw labels }
        for i:=0 to nl-1 do
        begin
            x:=2*lb + round(i*(lx-1)/(nl-1));

            { draw ticks }
            for y:=0 to lb-1 do
                for c:=0 to 2 do
                    fi.pt[y*fi.sl+x*fi.nc+c]:=r;
            for y:=ly-lb to ly-1 do
                for c:=0 to 2 do
                    fi.pt[y*fi.sl+x*fi.nc+c]:=r;

            { get drawing color }
            y:=(ly div 2)*fi.sl + x*fi.nc;
            s:=sqrt( colorize_l_r*sqr(fi.pt[y]) + colorize_l_g*sqr(fi.pt[y+1]) + colorize_l_b*sqr(fi.pt[y+2]) );
            if s<0.5*r then s:=r
            else s:=0;
            for c:=0 to 2 do
                cc.drawcolor[c]:=s;

            { draw string }
            cc.textBack:=false;
            cc.textVAlign:=0.5;
            cc.textHAlign:=-0.5;
            t:=map.l + (map.u-map.l)*i/(nl-1);
            if lm<0 then str:=int2str(round( t/map.u*100))+'%'
            else if lm=0 then str:=int2str(round(t*100))+'%'
            else str:=real2str(t, lm);
            y:=(ly+1) div 2;
            c:=ly-2*lb;
            if c>44 then SImg_Write12x22(cc, fi, x, y, str)
            else if c>32 then SImg_Write8x16(cc, fi, x, y, str)
            else SImg_Write8x8(cc, fi, x, y, str);
        end;

    end;


    SImg_free(cc,fc);
    SImg_DecDebugLevel;
end;


