{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* SImg_FFTImage routines ******************************************}
{*************************************************************************}
{*************************************************************************}

{*************************************************************************}
{******* SImg_FFTFindBase ************************************************}
{*************************************************************************}
{ return smallest number with optimal base larger than n }
function SImg_FFT_FindBase(n:longint):longint;
var b2  : longint;
begin
    b2:=0;
    while (n>22) or (n=15) do
    begin
        n:=(n+1) shr 1;
        b2+=1;
    end;
    result:=n shl b2;
end;


{*************************************************************************}
{******* SImg_GetImageTableEntry *****************************************}
{*************************************************************************}
function SImg_GetImageTableEntry(fi:TSImg_FFTImage):longint;
begin
    if fi.pt=nil then result:=-1
    else result:=SImg_GetImageTableEntry(fi.opt, fi.bytes);
end;


{*************************************************************************}
{******* SImg_New ********************************************************}
{*************************************************************************}
{ spt: nil: init with zeros; else: copy data from pointer }
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FFTImage; spt:pointer);  { nxr, ny must be set }
var i  : longint;
    d  : ^fftfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_Free(cc,r);
    SImg_IncDebugLevel;
    if (r.nxr<minxsize) or (r.nxr>maxxsize) or (r.ny<minysize) or (r.ny>maxysize) then cc.error('SImg_New: illegal dimensions');
    r.nxc:=r.nxr div 2 + 1;
    r.bytes:=(r.ny*r.nxc*2)*sizeof(fftfp) + TSImg_FFT_ALIGN-1 + 1024*1024;
    i:=SImg_GetEmptyImageTableEntry;
    SImg_Sem_Lock(SImg_Sem_Img);
    try
        with SImg_ImageTable[i] do
        begin
            size:=r.bytes;
            getmem2(pt,size);
            r.opt:=pt;
        end;
    finally
            SImg_Sem_Unlock(SImg_Sem_Img);
    end;
    r.iid:=i;
    cc.debug('SImg_New: allocated '+base1024str(r.bytes)+' bytes for '+SImg_InfoStr2(r)+' at '+ptr2str(r.opt));
    ptruint(r.pt):=(ptruint(r.opt)+TSImg_FFT_ALIGN-1) and not (TSImg_FFT_ALIGN-1);
    if spt=nil then
    begin
        d:=r.pt;
        for i:=0 to r.nxc*2*r.ny-1 do
            d[i]:=0;
    end
    else move(spt^,r.pt^, r.nxc*r.ny*2*sizeof(fftfp));
    if not SImg_Integrity(r) then cc.error('SImg_New: internal error');

    { create plans }
    fftw_plan_with_nthreads(cc.threads);
    r.plan_r2c:=fftw_plan_dft_r2c_2d(r.ny,r.nxr, r.pt, r.pt, FFTW_MEASURE);
    if r.plan_r2c=nil then cc.error('Unable to find forward FFT algorithm');
    r.plan_c2r:=fftw_plan_dft_c2r_2d(r.ny,r.nxr, r.pt, r.pt, FFTW_MEASURE);
    if r.plan_c2r=nil then cc.error('Unable to find reverse FFT algorithm');
    SImg_DecDebugLevel;
end;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FFTImage);
begin
    SImg_New(cc,r, nil);
end;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FFTImage; nxr,ny:longint);
begin
    r.nxr:=nxr; r.ny:=ny;
    SImg_New(cc,r, nil);
end;

procedure SImg_New_FindBase(cc:CSImg_Config; var r:TSImg_FFTImage; nx,ny:longint);
begin
    r.nxr:=SImg_FFT_FindBase(nx);
    r.ny:=SImg_FFT_FindBase(ny);
    SImg_New(cc,r, nil);
end;



{*************************************************************************}
{******* SImg_Free *******************************************************}
{*************************************************************************}
procedure SImg_Free(cc:CSImg_Config; var fi:TSImg_FFTImage);
var i  : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    i:=SImg_GetImageTableEntry(fi);
    if (i>=0) then
    begin
        with SImg_ImageTable[i] do
        begin
            cc.debug('SImg_Free: deallocating '+base1024str(size)+'bytes for '+SImg_InfoStr2(fi)+' at '+ptr2str(pt));
            SImg_Sem_Lock(SImg_Sem_Img);
            try
                freemem2(pt,size);
                pt:=nil; size:=0;
            finally
                SImg_Sem_Unlock(SImg_Sem_Img);
            end;
        end;
        fftw_destroy_plan(fi.plan_r2c);
        fftw_destroy_plan(fi.plan_c2r);
    end;
    fi.bytes:=0; fi.pt:=nil; fi.opt:=nil; fi.iid:=-1;
end;


{*************************************************************************}
{******* SImg_Move *******************************************************}
{*************************************************************************}
procedure SImg_Move(cc:CSImg_Config; var fi,fj:TSImg_FFTImage);  { fi -> fj }
begin
    if not SImg_Overwritealert(fi,fj) then
    begin
        SImg_Free(cc,fj);
        fj:=fi;
        fi.bytes:=0; fi.pt:=nil; fi.opt:=nil; fi.iid:=-1;
    end;
end;


{*************************************************************************}
{******* SImg_Integrity **************************************************}
{*************************************************************************}
function SImg_Integrity(fi:TSImg_FFTImage):boolean;
begin
    SImg_Integrity:=(fi.iid>=0) and (fi.iid<SImg_ImageTableSize) and
        (SImg_GetImageTableEntry(fi)=fi.iid) and
        (fi.nxr>=minxsize) and (fi.nxr<=maxxsize) and
        (fi.ny>=minysize) and (fi.ny<=maxysize) and
        (fi.nxc=(fi.nxr div 2 + 1));
end;


{*************************************************************************}
{******* SImg_OverwriteAlert *********************************************}
{*************************************************************************}
function SImg_OverwriteAlert(fi,fj:TSImg_FFTImage):boolean;
var i,j:longint;
begin
    i:=SImg_GetImageTableEntry(fi);
    j:=SImg_GetImageTableEntry(fj);
    SImg_OverwriteAlert:=(i=j) and (i>=0);
end;


{*************************************************************************}
{******* SImg_Duplicate **************************************************}
{*************************************************************************}
procedure SImg_Duplicate(cc:CSImg_Config; fi:TSImg_FFTImage; var fj:TSImg_FFTImage; dat:boolean);   { fi --> fj; dat=true: copy data }
var spt  : ^fftfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Integrity(fi) then cc.error('SImg_Duplicate: illegal image: fi');
    if SImg_Overwritealert(fi,fj) then
    begin
        if not dat then SImg_Clear(cc, fj);
        exit;
    end;
    spt:=nil;
    if dat then spt:=fi.pt;
    fj.nxr:=fi.nxr;
    fj.ny:=fi.ny;
    SImg_New(cc,fj,spt);
end;

{*************************************************************************}
{******* SImg_InfoStr ****************************************************}
{*************************************************************************}
function SImg_InfoStr(fi:TSImg_FFTImage):string19;
begin
    if SImg_GetImageTableEntry(fi)>=0
    then SImg_InfoStr:=int2str(fi.nxr)+'x'+int2str(fi.ny)
    else SImg_InfoStr:='(unused)';
end;

function SImg_InfoStr2(fi:TSImg_FFTImage):string37;
begin
    SImg_InfoStr2:='FPImg(id='+int2str(SImg_GetImageTableEntry(fi))+', fmt='+SImg_InfoStr(fi)+')';
end;


{*************************************************************************}
{******* SIMG_FFT_R2C ****************************************************}
{*************************************************************************}
procedure SImg_FFT_R2C(cc:CSImg_Config; fi:TSImg_FFTImage);
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Integrity(fi) then SImg_Error('SImg_FFT_R2C: invalid image');
    fftw_execute(fi.plan_r2c);
end;

{*************************************************************************}
{******* SIMG_FFT_C2R ****************************************************}
{*************************************************************************}
procedure SImg_FFT_c2r(cc:CSImg_Config; fi:TSImg_FFTImage);
var r  : double;
    i  : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Integrity(fi) then SImg_Error('SImg_FFT_c2r: invalid image');
    { normalize }
    r:=1/(fi.nxr*fi.ny);
    for i:=0 to fi.nxc*fi.ny*2 do
        fi.pt[i]*=r;
    fftw_execute(fi.plan_C2R);
end;

{*************************************************************************}
{******* SIMG_FFT_R2C_flops **********************************************}
{*************************************************************************}
function SImg_FFT_r2c_flops(cc:CSImg_Config; fi:TSImg_FFTImage):int64;
var a,m,ma  : double ;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Integrity(fi) then SImg_Error('SImg_FFT_r2c: invalid image');
    fftw_flops(fi.plan_r2c, @a,@m,@ma);
    result:=round(a+m+ma*2);
end;

{*************************************************************************}
{******* SIMG_FFT_C2R_flops **********************************************}
{*************************************************************************}
function SImg_FFT_c2r_flops(cc:CSImg_Config; fi:TSImg_FFTImage):int64;
var a,m,ma  : double ;
begin
    if cc=nil then cc:=SImg_Config_default;
    if not SImg_Integrity(fi) then SImg_Error('SImg_FFT_c2r: invalid image');
    fftw_flops(fi.plan_c2r, @a,@m,@ma);
    result:=round(a+m+ma*2) + fi.nxc*fi.ny*2;  { including normalization }
end;


{*************************************************************************}
{******* Initialization / finalization ***********************************}
{*************************************************************************}
{******* SImg_InitFFT ****************************************************}
procedure SImg_InitFFT;
begin
    fftw_init_threads();
    fftw_set_timelimit(fftw_timelimit);
end;

{******* SImg_DoneFFT ****************************************************}
procedure SImg_DoneFFT;
begin
    fftw_cleanup_threads();
end;
