{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* CSImg_MetaInfo **************************************************}
{*************************************************************************}
const SImg_MetaInfoSizeInc = 256;

{ meta infos often used by simg:
BYTEORDR        / BIG ENDIAN or LITTLE_ENDIAN
DATAMIN         / Minimum valid value
DATAMAX         / Maximum valid value
BAYERPAT        / Bayer color pattern, e.g. RG/GB or 01/23
EGAIN           / ADC resolution in electrons/<image unit>
BUNIT           / Image unit
RDNOISE         / Read out noise in <image unit>'s
BGNOISE         / Background noise in <image unit>'s
DCNOISE         / Dark current noise in <image unit>'s
CROPCOMP        / T to mark that there are cropped values (this will be compensated during integer to floating point conversion)
EXPTIME         / Exposure time
FWHMPIX         / Star FWHM in pixels
NOISEFLR        / Noise floor
}

{******* create **********************************************************}
constructor CSImg_MetaInfo.create;
begin
    initCriticalSection(cs);
    reserved:=0;
    cnt:=0;
    keys:=nil;
    vals:=nil;
    coms:=nil;
end;

constructor CSImg_MetaInfo.create(mi:CSImg_MetaInfo);
var i  : longint;
begin
    create;
    if mi=nil then exit;
    for i:=0 to mi.size-1 do
        if mi.key[i]<>'' then add(mi.key[i], mi.val[i], mi.comment[i]);
end;

{******* reset ***********************************************************}
procedure CSImg_MetaInfo.reset;
var i    : longint;
begin
    enterCriticalSection(cs);
    for i:=0 to cnt-1 do
    begin
        keys[i]:='';
        vals[i]:='';
        coms[i]:='';
    end;
    cnt:=0;
    leaveCriticalSection(cs);
end;

{******* destroy *********************************************************}
destructor CSImg_MetaInfo.destroy;
begin
    reset;
    freemem(keys);
    freemem(vals);
    freemem(coms);
    doneCriticalSection(cs);
end;

{******* add *************************************************************}
procedure CSImg_MetaInfo.add(const theKey,theVal,theComment:ansistring);
var i : longint;
begin
    if theKey='' then exit;
    enterCriticalSection(cs);
    i:=0;
    while (i<cnt) and (keys[i]<>theKey) do
        inc(i);
    if i>=reserved then
    begin
        reserved+=SImg_MetaInfoSizeInc;
        reallocmem(keys,reserved*sizeof(ansistring));
        reallocmem(vals,reserved*sizeof(ansistring));
        reallocmem(coms,reserved*sizeof(ansistring));
        fillchar(keys[reserved-SImg_MetaInfoSizeInc], SImg_MetaInfoSizeInc*sizeof(ansistring), 0);
        fillchar(vals[reserved-SImg_MetaInfoSizeInc], SImg_MetaInfoSizeInc*sizeof(ansistring), 0);
        fillchar(coms[reserved-SImg_MetaInfoSizeInc], SImg_MetaInfoSizeInc*sizeof(ansistring), 0);
    end;
    if i=cnt then inc(cnt);
    keys[i]:=theKey;
    vals[i]:=theVal;
    coms[i]:=theComment;
    leaveCriticalSection(cs);
end;

{ add encoded key of the form <key>=<value>[/<comment>] and return an error string if an error occurred or '' otherwise }
function CSImg_MetaInfo.add(sa:ansistring):ansistring;
var j    : longint;
    k,c  : ansistring;
    q    : boolean;
begin
    result:='';
    j:=pos('=',sa);
    if (j<=1) or (j>=length(sa)) then
    begin
        result:='No ''='' found';
        exit;
    end;
    k:=trim(copy(sa,1,j-1));
    sa:=trim(copy(sa,j+1,length(sa)-j));
    j:=1;
    q:=false;
    while (j<=length(sa)) and (q or (sa[j]<>'/')) do
    begin
        if sa[j]='''' then q:=not q;
        inc(j);
    end;
    if j=1 then
    begin
        result:='Empty value';
        exit;
    end;
    c:='';
    if j<length(sa) then c:=trim(copy(sa,j+1,length(sa)-j));
    add(k, trim(copy(sa,1,j-1)), c);
end;

procedure CSImg_MetaInfo.add(mi:CSImg_MetaInfo);
var i  : longint;
begin
    if mi=nil then exit;
    for i:=0 to mi.size-1 do
        if mi.key[i]<>'' then add(mi.key[i], mi.val[i], mi.comment[i]);
end;

{******* delete **********************************************************}
procedure CSImg_MetaInfo.delete(const theKey:ansistring);
begin
    rename(theKey,'');
end;

{******* delete **********************************************************}
{ delete standard raw attributes }
procedure CSImg_MetaInfo.deleteRaw;
begin
    delete('DATAMIN');
    delete('DATAMAX');
    delete('BAYERPAT');
    delete('EGAIN');
    delete('BUNIT');
    delete('RDNOISE');
    delete('CROPCOMP');
    delete('BGNOISE');
end;

{******* rename **********************************************************}
procedure CSImg_MetaInfo.rename(const oldKey,newKey:ansistring);
var i  : longint;
begin
    if oldKey='' then exit;
    enterCriticalSection(cs);
    for i:=0 to cnt-1 do
        if keys[i]=oldKey then keys[i]:=newKey;
    leaveCriticalSection(cs);
end;

{******* get *************************************************************}
function CSImg_MetaInfo.get(const theKey:ansistring; var theVal:ansistring; theComment:pansistring):boolean;
var i   : longint;
begin
    result:=false;
    if theKey='' then exit;
    //enterCriticalSection(cs);
    for i:=0 to cnt-1 do
        if keys[i]=theKey then
        begin
            theVal:=vals[i];
            if theComment<>nil then theComment^:=coms[i];
            result:=true;
        end;
    //leaveCriticalSection(cs);
end;

function CSImg_MetaInfo.get(const theKey:ansistring; var theVal:ansistring):boolean;
begin
    result:=get(theKey, theVal, nil);
end;

function CSImg_MetaInfo.get(const theKey:ansistring; var d:double; theComment:pansistring):boolean;
var i   : longint;
    v   : ansistring;
    r   : double;
begin
    result:=get(theKey, v, theComment);
    if not result then exit;
    system.val(v,r,i);
    if i<>0 then
    begin
        result:=false;
        exit;
    end;
    d:=r;
end;

function CSImg_MetaInfo.get(const theKey:ansistring; var d:double):boolean;
begin
    result:=get(theKey,d,nil);
end;

function CSImg_MetaInfo.get(const theKey:ansistring; var s:single):boolean;
var d : double;
begin
    result:=get(theKey,d,nil);
    s:=d;
end;


{******* getDeg **********************************************************}
{ Get value in degree. Sexagesimal values (separators ':' or ' ') are multiplied with cf. }
function CSImg_MetaInfo.getDeg(const theKey:ansistring; cf:double; var d:double; theComment:pansistring):boolean;
var i,j,k  : longint;
    v      : ansistring;
    r,s    : double;
begin
    result:=get(theKey, v, theComment);
    if not result then exit;

    { decimal conversion }
    system.val(v,r,i);
    if i=0 then
    begin
        d:=r;
        exit;
    end;

    { sexagesimal conversion }
    v:=trimquotes(v);
    trimstr(v);
    if length(v)<1 then
    begin
        result:=false;
        exit;
    end;

    r:=0;
    j:=1;
    while j<=length(v) do
    begin
        k:=j;
        while (j<=length(v)) and (v[j]>' ') and (v[j]<>':') do
            j+=1;

        if (v[k]='-') then
        begin
            cf:=-cf;
            k+=1;
        end;

        system.val(copy(v,k,j-k),s,i);
        if i<>0 then
        begin
            result:=false;
            exit;
        end;

        r+=s*cf;
        cf/=60;

        while (j<=length(v)) and ((v[j]<=' ') or (v[j]=':')) do
            j+=1;
    end;
    d:=r;
end;


{******* getRaDec ********************************************************}
{ Get telescope position in degree. Return false if unknown }
function CSImg_MetaInfo.getRaDec(var ra,dec:double):boolean;
var r,d  : double;
begin
    result:=getDeg('RA',15,r,nil) or getDeg('OBJCTRA',15,r,nil);
    if not result then exit;
    result:=getDeg('DEC',1,d,nil) or getDeg('OBJCTDEC',1,d,nil);
    if result then
    begin
        ra:=r;
        dec:=d;
    end;
end;

{******* getDateTime *****************************************************}
{ Get start of observation date and time as unix epoch time in seconds. Return false if unknown }
function CSImg_MetaInfo.getDateTime(var dt:int64):boolean;
var ds,ts   : ansistring;
    dtd,dtt : TDateTime;
    i       : longint;
    f       : TFormatSettings;
begin
    result:=false;
    if not get('DATE-OBS',ds,nil) then exit;
    ds:=lowercase(trimquotes(ds));
    trimstr(ds);
    if ds[length(ds)]='z' then System.delete(ds,length(ds),1);

    ts:='';
    i:=pos('t',ds);
    if i>0 then
    begin
        ts:=copy(ds,i+1,length(ds)-i);
        ds:=copy(ds,1,i-1);
    end;

    if ts='' then
    begin
        if not get('TIME-OBS',ds,nil) then exit;
        ts:=trimquotes(ts);
        trimstr(ts);
    end;

    f:=defaultFormatSettings;
    f.shortDateFormat:='y/m/d';
    f.dateSeparator:='-';
    f.timeSeparator:=':';
    f.decimalSeparator:='.';
    if not tryStrToDate(ds,dtd,f) then exit;
    if not tryStrToTime(ts,dtt,f) then exit;
    dt:=DateTimeToUnix(ComposeDateTime(dtd,dtt));
    result:=true;
end;



{*************************************************************************}
{******* SImg_GetWCSData *************************************************}
{*************************************************************************}
{ returns false if meta WCS data is incomplete }
function SImg_GetWCSData(cc:CSImg_Config; mi:CSImg_MetaInfo; var wcs:TSImg_WCS):boolean;
var w     : TSImg_WCS;
    s     : ansistring;
    t1,t2 : string[3];
    c1,c2 : longint;  { axis coordinates: -1 unknown, 0:ra , 1dec }
    r1,r2 : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    // required basic values
    result:=false;
    if not mi.get('CRVAL1', w.ref_ra, nil) then exit;
    if not mi.get('CRVAL2', w.ref_dec, nil) then exit;
    if not mi.get('CRPIX1', w.ref_x, nil) then exit;
    if not mi.get('CRPIX2', w.ref_y, nil) then exit;
    if not mi.get('CD1_1', w.cd1_1, nil) then exit;
    if not mi.get('CD1_2', w.cd1_2, nil) then exit;
    if not mi.get('CD2_1', w.cd2_1, nil) then exit;
    if not mi.get('CD2_2', w.cd2_2, nil) then exit;

    // units
    r1:=1;
    if not mi.get('CUNIT1',s,nil) then cc.warning('SImg_GetWCSData: Key ''CUNIT1'' not defined, assuming ''deg''')
    else begin
        s:=lowercase(trimquotes(s));
        trimstr(s);
        if s='arcmin' then r1:=1/60
        else if s='arcsec' then r1:=1/3600
        else if s='rad' then r1:=180/pi
        else if s<>'deg' then cc.warning('SImg_GetWCSData: Unknown unit for axis 1: '''+s+''' assuming ''deg''');
    end;
    r2:=r1;
    w.ref_ra*=r1;
    r1*=pi/180;
    w.cd1_1*=r1;
    w.cd2_1*=r1;

    if not mi.get('CUNIT2',s,nil) then cc.warning('SImg_GetWCSData: Key ''CUNIT2'' not defined, assuming ''deg''')
    else begin
        s:=lowercase(trimquotes(s));
        trimstr(s);
        if s='arcmin' then r2:=1/60
        else if s='arcsec' then r2:=1/3600
        else if s='rad' then r2:=180/pi
        else if s<>'deg' then cc.warning('SImg_GetWCSData: Unknown unit for axis 2: '''+s+''' assuming ''deg''');
    end;
    w.ref_dec*=r2;
    r2*=pi/180;
    w.cd1_2*=r2;
    w.cd2_2*=r2;
    if abs(r1-r2)>1e-10 then cc.warning('SImg_GetWCSData: Different units for axis 1 and 2');

    // type
    t2:='';
    c2:=-1;
    if mi.get('CTYPE2',s,nil) then
    begin
        s:=trimquotes(s);
        trimstr(s);
        if (lowercase(copy(s,1,2))='ra') then c2:=0
        else if (lowercase(copy(s,1,3))='dec') then c2:=1;
        t2:=lowercase(copy(s,6,3));
    end;
    t1:=t2;
    c1:=-1;
    if mi.get('CTYPE1',s,nil) then
    begin
        s:=trimquotes(s);
        trimstr(s);
        if (lowercase(copy(s,1,2))='ra') then c1:=0
        else if (lowercase(copy(s,1,3))='dec') then c1:=1;
        t1:=lowercase(copy(s,6,3));
    end;
    if ((c1=0) and (c2=0)) or ((c1=1) and (c2=1)) or ((c1<0) and (c2<0)) then cc.warning('SImg_GetWCSData: No/unknown/invalid destination coordinate system defined using CTYPE1 and CTYPE2.')
    else if (c1=1) or (c2=0) then
    begin { swap ra - dec }
        r1:=w.ref_ra;
        w.ref_ra:=w.ref_dec;
        w.ref_dec:=r1;
    end;

    if t1='' then
    begin
        cc.warning('SImg_GetWCSData: No projection type given, assuming ''TAN''');
        t1:='tan';
    end
    else if (t2<>'') and (t1<>t2) then cc.warning('SImg_GetWCSData: Inconsistent projection types, using '''+t1+'''');
    w.t:=t1;

    // optional distortion terms
    if not mi.get('A_0_0',w.a0_0,nil) then w.a0_0:=0;
    if not mi.get('A_0_1',w.a0_1,nil) then w.a0_1:=0;
    if not mi.get('A_0_2',w.a0_2,nil) then w.a0_2:=0;
    if not mi.get('A_1_0',w.a1_0,nil) then w.a1_0:=0;
    if not mi.get('A_1_1',w.a1_1,nil) then w.a1_1:=0;
    if not mi.get('A_2_0',w.a2_0,nil) then w.a2_0:=0;
    if not mi.get('B_0_0',w.b0_0,nil) then w.b0_0:=0;
    if not mi.get('B_0_1',w.b0_1,nil) then w.b0_1:=0;
    if not mi.get('B_0_2',w.b0_2,nil) then w.b0_2:=0;
    if not mi.get('B_1_0',w.b1_0,nil) then w.b1_0:=0;
    if not mi.get('B_1_1',w.b1_1,nil) then w.b1_1:=0;
    if not mi.get('B_2_0',w.b2_0,nil) then w.b2_0:=0;

    wcs:=w;
    result:=true;
end;


{*************************************************************************}
{******* SImg_GetRaDec ***************************************************}
{*************************************************************************}
{ calculates equatorial position of a pixel using WCS data
  x,y :  WCS coordinates, bottom left pixel is (1,1) }
procedure SImg_GetRaDec(cc:CSImg_Config; const wcs:TSImg_WCS; x,y:double; var ra,dec:double);
const deg2rad = pi/180;
var u,v,z,a,s,c,r : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    with wcs do
    begin
        // distortions
        x-=ref_x; y-=ref_y;
        u:=x + a0_0 + a0_1*y + a1_0*x + a0_2*sqr(y) + a1_1*x*y + a2_0*sqr(x);
        v:=y + b0_0 + b0_1*y + b1_0*x + b0_2*sqr(y) + b1_1*x*y + b2_0*sqr(x);
        // intermediate world coordinates
        x:=cd1_1*u + cd1_2*v;
        y:=cd2_1*u + cd2_2*v;
        // coordinates on sphere
        if t='sin' then r:=1
        else begin
            r:=sqr(x)+sqr(y);
            if t='tan' then r:=1/sqrt(1+r)
            else if t='stg' then r:=4/(4+r)
            else if t='arc' then
            begin
                r:=sqrt(r);
                r:=(sin(r)+1e-20)/(r+1e-20)
            end
            else cc.error('SImg_GetRaDec: Unsupported projection type:'''+t+'''');
        end;
        x*=r; y*=r;
        z:=1-sqr(x)-sqr(y);
        if z<0 then cc.error('SImg_GetRaDec: Internal error');
        z:=sqrt(z);
        // rotation of sphere
        {sincos(-(90-ref_dec)*deg2rad, s, c);
        r:=c*z - s*y;
        y:=s*z + c*y;}
        sincos(ref_dec*deg2rad, c, s);
        r:= c*z + s*y;
        y:= c*y - s*z;
        // calculation of coordinates
        a:=180/pi;
        dec:=arcsin(r)*a;
        if abs(y)>1e-100
        then ra:=arctan2(x,-y)*a
        else if x>0 then ra:=90
        else ra:=-90;
        ra+=ref_ra;
        if ra>360 then ra-=360
        else if ra<0 then ra+=360;
    end;
end;

{ calculates equatorial position of a pixel using WCS data
  x,y :  picture coordinate, top left pixel is (0,0) }
procedure SImg_GetRaDec(cc:CSImg_Config; const wcs:TSImg_WCS; x,y,height:double; var ra,dec:double);
begin
    SImg_GetRaDec(cc,wcs, x+1,height-y, ra,dec);
end;

{*************************************************************************}
{******* SImg_GetPos *****************************************************}
{*************************************************************************}
{ Calculates approximate pixel position in fits coordinates using WCS data. (Distortions are ignored.)
  ra,dec  : equatorial position
  result  : false if wsc transformation matrix is singular
}
function SImg_GetPos(const wcs:TSImg_WCS; ra,dec:double; var xo,yo:double):boolean;
const deg2rad = pi/180;
var x,y,y2,z,s,c  : double;
begin
    result:=true;
    { 1. rotate -wcs.ra_ref }
    ra-=wcs.ref_ra;
    { 2. convert to spherical coordinates }
    sincos(dec*deg2rad,y2,z);
    sincos(ra*deg2rad,s,c);
    x:=s*z;
    z:=c*z;
    { 3. rotate -wcs.dec_ref }
    sincos(wcs.ref_dec*deg2rad,s,c);
    y:=y2*c - s*z ;
    z:= z*c + s*y2;
    { 4. projection }
    if wcs.t='sin' then s:=1
    else if (wcs.t='tan') and (z>1e-10) then s:=1/z
    else s:=2/(z+1);  { stereographic in all other cases }
    x*=s; y*=s;
    { 5. convert to fits coordinates }
    s:=wcs.cd1_1*wcs.cd2_2 - wcs.cd1_2*wcs.cd2_1;
    if abs(s)<1e-50 then
    begin
        result:=false;
        exit;
    end;
    s:=1/s;
    xo:=s*(wcs.cd2_2*x - wcs.cd1_2*y) + wcs.ref_x;
    yo:=s*(wcs.cd1_1*y - wcs.cd2_1*x) + wcs.ref_y;
end;

{ Calculates approximate pixel position in fits coordinates using WCS data. (Distortions are ignored.)
  ra,dec  : equatorial position
  height  : picture height
  result  : false if WCS transformation matrix is singular
}
function SImg_GetPos(const wcs:TSImg_WCS; ra,dec,height:double; var xo,yo:double):boolean;
begin
    result:=SImg_GetPos(wcs, ra,dec, xo,yo);
    xo-=1;
    yo:=height-yo;
end;


{*************************************************************************}
{******* SImg_GetPixelScale **********************************************}
{*************************************************************************}
{ calculates pixel scale at reference point in arcsecs. dx is positive, dy has the same sign as the determinant }
procedure SImg_GetPixelScale(const wcs:TSImg_WCS; var dx,dy:double);
begin
    dx:=sqrt(sqr(wcs.cd1_1)+sqr(wcs.cd2_1))*180*3600/pi;
    dy:=sqrt(sqr(wcs.cd1_2)+sqr(wcs.cd2_2))*180*3600/pi;
    if wcs.cd1_1*wcs.cd2_2-wcs.cd1_2*wcs.cd2_1 > 0 then dy:=-dy;
end;
