{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[sdb_debug]

{ sub-grid size }
const SImg_StackDB_subgrid_m = 5;
      SImg_StackDB_subgrid_n = 1 shl SImg_StackDB_subgrid_m;

{ sub-grid info }
type TSImg_StackDB_subgrid = record
        nx,ny   : longint;              { size }
        xo,yo   : longint;              { offsets }
        end;

{ in/out info }
type TSImg_StackDB_inout = record
        b         : ^byte;
        smax,smin : single;             { maximum/minimum scale as <image pixels width>/<stackdb pixel width> }
        end;


{*******************************************************************************}
{********* SImg_StackDB_GetInfo ***********************************************}
{*******************************************************************************}
{ nx,ny : database dimensions }
procedure SImg_StackDB_GetInfo(cc:CSImg_Config; var si:TSImg_SDBInfo; mi:CSImg_MetaInfo; nx,ny:longint);
var str    : ansistring;
    s,rot  : single;
    flip   : boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    with si do
    begin
        flip:=false;
        if not mi.get('SDBNFOX0', x0) then cc.error('SImg_StackDB_ReadInfo: Attribute not found: "SDBNFOXC"');
        if not mi.get('SDBNFOY0', y0) then cc.error('SImg_StackDB_ReadInfo: Attribute not found: "SDBNFOYC"');
        if not mi.get('SDBNFOS' , s)  then cc.error('SImg_StackDB_ReadInfo: Attribute not found: "SDBNFOS"');
        if not mi.get('SDBNFOR' , rot) then rot:=0;
        if mi.get('SDBNFOFY', str) then
        begin
            if (str='T') or (str='t') then flip:=true
            else if (str<>'F') and (str<>'f') then cc.error('SImg_StackDB_ReadInfo: Invalid value for key "SDBNFOFY", "T" or "F" expected');
        end;
        xs:=(nx-1) shr SImg_StackDB_subgrid_m + 1;
        ys:=(ny-1) shr SImg_StackDB_subgrid_m + 1;
        xso:=(nx - xs shl SImg_StackDB_subgrid_m) div 2;
        yso:=(ny - ys shl SImg_StackDB_subgrid_m) div 2;
        SImg_StackDB_InitInfo(si,s,rot,flip);
    end;
end;

{*******************************************************************************}
{********* SImg_StackDB_InitInfo ***********************************************}
{*******************************************************************************}
{ Calculate the transformation matrix }
procedure SImg_StackDB_InitInfo(var si:TSImg_SDBInfo; scale,rot:single; flip:boolean);
const deg2rad = pi/180.0;
var c,s,sy  : double;
begin
    if flip then sy:=-scale
    else sy:=scale;
    sincos(rot*deg2rad,s,c);
    si.s11:= c*scale; si.s12:=s*sy;
    si.s21:=-s*scale; si.s22:=c*sy;
end;

{*******************************************************************************}
{********* SImg_StackDB_WriteInfo **********************************************}
{*******************************************************************************}
procedure SImg_StackDB_WriteInfo(const si:TSImg_SDBInfo; mi:CSImg_MetaInfo);
const rad2deg = 180.0/pi;
begin
    mi.add('SDBNFOX0', real2str(si.x0,-2), 'Stack database: X origin');
    mi.add('SDBNFOY0', real2str(si.y0,-2), 'Stack database: Y origin');
    mi.add('SDBNFOS' , real2str(sqrt(sqr(si.s11)+sqr(si.s12)), -6), 'Stack database: pixel scale');
    mi.add('SDBNFOR' , real2str(arctan2(-si.s21,si.s11)*rad2deg, -6), 'Stack database: Clockwise rotation in degree');
    mi.add('SDBNFOFY' , condstring(si.s11*si.s22<0,'T','F'), 'Stack database: Flipping in Y direction (applied after rotation)');
end;


{*******************************************************************************}
{********* SImg_StackDB_inout **************************************************}
{*******************************************************************************}
{ calcultes whether a database points an a grid it within (1) or without (0) an image.
  Returns a byte array with the results
  si          : stack database info
  fi_nx,fi_ny : image dimensions
  r           : alignment parameters, rotation and flipping must be corrected
}
function SImg_StackDB_inout(si:TSImg_SDBInfo; fi_nx,fi_ny:longint; r:TSImg_AlignmentParams):TSImg_StackDB_inout;
var xg,yg,xj,yj         : longint;
    y1,y2,x1,x2,a       : double;
    xc,yc,y0            : double;
    io                  : ^byte;
    rowbuf              : ^single;
begin
    getmem2(io,(si.xs+1)*(si.ys+1));
    getmem2(rowbuf,(si.xs+1)*sizeof(single)*2);
    result.b:=io;
    result.smin:=1e30;
    result.smax:=0;
    si.x0-=si.xso;
    si.y0-=si.yso;
    xc:=0; yc:=0;       // makes the compiler happy
    for yg:=0 to si.ys do
    begin
        y0:=yg shl SImg_StackDB_subgrid_m - si.y0;
        for xg:=0 to si.xs do
        begin
            x1:=xg shl SImg_StackDB_subgrid_m - si.x0;
            y1:=si.s21*x1 + si.s22*y0;   y2:=sqr(y1);
            x1:=si.s11*x1 + si.s12*y0;   x2:=sqr(x1);
            { alignment: coordinate transformation }
            a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
            x1+=x1*a + r.c3*x2 + r.c5*y2;
            y1+=y1*a + r.c4*y2 + r.c6*x2;
            x2:=r.b11*x1 + r.b12*y1 + r.ax - r.xmin;  xj:=round(x2);
            y2:=r.b21*x1 + r.b22*y1 + r.ay - r.ymin;  yj:=round(y2);
            if (xj<0) or (xj>=fi_nx) or (yj<0) or (yj>=fi_ny) then io^:=0
            else begin
                io^:=1;
                { scale }
                if (xg>0) and (io[-1]<>0) then
                begin
                    a:=sqr(x2-xc)+sqr(y2-yc);
                    if a<result.smin then result.smin:=a;
                    if a>result.smax then result.smax:=a;
                end;
                if (yg>0) and (io[-si.xs-1]<>0) then
                begin
                    a:=sqr(x2-rowbuf[xg*2])+sqr(y2-rowbuf[xg*2+1]);
                    if a<result.smin then result.smin:=a;
                    if a>result.smax then result.smax:=a;
                end;
            end;
            xc:=x2;  rowbuf[xg*2  ]:=x2;
            yc:=y2;  rowbuf[xg*2+1]:=y2;
            io+=1;
        end;
    end;
    result.smin:=sqrt(result.smin)/SImg_StackDB_subgrid_n;
    result.smax:=sqrt(result.smax)/SImg_StackDB_subgrid_n;
    freemem2(rowbuf,(si.xs+1)*sizeof(single)*2);
end;

{*******************************************************************************}
{********* SImg_StackDB_FWHMonSG ***********************************************}
{*******************************************************************************}
{ Calculate FWHM on sub-grid point.
   si    : stack database info
   fi    : FWHM map
   r     : Align parameter, must be corrected using SImg_CorrectAlignement
   x,y   : sub-grid position
   sx,sy : FWHM image scale factors: sx = fi.nx/(r^.xmax-r^.xmin+1) , ...
   fwhm  : FWHM output scaled to reference image
   result: true if within bounds
}
function SImg_StackDB_FWHMonSG(const si:TSImg_SDBInfo; const r:TSImg_AlignmentParams; const fi:TSIMG_FPImage; x,y:longint; sx,sy:single; var fwhm:single):boolean;
var a,x1,y1,x2,y2,x3,y3  : double;
begin
    result:=false;
    x1:=(2*x+1) shl (SImg_StackDB_subgrid_m-1) + si.xso - si.x0;
    y2:=(2*y+1) shl (SImg_StackDB_subgrid_m-1) + si.yso - si.y0;
    { sub-grid center }
    y1:=si.s21*x1 + si.s22*y2;   y3:=sqr(y1);
    x1:=si.s11*x1 + si.s12*y2;   x3:=sqr(x1);
    a:=r.d*(x3+y3) + r.c1*x1 + r.c2*y1;
    x2:=x1 + x1*a + r.c3*x3 + r.c5*y3;
    y2:=y1 + y1*a + r.c4*y3 + r.c6*x3;
    x3:=r.b11*x2 + r.b12*y2;
    y3:=r.b21*x2 + r.b22*y2;
    x:=round( (x3 + r.ax - r.xmin)*sx );
    y:=round( (y3 + r.ay - r.ymin)*sy );
    result:=(x>=-1) and (x<=fi.nx) and (y>=-1) and (y<fi.ny);
    if x<0 then x:=0;
    if x>=fi.nx then x:=fi.nx-1;
    if y<0 then y:=0;
    if y>=fi.ny then y:=fi.ny-1;
    { subgrid center + (1,1) for exact sale calculation }
    x1+=si.s11+si.s12;  x2:=sqr(x1);
    y1+=si.s21+si.s22;  y2:=sqr(y1);
    a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
    x1+=x1*a + r.c3*x2 + r.c5*y2;
    y1+=y1*a + r.c4*y2 + r.c6*x2;
    x2:=r.b11*x1 + r.b12*y1 - x3;
    y2:=r.b21*x1 + r.b22*y1 - y3;
    { scale fwhm }
    a:=sqrt(0.5*(sqr(x2)+sqr(y2)));
    if a<1e-6 then a:=1e-6;
    fwhm:=fi.pt[y*fi.sl+x]/a;
    result:=true;
end;

{*******************************************************************************}
{********* SImg_StackDB_add ****************************************************}
{*******************************************************************************}
{ add image to stack database
  ti    : thread index
  tn    : number of threads
  inout : in out data in corners of sub-grid
  si    : stack database info
  fs    : stack database data
  fi    : image to add
  fw    : pixel weights image, first color expected be 1/sqr(sigma)
  ffl   : FWHM limits
  ffm   : FWHM map
  r     : alignment parameters must be corrected using SImg_CorrectAlignement
}
threadprocedure SImg_StackDB_add_thread(ti:longint; tn:longint; inout:TSImg_StackDB_inout; si:TSImg_SDBInfo; fs:TSImg_FPImage; fi:TSImg_FPImage; fw:TSImg_FPImage; ffl:TSImg_FPImage; ffm:TSImg_FPImage; r:TSImg_AlignmentParams);
const sub_sampling_size = 0.3;  { in destination pixel width's }
var d                     : ^simgfp;
    io                    : ^byte;
    j,xi,yi,xj,yj,c,cz    : longint;
    xg,yg,xi0,xi1,yi0,yi1 : longint;
    xs,ys,ns              : longint;
    a,b,ff,ds,sx,sy       : double;
    x0,y0,x1,x2,y1,y2     : double;
    s,y4                  : single;
begin
    if inout.smax<inout.smin then exit;         // means no overlapping
    { supersampling }
{    a:=sqrt(sqr(r.b11)+sqr(r.b21));
    b:=sqrt(sqr(r.b12)+sqr(r.b22));
    if b>a then a:=b;
    if ti=0 then writeln(stderr,'sold=',a*si.s:0:4,'   smax=',inout.smax:0:4,'   smin=',inout.smin:0:4);}
//    if ti=0 then writeln(stderr,'smax=',inout.smax:0:4,'   smin=',inout.smin:0:4,'  s11=',si.s11:0:3,' s12=',si.s12:0:3,' s21=',si.s21:0:3,' s22=',si.s22:0:3);
    ns:=round(inout.smax/sub_sampling_size+0.5);
    ds:=1/ns;
    { factors, offsets}
    ff:=r.weight/sqr(ns*r.fact);
    x0:=si.x0 + 0.5 - 0.5*ds;
    y0:=si.y0 + 0.5 - 0.5*ds;
    { fwhm limits }
    if fs.nc<4 then ffl.pt:=nil;
    if ffl.pt=nil then ffm.pt:=nil;
    sx:=1; sy:=1;
    cz:=0;
    if ffm.pt<>nil then
    begin
        sx:=ffm.nx/fi.nx;
        sy:=ffm.ny/fi.ny;
        cz:=fs.nc div 2 - 1;
        if ffl.nc<cz then cz:=ffl.nc;
    end;

    for yg:=0 to si.ys-1 do if yg mod tn = ti then
    begin
        io:=inout.b+yg*(si.xs+1);
        for xg:=0 to si.xs-1 do if (io[xg]<>0) or (io[xg+1]<>0) or (io[xg+si.xs+1]<>0) or (io[xg+si.xs+2]<>0) then
        begin
            { FWHM range }
            c:=0;
            if ffm.pt<>nil then
            begin
                SImg_StackDB_FWHMonSG(si, r, ffm, xg,yg, sx,sy, s);
                d:=ffl.pt + yg*ffl.sl + xg*ffl.nc;
                while (c<cz) and (s>=d[c]) do
                    c+=1;
            end;
            { grid field coordinates }
            yi0:=yg shl SImg_StackDB_subgrid_m + si.yso;
            if yi0<0 then yi0:=0;
            xi0:=xg shl SImg_StackDB_subgrid_m + si.xso;
            if xi0<0 then xi0:=0;
            yi1:=(yg+1) shl SImg_StackDB_subgrid_m + si.yso;
            if yi1>fs.ny then yi1:=fs.ny;
            xi1:=(xg+1) shl SImg_StackDB_subgrid_m + si.xso;
            if xi1>fs.nx then xi1:=fs.nx;
            { grid field }
            for yi:=yi0 to yi1-1 do
            begin
                d:=fs.pt + yi*fs.sl + xi0*fs.nc + c*2;
                for xi:=xi0 to xi1-1 do
                begin
                    { supersampling }
                    for ys:=0 to ns-1 do
                    begin
                        y4:=yi - y0 + ys*ds;  { = yi - si.y0 - 0.5 + (ys+0.5)/ns }
                        for xs:=0 to ns-1 do
                        begin
                            { transformation of stack database coordinates ( xi-0.5+(xs+0.5)/ns , yi-0.5+(ys+0.5)/ns ) into reference image coordinates }
                            x1:=xi - x0 + xs*ds;  { = xi - si.x0 - 0.5 + (xs+0.5)/ns }
                            y1:=si.s21*x1 + si.s22*y4;   y2:=sqr(y1);
                            x1:=si.s11*x1 + si.s12*y4;   x2:=sqr(x1);
                            { alignment: coordinate transformation }
                            a:= r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
                            x1+=x1*a + r.c3*x2 + r.c5*y2;
                            y1+=y1*a + r.c4*y2 + r.c6*x2;
                            xj:=round(r.b11*x1 + r.b12*y1 + r.ax - r.xmin);
                            yj:=round(r.b21*x1 + r.b22*y1 + r.ay - r.ymin);
                            if (xj>=0) and (xj<fi.nx) and (yj>=0) and (yj<fi.ny) then
                            begin
                                j:=yj*fi.nx + xj;
                                a:=fi.pt[j]*r.fact;
                                b:=fw.pt[j]*ff;
//                              if a<-1000 then SImg_error('xg='+int2str(xg)+'  yg='+int2str(yg)+' xj='+int2str(xj)+' yj='+int2str(yj)+' fact='+real2str(r.fact));
                                d[0]+=a*b;
                                d[1]+=b;
                            end;
                        end;
                    end;
                    d+=fs.nc;
                end;
            end;
        end;
    end;
end;

{ add image to stack database
  si    : stack database info
  fs    : stack database
  fi    : image to add
  fw    : weight image, expected to be 1/sqr(sigma)
  ffl   : FWHM limits
  ffm   : FWHM map
  res   : alignment parameters
}
procedure SImg_StackDB_add(cc:CSImg_Config; si:TSImg_SDBInfo; fs,fi,fw,ffl,ffm:TSImg_FPImage; res:TSImg_AlignmentParams);
var t,tn  : longint;
    inout : TSImg_StackDB_inout;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_add: fs='+SImg_InfoStr2(fs)+' fi='+SImg_InfoStr2(fi)+'  fw='+SImg_InfoStr2(fw)+'  ffl='+SImg_InfoStr2(ffl)+'  ffm='+SImg_InfoStr2(ffm));
    if (not SImg_Recover(cc,fs)) or (fs.nc and 1<>0) then cc.error('SImg_StackDB_add: Illegal stack database image');
    if not SImg_Recover(cc,fi) then cc.error('SImg_StackDB_add: Illegal input image');
    if fi.nc<>1 then cc.error('SImg_StackDB_add: Only one color images are supported');
    if not SImg_Recover(cc,fw) then cc.error('SImg_StackDB_add: Illegal weight image');
    if (fw.pt<>nil) and ((fw.nx<>fi.nx) or (fw.ny<>fi.ny) or (fw.nc<>1)) then cc.error('SImg_StackDB_add: Invalid dimensions of weight image');

    if (ffl.pt<>nil) and (not SImg_Recover(cc,ffl)) then
    begin
        cc.warning('SImg_StackDB_add: Illegal FWHM limit image, ignoring it');
        ffm.pt:=nil;
    end;
    if (ffl.pt<>nil) and ((ffl.nx<>si.xs) or (ffl.ny<>si.ys)) then
    begin
        cc.warning('SImg_StackDB_add: Invalid dimensions of FWHM limit image. ignoring it');
        ffl.pt:=nil;
    end;
    if (ffl.pt<>nil) and (ffl.nc*2+2<>fs.nc) then cc.warning('SImg_StackDB_add: Number of colors of FWHM limit images does not match to number of FWHM ranges of stack database');
    if ffl.pt=nil then ffm.pt:=nil;

    if (ffm.pt<>nil) and ((not SImg_Recover(cc,ffm)) or (ffm.nc<>1)) then
    begin
        cc.warning('SImg_StackDB_add: Illegal FWHM map, ignoring it');
        ffm.pt:=nil;
    end;
    if ffm.pt=nil then ffl.pt:=nil;

    inout:=SImg_StackDB_inout(si, fi.nx,fi.ny, res);
    if inout.smax<inout.smin then
    begin
        cc.warning('SImg_StackDB_add: No overlapping between source and destination.');
        freemem2(inout.b);
        exit;
    end;

    tn:=cc.threads;
    for t:=0 to tn-1 do
        dothread SImg_StackDB_add_thread(t,tn, inout, si, fs,fi,fw,ffl,ffm, res);
    threadwait;

    freemem2(inout.b);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_shift **************************************************}
{*******************************************************************************}
{ shift referee image to destination
  ti    : thread index
  tn    : number of threads
  inout : in/out data in corners of sub-grid
  si    : stack database info
  fr    : reference image
  fi    : destination image
  fw    : weight buffer
  r     : alignment parameters, must be corrected using SImg_CorrectAlignement
}
threadprocedure SImg_StackDB_shift_thread(ti:longint; tn:longint; inout:TSImg_StackDB_inout; si:TSImg_SDBInfo; fr:TSImg_FPImage; fi:TSImg_FPImage; wi:pbyte; r:TSImg_AlignmentParams);
const sub_sampling_size = 0.6;  { in destination pixel width's }
var d,e                   : ^simgfp;
    io                    : ^byte;
    j,k,xi,yi,xj,yj,c     : longint;
    xg,yg,xi0,xi1,yi0,yi1 : longint;
    xs,ys,ns              : longint;
    a,ds                  : double;
    x0,y0,x1,x2,y1,y2     : double;
    y4                    : single;
begin
    if inout.smax<inout.smin then exit;         // means no overlapping
    { supersampling }
    ns:=round(inout.smax/sub_sampling_size+0.5);
    ds:=1/ns;
    { factors, offsets}
    x0:=si.x0 + 0.5 - 0.5*ds;
    y0:=si.y0 + 0.5 - 0.5*ds;

    for yg:=0 to si.ys-1 do if yg mod tn = ti then
    begin
        io:=inout.b+yg*(si.xs+1);
        for xg:=0 to si.xs-1 do if (io[xg]<>0) or (io[xg+1]<>0) or (io[xg+si.xs+1]<>0) or (io[xg+si.xs+2]<>0) then
        begin
            { grid field coordinates }
            yi0:=yg shl SImg_StackDB_subgrid_m + si.yso;
            if yi0<0 then yi0:=0;
            xi0:=xg shl SImg_StackDB_subgrid_m + si.xso;
            if xi0<0 then xi0:=0;
            yi1:=(yg+1) shl SImg_StackDB_subgrid_m + si.yso;
            if yi1>fr.ny then yi1:=fr.ny;
            xi1:=(xg+1) shl SImg_StackDB_subgrid_m + si.xso;
            if xi1>fr.nx then xi1:=fr.nx;
            { grid field }
            for yi:=yi0 to yi1-1 do
            begin
                d:=fr.pt + yi*fr.sl + xi0*fr.nc;
                for xi:=xi0 to xi1-1 do
                begin
                    { supersampling }
                    for ys:=0 to ns-1 do
                    begin
                        y4:=yi - y0 + ys*ds;  { = yi - si.y0 - 0.5 + (ys*0.5)/ns }
                        for xs:=0 to ns-1 do
                        begin
                            { transformation of stack database coordinates ( xi-0.5+(xs+0.5)/ns , yi-0.5+(ys+0.5)/ns ) into reference image coordinates }
                            x1:=xi - x0 + xs*ds; { = xi - si.x0 - 0.5 + (xs*0.5)/ns }
                            y1:=si.s21*x1 + si.s22*y4;   y2:=sqr(y1);
                            x1:=si.s11*x1 + si.s12*y4;   x2:=sqr(x1);
                            { reference coordinates to destination coordinates }
                            a:=r.d*(x2+y2) + r.c1*x1 + r.c2*y1;
                            x1+=x1*a + r.c3*x2 + r.c5*y2;
                            y1+=y1*a + r.c4*y2 + r.c6*x2;
                            xj:=round(r.b11*x1 + r.b12*y1 + r.ax - r.xmin);
                            yj:=round(r.b21*x1 + r.b22*y1 + r.ay - r.ymin);
                            if (xj>=0) and (xj<fi.nx) and (yj>=0) and (yj<fi.ny) then
                            begin
                                j:=yj*fi.nx + xj;
                                e:=fi.pt + j*fi.nc;
                                k:=wi[j];
                                if k=0 then
                                    for c:=0 to fi.nc-1 do
                                        e[c]:=d[c]
                                else for c:=0 to fi.nc-1 do
                                        e[c]+=d[c];
                                wi[j]:=k+1;
                            end;
                        end;
                    end;
                    d+=fr.nc;
                end;
            end;
        end;
    end;
end;

{ shift reference image to destination
  si    : stack dtatabse info
  fr    : reference image
  fi    : Destination image. Only defined pixels are written.
  fw    : Weight image. Optional. If given, undefined pixels are masked out.
  res   : alignment parameters
}
function SImg_StackDB_shift(cc:CSImg_Config; si:TSImg_SDBInfo; fr,fi,fw:TSImg_FPImage; res:TSImg_AlignmentParams):boolean;
var t,tn,c,n : longint;
    wi       : ^byte;
    r        : double;
    inout    : TSImg_StackDB_inout;
    d        : ^simgfp;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    result:=false;
    if res.fact<1e-30 then
    begin
        cc.warning('SImg_StackDB_shift: Brightness scaling to small, destination undefined');
        exit;
    end;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_shift: fr='+SImg_InfoStr2(fr)+' fi='+SImg_InfoStr2(fi));
    if (not SImg_Recover(cc,fr)) then cc.error('SImg_StackDB_shift: Invalid reference image');
    if (not SImg_Recover(cc,fi)) or (fi.nc<>fr.nc) then cc.error('SImg_StackDB_shift: Invalid destination image');
    if (fw.pt<>nil) and (not SImg_Recover(cc,fw)) then
    begin
        cc.warning('SImg_StackDB_shift: Invalid pixel weight image: Ignored.');
        fw.pt:=nil;
    end;
    if (fw.pt<>nil) and (fw.nx<>fi.nx) and (fw.ny<>fi.ny) and (fw.nc<>1) then
    begin
        cc.warning('SImg_StackDB_shift: Invalid dimensions of pixel weight image: Ignored.');
        fw.pt:=nil;
    end;

    inout:=SImg_StackDB_inout(si, fi.nx,fi.ny, res);
    if inout.smax<inout.smin then
    begin
        cc.warning('SImg_StackDB_add: No overlapping between source and destination.');
        freemem2(inout.b);
        exit;
    end;

    n:=fi.nx*fi.ny;
    getmem2(wi,n);
    fillchar(wi^, n, 0);

    tn:=cc.threads;
    for t:=0 to tn-1 do
        dothread SImg_StackDB_shift_thread(t,tn, inout, si, fr,fi,wi, res);
    threadwait;

    d:=fi.pt;
    for t:=0 to n-1 do
    begin
        tn:=wi[t];
        if tn>0 then
        begin
            r:=1/(tn*res.fact);
            for c:=0 to fi.nc-1 do
                d[c]*=r;
        end
        else if fw.pt<>nil then fw.pt[t]:=0;
        d+=fi.nc;
    end;
    result:=true;

    freemem2(wi,n);
    freemem2(inout.b);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_Correct ************************************************}
{*******************************************************************************}
{ correct value biases bewteen fwhm range }
procedure SImg_StackDB_Correct(cc:CSImg_Config; fs:TSImg_FPImage);
var a                   : array of double;
    d                   : ^simgfp;
    r,s,t,u,v           : double;
    xg,yg,xi,yi,c,nr,i  : longint;
    xi0,yi0,xi1,yi1     : longint;
    xs,ys,xso,yso       : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if fs.nc<4 then exit;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_Correct: fs='+SImg_InfoStr2(fs));
    if not SImg_Recover(cc,fs) or ((fs.nc and 1)<>0) then cc.error('SImg_StackDB_Correct: Illegal stack database');

    setLength(a,fs.nc);
    nr:=fs.nc div 2;

    xs:=(fs.nx-1) shr SImg_StackDB_subgrid_m + 1;
    ys:=(fs.ny-1) shr SImg_StackDB_subgrid_m + 1;
    xso:=(fs.nx - xs shl SImg_StackDB_subgrid_m) div 2;
    yso:=(fs.ny - ys shl SImg_StackDB_subgrid_m) div 2;

    d:=fs.pt;
    t:=0; u:=0; v:=0;
    for i:=0 to fs.nx*fs.ny-1 do
    begin
        r:=0;  s:=0;
        for c:=0 to nr-1 do
        begin
            r+=d[c*2];
            s+=d[c*2+1];
        end;
        d+=fs.nc;
        if s<1e-30 then continue;
        t+=r;
        u+=sqr(r)/s;
        v+=s;
    end;
    if v>1e-30 then v:=1/v;
    t*=v;
    u:=u*v-sqr(t);
    if u<0 then u:=0;
//    writeln(stderr,'avg='+real2str(t,6)+'  var='+real2str(sqrt(u),6));
    t+=sqrt(u)*2;

    for yg:=0 to ys-1 do
        for xg:=0 to xs-1 do
        begin
            { grid field coordinates }
            yi0:=yg shl SImg_StackDB_subgrid_m + yso;
            if yi0<0 then yi0:=0;
            xi0:=xg shl SImg_StackDB_subgrid_m + xso;
            if xi0<0 then xi0:=0;
            yi1:=(yg+1) shl SImg_StackDB_subgrid_m + yso;
            if yi1>fs.ny then yi1:=fs.ny;
            xi1:=(xg+1) shl SImg_StackDB_subgrid_m + xso;
            if xi1>fs.nx then xi1:=fs.nx;

            for c:=0 to fs.nc-1 do
                a[c]:=0;
            for yi:=yi0 to yi1-1 do
            begin
                d:=fs.pt + yi*fs.sl + xi0*fs.nc;
                for xi:=xi0 to xi1-1 do
                begin
                    for c:=0 to nr-1 do
                    begin
                        r:=d[c*2  ];
                        s:=d[c*2+1];
                        if r>t*s then r:=t*s;
                        a[c*2  ]+=r;
                        a[c*2+1]+=s;
                    end;
                    d+=fs.nc;
                end;
            end;

            r:=0; s:=0;
            for c:=0 to nr-1 do
            begin
                r+=a[c*2  ];
                s+=a[c*2+1];
            end;
            if s<1e-30 then continue;
            r/=s;

            for c:=0 to nr-1 do
            begin
                s:=a[c*2+1];
                if s>1e-30 then a[c*2]:=a[c*2]/s-r
                else a[c*2]:=0;
            end;

            for yi:=yi0 to yi1-1 do
            begin
                d:=fs.pt + yi*fs.sl + xi0*fs.nc;
                for xi:=xi0 to xi1-1 do
                begin
                    for c:=0 to nr-1 do
                        d[c*2]-=a[c*2]*d[c*2+1];
                    d+=fs.nc;
                end;
            end;
        end;
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_Convert ************************************************}
{*******************************************************************************}
{ convert resolution optimized values and weight
  fi : output image
  fw : weight = 1/sqr(sigma)
  snr: maximum snr, if <1 interpreted as quantile, <=0 to disable resolution optimization
}
procedure SImg_StackDB_Convert(cc:CSImg_Config; fs:TSImg_FPImage; var fi,fw:TSImg_FPImage; snr:double);
const reg = 1;  // signal regularization for snr calculation
var fv,fn       : TSImg_FPImage;
    mask        : TSImg_RDImage;
    r,s,t,u,a,b : double;
    i,c,nr      : longint;
    d           : ^simgfp;
    ir          : boolean;
label lInterpolation;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_convert: fs='+SImg_InfoStr2(fs)+' fi='+SImg_InfoStr2(fi)+' fw='+SImg_InfoStr2(fw)+' snr='+real2str(snr));
    if not SImg_Recover(cc,fs) or ((fs.nc and 1)<>0) then cc.error('SImg_StackDB_Convert: Illegal stack database');
    if not SImg_OverWriteAlert(fs,fi) then SImg_Free(cc,fi);
    if not SImg_OverWriteAlert(fs,fw) then SImg_Free(cc,fw);

    { total values, weight, weights average/variance }
    fn.pt:=nil; fv.pt:=nil;
    SImg_New(cc, fn, fs.nx,fs.ny, 1);
    SImg_New(cc, fv, fs.nx,fs.ny, 1);
    SImg_New(cc, mask, SImg_RawByte, fv.nx, fv.ny, 1, true);
    nr:=fs.nc div 2; { number of ranges }
    ir:=false;

    { only two colors }
    if (fs.nc=2) or (snr<simg_mina) then
    begin
        SImg_New(cc, fn, fs.nx,fs.ny, 1);
        SImg_New(cc, fv, fs.nx,fs.ny, 1);
        d:=fs.pt;
        for i:=0 to fv.size-1 do
        begin
            r:=0; s:=0;
            for c:=0 to nr-1 do
            begin
                r+=d[2*c];
                s+=d[2*c+1];
            end;
            d+=fs.nc;
            if s>1e-30 then fv.pt[i]:=r/s
            else begin
                fv.pt[i]:=0;
                mask.ptb[i]:=1;
                ir:=true;
            end;
            fn.pt[i]:=s;
        end;
        goto lInterpolation;
    end;

    { total values, weight, weighted average/variance }
    d:=fs.pt;
    a:=0; b:=0; u:=0;
    for i:=0 to fn.size-1 do
    begin
        r:=0;  s:=0;
        for c:=0 to nr-1 do
        begin
            r+=d[c*2];
            s+=d[c*2+1];
        end;
        d+=fs.nc;
        if s<simg_mina then continue;
        t:=1/s;
        a+=r;
        b+=sqr(r)*t;
        u+=s;
        fv.pt[i]:=r*t;
        fn.pt[i]:=s;
    end;
    if u>1e-30 then u:=1/u;
    a*=u;
    b:=b*u-sqr(a);
    if b<0 then b:=0;
    b:=sqrt(b)*reg;
//    writeln(stderr,'avg='+real2str(a,6)+'  reg='+real2str(b,6));

    { fwhm, snr }
    t:=SImg_GetSharpnessFWHM(cc, fv);
    cc.info('sdbconvert: Old FWHM='+real2str(t,-1));
    for i:=0 to fn.size-1 do
    begin
        r:=fv.pt[i]-a;
        if r<0 then r:=0;
        r+=b;
        fv.pt[i]:=r*sqrt(fn.pt[i]);     // regularized SNR
    end;
    if snr<1.0000000001 then snr:=SImg_GetQuantile(cc, fv, abs(snr) );
//ifdef[sdb_debug]
//    SImg_WriteImageFile(cc,'sdbconvert-snr.fits',fv);
//endif

    { reduction mask }
    for i:=0 to fn.size-1 do
    begin
        r:=fv.pt[i];
        if r<snr then r:=1
        else begin
            r:=sqrt(snr/r);
        end;
        fv.pt[i]:=r;
    end;
//ifdef[sdb_debug]
//    SImg_WriteImageFile(cc,'sdbconvert-mask0.fits',fv);
//endif
    SImg_FBGFilter(cc, fv,fv,t,true);
//ifdef[sdb_debug]
    SImg_WriteImageFile(cc,'sdbconvert-mask.fits',fv);
//endif

    { combine everything }
    d:=fs.pt+2;
    for i:=0 to fn.size-1 do
    begin
        r:=d[-2];
        s:=d[-1];
        t:=fn.pt[i]*sqr(sqr(fv.pt[i]));
        if s<t then
        begin
            for c:=0 to nr-2 do
            begin
                u:=d[c*2+1];
                if (s+u>t*0.99999) and (u>1e-30) then
                begin
                    r+=d[c*2]*(t-s)/u;
                    s:=t;
                    break;
                end;
                r+=d[c*2];
                s+=u;
            end;
        end;
        d+=fs.nc;
        if s>1e-30 then r/=s
        else begin
            r:=0;
            mask.ptb[i]:=1;
            ir:=true;
        end;
        fv.pt[i]:=r;
        fn.pt[i]:=s;
    end;

lInterpolation:
    if ir then
    begin
        cc.warning('Interpolation required');
        SImg_InterpolateMasked(cc, fv, mask, 1);
//ifdef[sdb_debug]
        SImg_WriteImageFile(cc,'convertVal-mask.fits',mask);
//endif
    end;
    SImg_Free(cc,mask);


    SImg_Move(cc, fv,fi);
    SImg_Move(cc, fn,fw);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SI2mg_StackDB_ConvertVal *********************************************}
{*******************************************************************************}
{ convert weighted averages for all weights }
procedure SImg_StackDB_ConvertVal(cc:CSImg_Config; fs:TSImg_FPImage; var fi:TSImg_FPImage; sc:boolean);
var fj     : TSImg_FPImage;
    mask   : TSImg_RDImage;
    r,s    : double;
    i,c,cz : longint;
    d      : ^simgfp;
    b      : boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_convertVal: fs='+SImg_InfoStr2(fs)+' fi='+SImg_InfoStr2(fi)+' sc='+boolean2str(sc));
    if not SImg_Recover(cc,fs) or ((fs.nc and 1)<>0) then cc.error('SImg_StackDB_ConvertVal: Illegal stack database');
    if not SImg_OverWriteAlert(fs,fi) then SImg_Free(cc,fi);

    mask.pt:=nil;
    fj.pt:=nil;
    b:=false;
    if sc or (fs.nc=2) then
    begin
        SImg_New(cc,fj, fs.nx, fs.ny, fs.nc>>1);
        SImg_New(cc, mask, SImg_RawByte, fj.nx, fj.ny, fj.nc, true);
        for i:=0 to fj.size-1 do
        begin
            r:=fs.pt[i*2+1];
            if r>simg_mina
            then fj.pt[i]:=fs.pt[i*2]/r
            else begin
                fj.pt[i]:=0;
                mask.ptb[i]:=1;
                b:=true;
            end;
        end;
    end else
    begin
        SImg_New(cc,fj, fs.nx, fs.ny, 1);
        SImg_New(cc, mask, SImg_RawByte, fj.nx, fj.ny, 1, true);
        cz:=fs.nc div 2-1;
        d:=fs.pt;
        for i:=0 to fj.nx*fj.ny-1 do
        begin
            r:=0;  s:=0;
            for c:=0 to cz do
            begin
                s+=d[c*2];
                r+=d[c*2+1];
            end;
            d+=fs.nc;
//          if r<simg_mina then cc.info('s='+real2str(s)+'  r='+real2str(r));
            if r>simg_mina
            then fj.pt[i]:=s/r
            else begin
                fj.pt[i]:=0;
                mask.ptb[i]:=1;
                b:=true;
            end;
        end;
    end;

    if b then
    begin
        cc.warning('Interpolation required');
        SImg_InterpolateMasked(cc, fj, mask, 1);
//ifdef[sdb_debug]
        SImg_WriteImageFile(cc,'convertVal-mask.fits',mask);
//endif
    end;
    SImg_Free(cc,mask);

    SImg_Move(cc, fj,fi);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SI2mg_StackDB_ConvertValWeight **************************************}
{*******************************************************************************}
{ convert weighted averages and weight, do not interpolate }
procedure SImg_StackDB_ConvertValWeight(cc:CSImg_Config; fs:TSImg_FPImage; var fi,fw:TSImg_FPImage);
var fj,fk       : TSImg_FPImage;
    r,s         : double;
    i,c,cz      : longint;
    d           : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_convertValWeight: fs='+SImg_InfoStr2(fs)+' fi='+SImg_InfoStr2(fi)+' fw='+SImg_InfoStr2(fw));
    if not SImg_Recover(cc,fs) or ((fs.nc and 1)<>0) then cc.error('SImg_StackDB_ConvertVal: Illegal stack database');
    if not SImg_OverWriteAlert(fs,fi) then SImg_Free(cc,fi);
    if not SImg_OverWriteAlert(fs,fw) then SImg_Free(cc,fw);

    fj.pt:=nil; fk.pt:=nil;
    SImg_New(cc,fj, fs.nx, fs.ny, 1);
    SImg_New(cc,fk, fs.nx, fs.ny, 1);
    cz:=fs.nc div 2-1;
    d:=fs.pt;
    for i:=0 to fj.nx*fj.ny-1 do
    begin
        r:=0;  s:=0;
        for c:=0 to cz do
        begin
            s+=d[c*2];
            r+=d[c*2+1];
        end;
        d+=fs.nc;
        if r>simg_mina
        then fj.pt[i]:=s/r
        else fj.pt[i]:=0;
        fk.pt[i]:=r;
    end;

    SImg_Move(cc, fj,fi);
    SImg_Move(cc, fk,fw);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_ConvertSNR *********************************************}
{*******************************************************************************}
{ convert SNR for all resolutions }
procedure SImg_StackDB_ConvertSNR(cc:CSImg_Config; fs:TSImg_FPImage; var fi:TSImg_FPImage; sc:boolean);
var fj     : TSImg_FPImage;
    r,s    : double;
    i,c,cz : longint;
    d      : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_convertSNR: fs='+SImg_InfoStr2(fs)+' fi='+SImg_InfoStr2(fi)+' s='+boolean2str(sc));
    if not SImg_Recover(cc,fs) or ((fs.nc and 1)<>0) then cc.error('SImg_StackDB_ConvertSNR: Illegal stack database');
    if not SImg_OverWriteAlert(fs,fi) then SImg_Free(cc,fi);

    fj.pt:=nil;
    if sc or (fs.nc=2) then
    begin
        SImg_New(cc,fj, fs.nx, fs.ny, fs.nc>>1);
        for i:=0 to fj.size-1 do
        begin
            r:=fs.pt[i*2+1];
            if r>1e-30  then fj.pt[i]:=fs.pt[i*2]/sqrt(r)
            else fj.pt[i]:=0;
        end;
    end else
    begin
        SImg_New(cc,fj, fs.nx, fs.ny, 1);
        cz:=fs.nc div 2-1;
        d:=fs.pt;
        for i:=0 to fj.nx*fj.ny-1 do
        begin
            r:=0;  s:=0;
            for c:=0 to cz do
            begin
                s+=d[c*2];
                r+=d[c*2+1];
            end;
            d+=fs.nc;
            if r>1e-30 then fj.pt[i]:=s/sqrt(r)
            else fj.pt[i]:=0;
        end;
    end;
    SImg_Move(cc, fj,fi);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_ConvertWeight ******************************************}
{*******************************************************************************}
{ convert all weights }
procedure SImg_StackDB_ConvertWeight(cc:CSImg_Config; fs:TSImg_FPImage; var fi:TSImg_FPImage; sc:boolean);
var fj     : TSImg_FPImage;
    r      : double;
    i,c,cz : longint;
    d      : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_convertWeight: fs='+SImg_InfoStr2(fs)+' fi='+SImg_InfoStr2(fi)+' s='+boolean2str(sc));
    if not SImg_Recover(cc,fs) or ((fs.nc and 1)<>0) then cc.error('SImg_StackDB_ConvertWeight: Illegal stack database');
    if not SImg_OverWriteAlert(fs,fi) then SImg_Free(cc,fi);

    fj.pt:=nil;
    if sc or (fs.nc=2) then
    begin
        SImg_New(cc,fj, fs.nx, fs.ny, fs.nc>>1);
        for i:=0 to fj.size-1 do
            fj.pt[i]:=fs.pt[i*2+1];
    end else
    begin
        SImg_New(cc,fj, fs.nx, fs.ny, 1);
        cz:=fs.nc div 2-1;
        d:=fs.pt;
        for i:=0 to fj.nx*fj.ny-1 do
        begin
            r:=0;
            for c:=0 to cz do
                r+=d[c*2+1];
            d+=fs.nc;
            fj.pt[i]:=r;
        end;
    end;
    SImg_Move(cc, fj,fi);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_CropInfo ***********************************************}
{*******************************************************************************}
{ Calaculate a crop region from weight
    thr : threshold. Recommended: 0.1..0.9
}
procedure SImg_StackDB_CropInfo(cc:CSImg_Config; fs:TSImg_FPImage; thr:double; var x1,y1,x2,y2:longint);
var wx,wy  : ^double;
    r,s    : double;
    i,j    : longint;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_cropInfo: fs='+SImg_InfoStr2(fs)+' thr='+real2str(thr));
    if not SImg_Recover(cc,fs) or (fs.nc<2) then cc.error('SImg_StackDB_cropInfo: Illegal stack database');

    { row and column weight sums }
    getmem2(wx,fs.nx*sizeof(double));
    for i:=0 to fs.nx-1 do
        wx[i]:=0;
    getmem2(wy,fs.ny*sizeof(double));
    for i:=0 to fs.ny-1 do
        wy[i]:=0;
    for i:=0 to fs.ny-1 do
    begin
        s:=0;
        for j:=0 to fs.nx-1 do
        begin
            r:=fs.pt[1];
            wx[j]+=r;
            s+=r;
            fs.pt+=fs.nc;
        end;
        wy[i]:=s;
    end;

    { x range }
    s:=wx[0]; j:=0;
    for i:=1 to fs.nx-1 do
        if wx[i]>s then
        begin
            s:=wx[i];
            j:=i;
        end;
    i:=j+1;
    j-=1;
    s*=thr;
    while (j>=0) and (wx[j]>s) do
        j-=1;
    while (i<fs.nx) and (wx[i]>s) do
        i+=1;
    x1:=j+1;
    x2:=i-1;

    { y range }
    s:=wy[0]; j:=0;
    for i:=1 to fs.ny-1 do
        if wy[i]>s then
        begin
            s:=wy[i];
            j:=i;
        end;
    i:=j+1;
    j-=1;
    s*=thr;
    while (j>=0) and (wy[j]>s) do
        j-=1;
    while (i<fs.ny) and (wy[i]>s) do
        i+=1;
    y1:=j+1;
    y2:=i-1;

    freemem2(wx);
    freemem2(wy);

    SImg_DecDebugLevel;
end;



{*******************************************************************************}
{********* SImg_StackDB_FWHMLimits *********************************************}
{*******************************************************************************}
const SImg_StackDB_FWHM_histSize = 64;

threadprocedure SImg_StackDB_FWHMLimits_thread(minb:psimgfp; maxb:psimgfp; hist:pword; fi:TSImg_FPImage; t:longint; tn:longint);
const fwhm_offs = 0.25;   { offset at FWHM=1 }
var a,b,rn      : single;
    i,j,k,l,m,n,d       : longint;
    x,y,ya,yz           : longint;
    w                   : ^word;
    e                   : ^simgfp;
begin
    n:=fi.nc+1;
    rn:=1/n;
    ya:=fi.ny*t div tn;
    yz:=fi.ny*(t+1) div tn-1;
    for y:=ya to yz do
    begin
        d:=y*fi.nx;
        e:=fi.pt + y*fi.sl;
        for x:=0 to fi.nx-1 do
        begin
            a:=minb[d];
            b:=1/maxb[d];
            w:=hist + d*SImg_StackDB_FWHM_histSize;
            d+=1;
            k:=0;
            for i:=0 to SImg_StackDB_FWHM_histSize-1 do
                k+=w[i];
            l:=0; i:=0;
            for j:=1 to n-1 do
            begin
                m:=round(sqr(j*rn)*k);
                while (i<SImg_StackDB_FWHM_histSize) and (l+w[i]<m) do
                begin
                    l+=w[i];
                    i+=1;
                end;
                e^:=(i*b+a)*(1+0.5*fwhm_offs) + 0.5*fwhm_offs;
                e+=1;
            end;
        end;
    end;
end;


procedure SImg_StackDB_FWHMLimits(cc:CSImg_Config; si:TSImg_SDBInfo; const fns:array of ansistring; const res:array of TSImg_AlignmentParams; var fj:TSImg_FPImage; n:longint);
var fi                  : TSImg_FPImage;
    minb,maxb,d0,d1     : ^simgfp;
    a,mina,maxa         : simgfp;
    sx,sy               : single;
    i,j,x,y,len         : longint;
    r                   : TSImg_AlignmentParams;
    hist,e              : ^word;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_FWHMLimits: called with: out='+SImg_Infostr2(fj)+' n='+int2str(n)+' fns.length='+int2str(high(fns))+' aps.length='+int2str(high(res)));
    if high(fns)<>high(res) then cc.error('SImg_StackDB_FWHMLimits: length of fns and aps must be equal');
    len:=high(fns)+1;
    SImg_Free(cc,fj);

    si.x0-=si.xso + 0.5*SImg_StackDB_subgrid_n;
    si.y0-=si.yso + 0.5*SImg_StackDB_subgrid_n;

    getmem2(minb, si.xs*si.ys*sizeof(simgfp));
    getmem2(maxb, si.xs*si.ys*sizeof(simgfp));
    for i:=0 to si.xs*si.ys-1 do
    begin
        minb[i]:= simg_large;
        maxb[i]:=-simg_large;
    end;

    cc.info(cc.name+': Pass 1 -- finding minima/maxima');
    fi.pt:=nil;
    mina:= simg_large;
    maxa:=-simg_large;
    for i:=0 to len-1 do
    begin
        r:=res[i];
        SImg_ReadImageFile(cc,fns[i],fi);
        if fi.nc<>1 then cc.error(cc.name+': FWHM map ''' + fns[i] + ''' contains more than one color');
        sx:=fi.nx/(r.xmax-r.xmin+1);
        sy:=fi.ny/(r.ymax-r.ymin+1);

        d0:=minb;
        d1:=maxb;
        for y:=0 to si.ys-1 do
            for x:=0 to si.xs-1 do
            begin
                if SImg_StackDB_FWHMonSG(si, r, fi, x,y, sx,sy, a) then
                begin
                    if a<d0^ then d0^:=a;
                    if a>d1^ then d1^:=a;
                    if a<mina then mina:=a;
                    if a>maxa then maxa:=a;
                end;
                d0+=1;
                d1+=1;
            end;
    end;
//ifdef[sdb_debug]
    cc.info('mina='+real2str(mina)+'  maxa='+real2str(maxa));
//endif

    cc.info(cc.name+': Pass 2 -- calculating histogram');
    for i:=0 to si.xs*si.ys-1 do
    begin
        if mina<minb[i] then minb[i]:=mina;
        if maxa>maxb[i] then maxb[i]:=maxa;
        a:=maxb[i]-minb[i];
        if a<1e-20 then a:=1e-20;
        maxb[i]:=SImg_StackDB_FWHM_histSize/a;
    end;
    getmem2(hist, si.xs*si.ys*sizeof(word)*SImg_StackDB_FWHM_histSize);
    fillchar(hist^, si.xs*si.ys*sizeof(word)*SImg_StackDB_FWHM_histSize, 0);

    for i:=0 to len-1 do
    begin
        r:=res[i];
        SImg_ReadImageFile(cc,fns[i],fi);
        sx:=fi.nx/(r.xmax-r.xmin+1);
        sy:=fi.ny/(r.ymax-r.ymin+1);
        d0:=minb;
        d1:=maxb;
        e:=hist;
        for y:=0 to si.ys-1 do
            for x:=0 to si.xs-1 do
            begin
                if SImg_StackDB_FWHMonSG(si, r, fi, x,y, sx,sy, a) then
                begin
                    j:=round((a-d0^)*d1^-0.5);
                    if j<0 then j:=0;
                    if j>=SImg_StackDB_FWHM_histSize then j:=SImg_StackDB_FWHM_histSize-1;
                    e[j]+=1;
                end;
                d0+=1;  d1+=1;
                e+=SImg_StackDB_FWHM_histSize;
            end;
    end;

    cc.info(cc.name+': Pass 3 -- calculating limits using '+int2str(cc.threads)+' threads');
    SImg_New(cc,fi, si.xs, si.ys, n-1);
    j:=cc.threads;
    for i:=0 to j-1 do
        dothread SImg_StackDB_FWHMLimits_thread(minb, maxb, hist, fi, i,j);
    threadwait;

    freemem2(hist, si.xs*si.ys*sizeof(word)*SImg_StackDB_FWHM_histSize);
    freemem2(minb, si.xs*si.ys*sizeof(simgfp));
    freemem2(maxb, si.xs*si.ys*sizeof(simgfp));
    SImg_Move(cc, fi,fj);
    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_UnderweightBoundaries ******************************************}
{*******************************************************************************}
{ Underweight image boundaries
    fw: input and output
    wb: underweight size relative to half image size (0..1)
}
procedure SImg_UnderweightBoundaries(cc:CSImg_Config; fw:TSImg_FPImage; wb:double);
var x,y,c,j,k  : longint;
    r,s        : double;
    d          : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if wb<1e-4 then
    begin
        cc.warning('SImg_UnderweightBoundaries: Nothing to do');
        SImg_DecDebugLevel;
        exit;
    end;
    SImg_IncDebugLevel;
    cc.debug('SImg_UnderweightBoundaries: called with: fw='+SImg_InfoStr2(fw)+' wb='+real2str(wb));
    if not SImg_Recover(cc,fw) then cc.error('SImg_UnderweightBoundaries: Invalid image');

    j:=fw.ny;
    if fw.nx<j then j:=fw.nx;
    r:=(j+1)*0.5*wb;
    j:=round(r-0.5);
    r:=1/r;
    for y:=0 to fw.ny-1 do
    begin
        x:=fw.ny-y;
        if x<=y then k:=x
        else k:=y+1;
        d:=fw.pt + y*fw.sl;
        for x:=1 to j do
        begin
            if x<k then s:=x*r
            else s:=k*r;
            for c:=0 to fw.nc-1 do
                d[c]*=s;
            d+=fw.nc;
        end;
        if k<=j then
        begin
            s:=k*r;
            for x:=0 to (fw.nx-2*j)*fw.nc-1 do
                d[x]*=s;
        end;
        d:=fw.pt + y*fw.sl + fw.nx-j;
        for x:=j downto 1  do
        begin
            if x<k then s:=x*r
            else s:=k*r;
            for c:=0 to fw.nc-1 do
                d[c]*=s;
            d+=fw.nc;
        end;
    end;

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_unmask *************************************************}
{*******************************************************************************}
{ ti : thread index
  tn : number of threads
  fi : background corrected image (input and output)
  fr : residual input
  fsn: square noise
  ff : weight factor output
  dt : unmask threshold relative to variance
  dr:  absolute value is unmask threshold relative to reference. If <0 lower threshold is sqrt(1+sqr(dl)).
  adr: ADC resolution in ADU/e
}
threadprocedure SImg_StackDB_unmask_thread(ti:longint; tn:longint; fi:TSImg_FPImage; fr:TSImg_FPImage; fsn:TSImg_FPImage; ff:TSImg_FPImage; dt:double; dr:double; adr:double);
const unmask_fact = ln(0.5);            { ln(<Unmask factor at threshold>) }
var i,x,y,ya,yz : longint;
    sl,sl2      : longint;
    q,r,s,t,u,v : double;
    w,drl,dru   : double;
    sqn         : double;
    d,e,f       : ^simgfp;
    by          : boolean;
begin
    ya:=(fi.ny*ti) div tn;
    yz:=(fi.ny*(ti+1)) div tn - 1;
    dru:=abs(dr);
    if dr<0 then drl:=sqrt(1+sqr(dru))
    else drl:=1.05*dru;
    dt:=sqr(dt);
    sqn:=sqr(adr)/12; { mean square of simulated quantization noise }
    for y:=ya to yz do
    begin
        v:=random;
        by:=(y>1) and (y<fi.ny-2);
        i:=y*fi.sl;
        for x:=0 to fi.nx-1 do
        begin
            d:=fr.pt+i;
            e:=fsn.pt+i;
            f:=fi.pt+i;
            v:=frac(pi*v+0.876);                { cheap random }
            w:=(v-0.5)*adr;                     { simulated quantization noise }
            r:=d^+w;                            { residual }
            s:=e^+sqn; if s<simg_mina then s:=simg_mina;        { square noise }
            t:=f^-w;                            { image }
            u:=t-r;                             { reference }
            if u<0 then u:=0;
            if r>0 then u*=dru
            else u*=drl;
            q:=sqr(r)/(dt*s+sqr(u));
            if by and (x>1) and (x<fi.nx-2) then
            begin
                { 3x3 avg }
                sl:=fr.sl; sl2:=sl*2;
                r+=d[-sl-1] + d[-sl] + d[-sl+1] +
                   d[   -1]          + d[   +1] +
                   d[ sl-1] + d[ sl] + d[ sl+1];
                s+=e[-sl-1] + e[-sl] + e[-sl+1] +
                   e[   -1]          + e[    1] +
                   e[ sl-1] + e[ sl] + e[ sl+1];
                if s<simg_mina then s:=simg_mina;
                t+=f[-sl-1] + f[-sl] + f[-sl+1] +
                   f[   -1]          + f[   +1] +
                   f[ sl-1] + f[ sl] + f[ sl+1];
                u:=t-r;
                if u<0 then u:=0;
                if r>0 then u*=dru
                else u*=drl;
                u:=sqr(r)/(dt*s+sqr(u));
                if u>q then q:=u;
                { 5x5 avg }
                r+=d[-sl2-2] + d[-sl2-1] + d[-sl2] + d[-sl2+1] + d[-sl2+2] +
                   d[-sl -2]                                   + d[-sl +2] +
                   d[    -2]                                   + d[    +2] +
                   d[ sl -2]                                   + d[ sl +2] +
                   d[ sl2-2] + d[ sl2-1] + d[ sl2] + d[ sl2+1] + d[ sl2+2];
                s+=e[-sl2-2] + e[-sl2-1] + e[-sl2] + e[-sl2+1] + e[-sl2+2] +
                   e[ -sl-2]                                   + e[-sl +2] +
                   e[    -2]                                   + e[    +2] +
                   e[  sl-2]                                   + e[ sl +2] +
                   e[ sl2-2] + e[ sl2-1] + e[ sl2] + e[ sl2+1] + e[ sl2+2];
                if s<simg_mina then s:=simg_mina;
                t+=f[-sl2-2] + f[-sl2-1] + f[-sl2] + f[-sl2+1] + f[-sl2+2] +
                   f[-sl -2]                                   + f[-sl +2] +
                   f[    -2]                                   + f[    +2] +
                   f[ sl -2]                                   + f[ sl +2] +
                   f[ sl2-2] + f[ sl2-1] + f[ sl2] + f[ sl2+1] + f[ sl2+2];
                u:=t-r;
                if u<0 then u:=0;
                if r>0 then u*=dru
                else u*=drl;
                u:=sqr(r)/(dt*s+sqr(u));
                if u>q then q:=u;
            end
            else q*=2;
//          ff.pt[i]:=exp(unmask_fact*sqr(q));
            ff.pt[i]:=exp(unmask_fact*q);
            i+=1;
        end;
    end;
end;

{ Unmask unlikely values.
    fi:  background corrected image (input and output)
    fr:  residual input
    fsn: square noise
    fw:  weight (input and output)
    dt:  unmask threshold relative to variance
    dr:  absolute value is unmask threshold relative to reference. If <0 lower threshold is sqrt(1+sqr(dl)).
    adr: ADC resolution in ADU/e
}
function SImg_StackDB_unmask(cc:CSImg_Config; fi,fr,fsn,fw:TSImg_FPImage; dt,dr,adr:double):double;
const min_uf  = 5e-3;           { Minimum unmask factor }
      rmin_uf = 1/min_uf;
var ff          : TSImg_FPImage;
    i,tn        : longint;
    r           : double;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_unmask: fi='+SImg_InfoStr2(fi)+' fr='+SImg_InfoStr2(fr)+' fsn='+SImg_InfoStr2(fsn)+' fw='+SImg_InfoStr2(fw)+' dt='+real2str(dt,-2)+' dr='+real2str(dr,-2)+' adr='+real2str(adr));
    if not SImg_Recover(cc,fi)  then cc.error('SImg_StackDB_unmask: Invalid image: fi');
    if not SImg_Recover(cc,fr)  then cc.error('SImg_StackDB_unmask: Invalid image: fr');
    if not SImg_Recover(cc,fsn) then cc.error('SImg_StackDB_unmask: Invalid image: fsn');
    if not SImg_Recover(cc,fw)  then cc.error('SImg_StackDB_unmask: Invalid image: fw');
    if fi.nc<>1 then cc.error('SImg_StackDB_unmask: Invalid dimensions: fi');
    if  (fr.nx<>fi.nx) or  (fr.ny<>fi.ny) or  (fr.nc<>1) then cc.error('SImg_StackDB_unmask: Invalid dimensions: fr');
    if (fsn.nx<>fi.nx) or (fsn.ny<>fi.ny) or (fsn.nc<>1) then cc.error('SImg_StackDB_unmask: Invalid dimensions: fsn');
    if  (fw.nx<>fi.nx) or  (fw.ny<>fi.ny) or  (fw.nc<>1) then cc.error('SImg_StackDB_unmask: Invalid dimensions: fw');
    dt:=abs(dt);

    ff.pt:=nil;
    SImg_New(cc, ff, fw.nx,fw.ny,1, false);
    tn:=cc.threads;
    for i:=0 to tn-1 do
        dothread SImg_StackDB_unmask_thread(i,tn, fi,fr,fsn,ff, dt,dr,adr);
    threadwait;

    result:=0;
    for i:=0 to fw.size-1 do
    begin
        r:=ff.pt[i];
        result+=(1-r);
        if r>=min_uf then fw.pt[i]*=r
        else begin
            fi.pt[i]-=(1-r*rmin_uf)*fr.pt[i];
            fw.pt[i]*=min_uf;
        end;
    end;
    SImg_Free(cc,ff);

    SImg_DecDebugLevel;
end;


{*******************************************************************************}
{********* SImg_StackDB_calib **************************************************}
{*******************************************************************************}
{ ti,tn : thread index, number of threads
  fd    : destination image
  fw    : pixel weight for destination image
  sl    : source list
  res   : alignment parameters
  ra    : search radius for background
  buf   : tw element buffer for output: weighted sum of destination and reference brightness
  bbw   : if <>nil it brightness*weight, background*weight and weight are stored to this buffer
}
threadprocedure SImg_StackDB_calib_thread(ti:longint; tn:longint; fd:TSImg_FPImage; fw:TSImg_FPImage; sl:TSImg_RDImage; res:TSImg_AlignmentParams; ra:longint; buf:pdouble; bbw:pdouble);
const ref_mag_filter    = 2;            { underweight sources if <destination brightness>*res.fact differs more than this factor form reference brightness. Filters out unlikely reference magnitudes. }
var ia,iz,sri,sra       : longint;
    xc,yc,x,y,i,j,k,l   : longint;
    a,x1,y1,x2,y2       : double;
    d,e                 : ^simgfp;
label endloop;
begin
    ia:=(sl.ny*ti) div tn;
    iz:=(sl.ny*(ti+1)) div tn - 1;
    sra:=sqr(ra+1);
    sri:=sra div 2;
    buf[0]:=0;
    buf[1]:=0;
    for i:=ia to iz do
    begin
        l:=i*sl.sl;
        { coordinate transformation }
        x1:=sl.ptd[l  ]; x2:=sqr(x1);
        y1:=sl.ptd[l+1]; y2:=sqr(y1);
        { reference coordinates to destination coordinates }
        a:=res.d*(x2+y2) + res.c1*x1 + res.c2*y1;
        x1+=x1*a + res.c3*x2 + res.c5*y2;
        y1+=y1*a + res.c4*y2 + res.c6*x2;
        xc:=round(res.b11*x1 + res.b12*y1 + res.ax - res.xmin);
        yc:=round(res.b21*x1 + res.b22*y1 + res.ay - res.ymin);
        if (xc<ra) or (yc<ra) or (xc>=fd.nx-ra) or (yc>=fd.ny-ra) then goto endloop;
        { find background and brightness }
        x1:=0;  x2:=0;
        y1:=0;  y2:=0;
        k:=0;
        for y:=-ra to ra do
        begin
            j:=(y+yc)*fd.sl + xc;
            d:=fd.pt + j;
            e:=fw.pt + j;
            for x:=-ra to ra do
            begin
                j:=sqr(x)+sqr(y);
                if j<=sri then
                begin
                    x1+=d[x];
                    a:=e[x];
                    if a<simg_mina then goto endloop;
                    x2+=1/a;
                    k+=1;
                end
                else if j<sra then
                begin
                    a:=e[x];
                    y1+=d[x]*a;
                    y2+=a;
                end;
            end;
        end;
        if y2<simg_mina then goto endloop;
        y1/=y2;         { background }
        x1-=k*y1;       { integrated flux minus background }
        if x1<simg_mina then goto endloop;
        x2:=1/x2;       { weight }
        if y1>simg_mina then x2*=sqr(x1/(x1+y1));
        { write results }
        if bbw<>nil then
        begin
            j:=i*3;
            bbw[j  ]:=x1;
            bbw[j+1]:=y1;
            bbw[j+2]:=x2;
        end;

        if sl.ptd[l+4]<simg_mina then goto endloop; { only 0 or 1 expected }
        a:=sl.ptd[l+2];                 { reference brightness }
        if a<simg_mina then goto endloop;
        if res.fact>simg_mina then      { underweight unlikely results / invalid magnitudes in reference source list }
        begin
            y2:=x1*res.fact;
            if y2*ref_mag_filter<a then x2*=y2*ref_mag_filter/a
            else if y2>ref_mag_filter*a then x2*=ref_mag_filter*a/y2;
        end;
        buf[0]+=x1*x2;
        buf[1]+=a*x2;
//      writeln(stderr,'bd=',real2str(x1),' br=',real2str(sl.ptd[l+2])+'  bg=',real2str(y1)+'  k=',k,'  w=',real2str(x2));
endloop:
    end;
end;

{ Calibrate (destination) image factor using source list of reference
  fd    : destination image
  fw    : pixel weight for destination image
  sl    : source list
  res   : alignment parameters
  fwhm  : star fwhm in order to determine calculation radius
  bbw   : if <>nil: brightness*weight, background*weight and weight are stored to this buffer
  result: calibration factor or -1 if it can't be determined
}
function SImg_StackDB_calib(cc:CSImg_Config; fd,fw:TSImg_FPImage; sl:TSImg_RDImage; res:TSImg_AlignmentParams; fwhm:double; bbw:pdouble):double;
const ref_mag_warning   = 1.5;  { emit a warning if resulting factors differs by a factor more than this from previous factor and re-iterate (see ref_mag_filter in SImg_StackDB_calib_thread) }
var buf : ^double;
    i,j : longint;
threadvars;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_StackDB_calib: fd='+SImg_InfoStr2(fd)+' fw='+SImg_InfoStr2(fw)+' sl='+SImg_InfoStr2(sl)+' fw='+SImg_InfoStr2(fw)+' fwhm='+real2str(fwhm,-2));
    if not SImg_Recover(cc,fd) then cc.error('SImg_StackDB_calib: Invalid image: fd');
    if not SImg_Recover(cc,fw) then cc.error('SImg_StackDB_calib: Invalid image: fw');
    if not (SImg_Recover(cc,sl) and (sl.fmt=SImg_RawDouble) and (sl.nx>=5) and (sl.nc=1)) then cc.Error('SImg_StackDB_calib: Invalid reference source list');
    if fd.nc<>1 then cc.error('SImg_StackDB_calib: Invalid dimensions: fd');
    if (fw.nx<>fd.nx) or (fw.ny<>fd.ny) or (fw.nc<>1) then cc.error('SImg_StackDB_calib: Invalid dimensions: fw');

    if fwhm<1 then fwhm:=1;
    getmem(buf,cc.threads*2*sizeof(double));

    if bbw<>nil then
        for i:=0 to sl.ny*3-1 do
            bbw[i]:=0;

    for j:=1 to 4 do { up to 4 iterations }
    begin
        for i:=0 to cc.threads-1 do
            dothread SImg_StackDB_calib_thread(i,cc.threads, fd,fw, sl, res, round(fwhm*2+0.45), buf+i*2, bbw);
        threadwait;
        for i:=1 to cc.threads-1 do
        begin
            buf[0]+=buf[i*2];
            buf[1]+=buf[i*2+1];
        end;
//        SImg_StackDB_calib_thread(0,1, fd,fw, sl, res, round(fwhm*2+0.45), buf, bbw);
        if buf[0]<simg_mina then result:=-1
        else result:=buf[1]/buf[0];

        if (result<0) or ((result*ref_mag_warning>res.fact) and (result<res.fact*ref_mag_warning)) then break;

        cc.warning(cc.name+': Calibrated image factor '+real2str(result,5)+' differs to much from ' + select(j=1,'initial','previous') + ' factor '+real2str(res.fact,5) + select(res.fact>simg_mina,'  (quotient '+real2str(result/res.fact,5)+')','')+': Re-iterate');
        res.fact:=result; { re-iterate with new filter }
        bbw:=nil;         { already calculated in 1st iteration }
    end;

//    cc.error('Factor: '+real2str(result));

    freemem(buf,cc.threads*2*sizeof(double));

    SImg_DecDebugLevel;
end;
