{  simg -- visualization library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{$mode objfpc}
{$packrecords C}
unit siv;

interface
uses glib2,gtk2,gdk2,gdk2pixbuf,simage;

{$I+}

{ ************************************************************************* }
{ ******** siv-misc ******************************************************* }
{ ************************************************************************* }
const siv_history_size = 16;
      siv_history_filename = 'visualhist';

type CSIV_History=class
       entries : array[0..siv_history_size-1] of simgstring;
       constructor init;                  { init with empty table }
       constructor init(fn:simgstring);   { try to read entries from file fn }
       constructor init(h:CSIV_History);  { init with values from h }
       procedure write(fn:simgstring);    { (try to) write values to file fn }
       procedure insert(s:simgstring);    { insert entry s }
       procedure remove(i:longint);       { remove value i }
       function search(s:simgstring):longint; { searches entry s; return entry number if found, -1 otherwise }
       function count:longint;            { count the values }
       end;

type CSiv_HScale=class  { like GtkHSkale but with spin buttons }
      hscale    : PGtkWidget;     { GtkHScale widget }
      thelabel  : PGtkWidget;     { label }
      adj       : PGtkAdjustment; { adjustment }
      constructor init(parent:PGtkWidget; value,lower,upper,step_increment,page_increment:double);  { parameters like gtk_adjustment_new }
      destructor done;
      function widget:PGtkWidget;       { returns hbox }
      procedure set_label(xsize:longint; s:simgstring); { init / change the label }
      procedure set_tooltip(s:simgstring);  { set the tooltip }
      procedure connect(proc:TGtkSignalFunc; data:gpointer);  { connect proc to 'value_changed' signal of adj }
      procedure stepcheck;              { check whether step button is pressed }
           { Pressing a step button causes a 'value_changed' signal. stepcheck should be called after
             the image was re-calculated }
      procedure show;                   { show hbox }
      procedure hide;                   { hide hbox }
     private
      hbox            : PGtkWidget;  { the hBox containing everything }
      vbox            : PGtkWidget;  { the vbox containing bottons }
      step_up         : PGtkWidget;  { step up button (GTKEventBox widget) }
      step_down       : PGtkWidget;  { step down button (GTKEventBox widget) }
      arrow_up        : PGtkWidget;  { PGtkImage, up arrow }
      arrow_down      : PGtkWidget;  { PGtkImage, down arrow }
      arrow_up_xpm    : array[0..1] of PGdkPixbuf;  { up arrows }
      arrow_down_xpm  : array[0..1] of PGdkPixbuf;  { down arrows }
      state_up        : boolean;     { state of up button }
      state_down      : boolean;     { state of down button }
      steptime        : longint;
      end;

var siv_global_history  : CSIV_History;
    siv_config          : CSImg_Config;

procedure siv_init(cc:CSImg_Config);
procedure siv_main;
procedure siv_done;
procedure siv_error(mes:simgstring);
procedure siv_warning(mes:simgstring);
procedure siv_set_rgb_color(var c:TGdkColor; r,g,b:double);
function siv_label_new(text:simgstring):PGtkWidget;
function siv_button_new_with_label(text:simgstring):PGtkWidget;
function siv_toggle_button_new_with_label(text:simgstring):PGtkWidget;

{ ************************************************************************* }
{ ******** siv-image ****************************************************** }
{ ************************************************************************* }
const siv_hist_ysize       = 100;      { y size of histogram images }
      siv_histl_xsize      = 288;      { x size of left histogram image }
      siv_histr_xsize      = 192;      { x size of right histogram image (must be <=siv_histl_xsize }
      siv_maxscale         = 9;        { maximum scale (-1) }
      siv_hist_stretchexp  = 0.7;      { minimal stretch of discretization intervals (fi -> img16) }
      siv_hist_bias        = 1e-3;     { to avoid several problems with small (<1) histogram values }
      siv_hist_smoth       = 0.25;     { smoothing avoids some strange effects with line histograms }
      siv_histed_res       = 4096;     { resolution of temporary equidistant discretized histogram }

type CSIV_Image=class
            { Parameters }
            scale          : longint;    { -3 means upscale 4:1; 0 means 1:1, 3 means down-scale 1:4 }
            red,green,blue : longint;    { colors }
            ll,ul          : double;     { lower/ upper limit (0...65535) }
            bet,gam        : double;     { beta, gamma }
            inv            : boolean;    { inverted mode }
            log            : boolean;    { logarithmic mode (for histograms) }
            update_scale   : boolean;    { scale value changed }
            update_limits  : boolean;    { values of ll/ul changed }
            update_colors  : boolean;    { values of red/green/blue changed }
            update_bg      : boolean;    { beta, gamma or inv changed }
            redraw_histl   : boolean;    { redrawing of hist16 needed }
            redraw_histr   : boolean;    { redrawing of hist8 needed }
            { rgb data for drawing histograms and other infos }
            histlimg       : ^byte;      { rgb image that shows the histogram of the 16 bit image and the transform curve (siv_hist12_xsize x siv_hist_ysize x 3 bytes) }
            histrimg       : ^byte;      { rgb image that shows the histogram of destination image (siv_hist8_xsize x siv_his_ysize x 3 bytes) }
            ql,qu          : double;     { lower and upper quantile }
            { misc }
            title,name     : simgstring;                       { title of the image }
            procedure update;
            constructor init(fi:TSImg_FPImage;fn:simgstring);  { fi : image, fn:filename (used for title) }
            constructor init(anc:CSIV_Image);                  { clone variant }
            procedure load(fi:TSImg_FPImage;fn:simgstring);    { inits fi; fn:filename (used for title) }
            procedure load(anc:CSIV_Image);                    { clone valiant }
            function scaledWidth:longint;                        { scaled original image width }
            function scaledHeight:longint;                       { scaled original image height }
            procedure getRGB(var buf:pbyte; var x0,y0,width,height:longint);
            destructor done;
        private
//          inv2          : boolean;
            { image properties }
            minv,maxv     : double;      { minimal/maximal value }
            mean,rms,meandev : double;
            { img16 stuff }
            img16         : TSimg_RDImage;  { 16 bit discretized image (not equidistant), as raw data image }
            delta         : double;
{ x    : normalized value of input image: 0..1
  x_   : average value
  y    : normalize discretized value
  y = x*(1-delta)/(1-delta*x)
  delta = 2 - 1/x_  (delta < 1)
  x = y/(1+delta*y-delta) }
            dist16        : array[0..65536] of longint;  { distribution of img16 (interval i goes form dp[i] to dp[i+1]; hist16[i]=dist16[i+1]-dist16[i]) }
            hist16c       : ^longint;       { histogram of img16 (separated colors) }
            { img8 stuff }
            xlat          : array[0..65535] of byte;
            { left and right histogram }
            histlc        : ^longint;    { left histogram (separated colors) }
            histrc        : ^longint;    { right histogram (separated colors) }
            histtrans     : array[0..siv_histr_xsize] of longint;
            { history marker }            { for drawing lines in histl and histr }
            marker_x,marker_y : longint;
            procedure img16_img8;        { converts img_16b to img_8, calculate hist8a and hist8c }
            procedure drawhistl;         { calculate histlimg }
            procedure drawhistr;         { calculate histrimg }
            procedure history_markers_int;
            procedure history_markers(x,y:longint);  { reset marker lines in hist16 and hist8 }
        end;


{ ************************************************************************* }
{ ******** CSIV_Draw ****************************************************** }
{ ************************************************************************* }
const siv_marker_frequency   : double = 2.0*pi*1.0;  { 1.0 Hz}
      siv_marker_update_time : longint = 100;        { update marker every every ... ms }

type TSivDraw=record
            GtkDrawingArea         : TGtkDrawingArea;
            { image data }
            image_width            : longint;        { total width of the image }
            image_height           : longint;        { total height of the image }
            { image data }
            hadj,vadj              : PGtkAdjustment; { adjustments }
            { image data }
            rgb                    : ^byte;          { rgb data }
            rgb_free               : ^byte;          { rgb that need to be freed }
            rgb_width,rgb_height   : longint;        { size of rgb data }
            rgb_x0,rgb_y0          : longint;        { position of the rgb data }
            { TSIV_Image }
            image                  : CSIV_Image;     { dynamic data source }
            { marker }
            marker_x,marker_y      : longint;        { position of marker }
            marker_time            : double;         { last time when marker was drawn }
            marker_id              : longint;        { id of then idle function }
            { attributes }
            background_gc          : PGdkGC;         { background gc (nil for none) }
            shadow_type            : TGtkShadowType; { shadow type }
            extra_width            : longint;        { shadow widh + extra_space }
            extra_height           : longint;        { shadow height + extra_space }
            { private }
            buf,back_buf           : PGdkPixmap;     { buffers }
            { position of rgb }
            left,top               : longint;        { position of first pixel of (drawn) rgb }
            right,bottom           : longint;        { position of last pixel of (drawn) rgb + 1 }
            { last position of drawing area }
            last_x,last_y,last_width,last_height : longint;
        end;
        PSivDraw=^TSivDraw;

        TSivDrawClass=record
            GtkDrawingAreaClass : TGtkDrawingAreaClass;
            old_unrealize       : procedure(widget:PGtkWidget); cdecl;
            old_size_allocate   : procedure(widget:PGtkWidget; allocation:PGdkRectangle); cdecl;
        end;
        PSivDrawClass=^TSivDrawClass;

{ Attributes:
    data          : rgb data
    widht,height  : dimensions
    bg            : background (nil for none)
    shadow        : shadow type
    extraspcae    : extra border }
function siv_draw_new(data:pbyte; width,height:longint; bg:PGdkGC; shadow:TGtkShadowType; extra_space:longint):PGtkWidget;  // initialize with fixed data
function siv_draw_new(img:CSIV_Image; bg:PGdkGC; shadow:TGtkShadowType; extra_space:longint):PGtkWidget;  // initialize with dynamic data
procedure siv_draw_set_hadj(draw:PSivDraw; hadj: PGtkAdjustment);
procedure siv_draw_set_vadj(draw:PSivDraw; vadj: PGtkAdjustment);
procedure siv_draw_reset(draw:PSivDraw; data:pbyte; width,height:longint);  // reset fixed data
procedure siv_draw_reset(draw:PSivDraw);                                    // reset dynamic data
procedure siv_draw_redraw(draw:PSivDraw);

procedure siv_draw_marker_expose(draw:PSIVDraw;x,y:longint);   { expose marker at position (marker_x, marker_y) and (x,y) }
procedure siv_draw_marker_move(draw:PSIVDraw; x,y:longint; expose:boolean);  { move marker to position (x,y); expose if expose=true }


{ ************************************************************************* }
{ ******** CSIV_ImageWidget *********************************************** }
{ ************************************************************************* }
type CSIV_ImageWidget = class
            widget             : PGtkWidget;     { ScrolledWindow widget }
            draw               : PGtkWidget;     { SivDraw widget }
            mask               : PGdkPixmap;     { background pattern }
            hadj,vadj          : PGtkAdjustment; { horizontal / vertical adjustment }
            hsb,vsb            : PGtkWidget;     { horizontal / vertical scroll bar }
            sb_height,sb_width : longint;        { height / width of horizontal / vertical scroll bar }
            image              : CSIV_Image;
            constructor init(parent:PGtkWidget; _image:CSIV_Image);
            destructor done;                  { destructor }
            procedure redraw;                 { redraw image }
            procedure enable_fit_request;     { next size_request returns the image size (for fitting) }
            procedure show;
            procedure hide;
        private
            b3_x,b3_y          : longint;     { last position during button3 pressed }
            b3_pressed         : boolean;     { button3 pressed }
            fit_request        : boolean;     { next size_request returns the image size (for fitting) }
        end;


{ ************************************************************************* }
{ ******** CSIV_Visual **************************************************** }
{ ************************************************************************* }
const siv_load_nothing = 0;  { load nothing }
      siv_load_fi      = 1;  { load image (as variable fi) }
      siv_load_fn      = 2;  { load image file }
      siv_load_clone   = 3;  { clone image from anc }

type CSIV_Visual=class;
     TSIV_ColorBuf=record    { used as data for color menu items }
       c      : longint;     { number of color }
       visual : CSIV_Visual; { owner }
       end;
     PSIV_ColorBuf = ^TSIV_ColorBuf;

     CSIV_Visual=class
       { constructor parameters:
          load : load strategy
          fi   : image (optional, can be invalid)
          adir : directory (optional, '' allowed)
          fn   : file name (used for title and file loading)
          h    : descendant history (nil allowed)
          shb  : if true: show histogram bar
          scb  : if true: show control bar
          cmte : if true: combine multi time exposure (at loading)
          anc  : ancestor for inheriting visualization parameters (optional, nil allowed)  }
            { window }
            window            : PGtkWidget;       { main window }
            vbox              : PGtkWidget;       { main vbox }
            { toolbar }
            toolbar           : PGtkWidget;       { tool bar (quit, open, ... ) }
            bopen             : PGtkWidget;       { open button }
            breload           : PGtkWidget;       { reload button }
            bclone            : PGtkWidget;       { clone button }
            bfit              : PGtkWidget;       { fit window button }
            binfo             : PGtkWidget;       { info button }
            bquit             : PGtkWidget;       { quit button }
            bcombinemte       : PGtkWidget;       { combine mte button }
            bnocombinemte     : PGtkWidget;       { not combine mte button }
            { history }
            history           : CSIV_History;     { file history }
            histcombo         : PGtkWidget;       { GtkCombo widget with history }
            { info bar }
            infobar           : PGtkWidget;       { shows position and color at mouse pointer }
            pointer_label     : PGtkWidget;       { label widget for mouse pointer }
            marker_label      : PGtkWidget;       { label widget for marker }
            { histbar }
            histbar           : PGtkWidget;       { control bar (histogram, resize, ... }
            bhide_hb          : PGtkWidget;       { hide control bar button }
            bshow_hb          : PGtkWidget;       { show control bar button }
            { controlbar }
            controlbar        : PGtkWidget;       { control bar (histogram, resize, ... }
            bhide_cb          : PGtkWidget;       { hide control bar button }
            bshow_cb          : PGtkWidget;       { show control bar button }
            { ul / ll }
            ul_scale          : CSIV_HScale;      { upper limit slider }
            ll_scale          : CSIV_HScale;      { lower limit slider }
            { other sliders }
            bet_scale         : CSIV_HScale;      { beta slider }
            gam_scale         : CSIV_HScale;      { gamma slider }
            scale_scale       : CSIV_HScale;      { scale slider }
            { toggle buttons }
            inv_tb            : PGtkWidget;       { inverted toggle button }
            swap_b            : PGtkWidget;       { swap limits button }
            log_tb            : PGtkWidget;       { logarithmic toggle button }
            { image widget }
            iw                : CSIV_ImageWidget; { image widget (scrollable image) }
            image             : CSIV_Image;       { CSIVImage }
            { hist16 }
            histl            : PGtkWidget;        { SivDraw widget for left histogram }
            { hist8 }
            histr             : PGtkWidget;       { SivDraw widget for right histogram }
            { marker }
            marker_x,marker_y : longint;          { position of marker }
            { info window }
            info_window       : PGtkWidget;
            { misc }
            dir               : simgstring;
            combinemte        : boolean;
            next              : CSIV_Visual;      { to make a list }
            { policy change  handler }
            policyChangeID    : longint;
            constructor init(load:longint; fi:TSImg_FPImage; fn,adir:simgstring; h:CSiv_History; scb,shb,cmte:boolean; anc:CSIV_Visual);
            destructor done;
            procedure redraw;                   { update + redraw }
            procedure fit;                      { fit window }
            procedure opennew(fn:simgstring);     { (try to) open image file fn in new window  }
            procedure setmarker(x,y:longint);
        private
            { colors }
            red_menu      : PGtkWidget;      { red menu }
            green_menu    : PGtkWidget;      { green menu }
            blue_menu     : PGtkWidget;      { blue menu }
            function create_color_menu(default:longint):PGtkWidget;
        end;


{ ************************************************************************* }
{ ************************************************************************* }
{ ******** IMPLEMENTATION ************************************************* }
{ ************************************************************************* }
{ ************************************************************************* }
implementation
uses sysutils,classes,compatun,siv_utils,siv_pixmaps,unixutil;

{$include config.inc}

const xsize=512;  { x size of for test image }
      ysize=512;  { y size of for test image }

{$include siv-misc.inc}
{$include siv-image.inc}
{$include siv-draw.inc}
{$include siv-imagewidget.inc}
{$include siv-visual.inc}

initialization
    SIV_InitSys;

finalization
    SIV_DoneSys;

end.
